/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */

/* -----------------------------------------------------------------------------------------------------------------------------------------------
   Functions called by C that contain calls to Java methods (See GetMethodID() function).
   Use "I" for int, "B" for byte, "S" for short, "J" for long, "F" for float, "D" for double, "C" for char, and "Z" for boolean and "V" for void.
   For arrays, include a prefix "[", e.g., "[Ljava/lang/Object;" for an array of Object; "[I" for an array of int.
   For strings,  the class descriptor is "Ljava/lang/String;". To convert char* (C) to java string see NewStringUTF() function.
*/

/**
   Refreshing the Graphical view.
 */
extern void
C2Java_refresh (boolean hasChanged)
{
  currentEnv->CallVoidMethod (currentObj, mid_Refresh, hasChanged);
}


///
extern void
C2Java_drawPolyline (int x[], int y[], int nb, int col, int width)
{

  // Lbesnard :
  // variant array : not accepted by visual C++;
  //   C2Java_drawPolyline is called with nb = 2;
  // int i; jint tamponx[nb]; for (i=0; i<nb; i++) {tamponx[i] = (jint)x[i];}

  jint tamponx[_ARRAY_MAX_SIZE_];
  mkJintArray (x, nb, tamponx);

  jintArray posx = currentEnv->NewIntArray (nb);
  currentEnv->SetIntArrayRegion (posx, 0, nb, tamponx);

  //  jint tampony[nb]; for (i=0; i<nb; i++) {tampony[i] = (jint)y[i];}
  jint tampony[_ARRAY_MAX_SIZE_];
  mkJintArray (y, nb, tampony);

  jintArray posy = currentEnv->NewIntArray (nb);
  currentEnv->SetIntArrayRegion (posy, 0, nb, tampony);
  currentEnv->CallVoidMethod (currentObj, mid_DrawPolyline, posx, posy, nb, col, width);
}

///
extern void
C2Java_drawLine (int a, int b, int c, int d, int col, int width)
{
  currentEnv->CallVoidMethod (currentObj, mid_DrawLine, (jint) a, (jint) b, (jint) c, (jint) d, (jint) col, (jint) width);
}

///
extern void
C2Java_FillPolygon (int x[], int y[], int nbPoints, int col, int c)
{
  //  jint tamponx[nbPoints]; jint tamponx[10]; for (i=0; i< nbPoints; i++) tamponx[i]=(jint)x[i];
  jint tamponx[_ARRAY_MAX_SIZE_];
  mkJintArray (x, nbPoints, tamponx);

  jintArray posx = currentEnv->NewIntArray (nbPoints);
  currentEnv->SetIntArrayRegion (posx, 0, nbPoints, tamponx);

  // jint tampony[nbPoints]; jint tampony[10];  for (i=0; i< nbPoints; i++) tampony[i]=(jint)y[i];
  jint tampony[_ARRAY_MAX_SIZE_];
  mkJintArray (y, nbPoints, tampony);

  jintArray posy = currentEnv->NewIntArray (nbPoints);
  currentEnv->SetIntArrayRegion (posy, 0, nbPoints, tampony);
  currentEnv->CallVoidMethod (currentObj, mid_FillPolygon, posx, posy, (jint) nbPoints, (jint) col, (jint) c);
}

///
extern void
C2Java_drawPolygon (int x[], int y[], int nbPoints, int col, int c)
{
  // jint tamponx[nbPoints]; jint tamponx[10]; for (i=0; i< nbPoints; i++) tamponx[i]=(jint)x[i];
  jint tamponx[_ARRAY_MAX_SIZE_];
  mkJintArray (x, nbPoints, tamponx);

  jintArray posx = currentEnv->NewIntArray (nbPoints);
  currentEnv->SetIntArrayRegion (posx, 0, nbPoints, tamponx);
  // jint tampony[nbPoints]; jint tampony[10]; for (i=0; i< nbPoints; i++) tampony[i]=(jint)y[i];

  jint tampony[_ARRAY_MAX_SIZE_];
  mkJintArray (y, nbPoints, tampony);

  jintArray posy = currentEnv->NewIntArray (nbPoints);
  currentEnv->SetIntArrayRegion (posy, 0, nbPoints, tampony);
  currentEnv->CallVoidMethod (currentObj, mid_DrawPolygon, posx, posy, (jint) nbPoints /*, (jint)col, (jint)c */ );
}

///
extern void
C2Java_drawRectangle (int a, int b, int c, int d, int col, int ep)
{
  currentEnv->CallVoidMethod (currentObj, mid_DrawRectangle, (jint) a, (jint) b, (jint) c, (jint) d, (jint) col, (jint) ep);
}

///
extern void
C2Java_FillRectangle (int a, int b, int c, int d, int col)
{
  currentEnv->CallVoidMethod (currentObj, mid_C2Java_FillRectangle, a, b, c, d, col);
}

///
extern void
C2Java_drawString (const char *c, int x, int y, int couleur, int typeFont)
{
  jstring chaine;
  chaine = currentEnv->NewStringUTF (c);
  currentEnv->CallVoidMethod (currentObj, mid_DrawString, chaine, x, y, couleur, typeFont);
}

///
extern void
C2Java_drawImage (const char *image, int x, int y)
{
  jstring chaine;
  chaine = currentEnv->NewStringUTF (image);
  currentEnv->CallVoidMethod (currentObj, mid_DrawImage, chaine, x, y);
}

///
extern int
C2Java_readMenuButton ()
{
  int i;
  i = currentEnv->CallIntMethod (currentObj, mid_C2Java_readMenuButton);
  return (i);
}


///
extern void
C2Java_razMenuButton ()
{
  currentEnv->CallVoidMethod (currentObj, mid_C2Java_razMenuButton);
}

///
extern int
C2Java_getMouseButton ()
{
  int i;
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getMouseButton", "()I");
  i = currentEnv->CallIntMethod (currentObj, mid);
  return (i);
}

///
extern boolean
C2Java_isMouseButtonClicked ()
{
  boolean b;
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "isMouseButtonClicked", "()Z");
  if (mid != null)
    b = currentEnv->CallBooleanMethod (currentObj, mid);
  else
    b = pk_false;
  return (b);
}

///
extern int
C2Java_getX ()
{
  int i;
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getXOrd", "()I");
  i = currentEnv->CallIntMethod (currentObj, mid);
  return (i);

}

///
extern int
C2Java_getY ()
{
  int i;
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getYOrd", "()I");
  i = currentEnv->CallIntMethod (currentObj, mid);
  return (i);
}

///
extern void
C2Java_setMoveMouseEvent ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "setMoveMouseEvent", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}

///
extern void
C2Java_resetMoveMouseEvent ()
{
  currentEnv->CallVoidMethod (currentObj, mid_C2Java_resetMoveMouseEvent);
}


///
extern void
C2Java_textHide (text_win_type w)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "textHide", "(I)V");
  currentEnv->CallVoidMethod (currentObj, mid, (jint) w);
}

///
extern void
C2Java_textShow (text_win_type w)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "textShow", "(I)V");
  currentEnv->CallVoidMethod (currentObj, mid, (jint) w);
}


///
extern void
C2Java_resetText (text_win_type w)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "resetText", "(I)V");
  currentEnv->CallVoidMethod (currentObj, mid, (jint) w);
}

///
extern int
C2Java_getTextColumns (text_win_type w)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getColumns", "(I)I");
  return (currentEnv->CallIntMethod (currentObj, mid, (jint) w));
}


///
extern void
C2Java_textAppend (text_win_type w, const char *text)
{
  jstring chaine;
  chaine = currentEnv->NewStringUTF (text);
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "appendText", "(ILjava/lang/String;)V");
  currentEnv->CallVoidMethod (currentObj, mid, (jint) w, chaine);
}

///
// extern void C2Java_setTitle(text_win_type w, const char *text)
// {
  // jstring chaine;
  // chaine = currentEnv->NewStringUTF(text);
  // mid = currentEnv->GetMethodID(currentEnv->GetObjectClass(currentObj),"setTitle","(ILjava/lang/String;)V");
  // currentEnv->CallVoidMethod(currentObj, mid, (jint)w, chaine);
// }

///
extern void
C2Java_showNamePort (char *text)
{
  jstring chaine;
  chaine = currentEnv->NewStringUTF (text);
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "showNamePort", "(Ljava/lang/String;)V");
  currentEnv->CallVoidMethod (currentObj, mid, chaine);
}


///
extern void
C2Java_showPortHide ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "showPortHide", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}

///
extern typChoice
C2Java_getUserChoice ()
{
  int choice;
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getUserChoice", "()I");
  choice = currentEnv->CallIntMethod (currentObj, mid);
  if (choice == 1) return OK;
  if (choice == 2) return SKIP;
  return CANCEL;
}


///
extern void
C2Java_getIdentsForReplace_old (ptrepstr * ident1, ptrepstr * ident2)
{
  jclass currentCls;
  jstring jchaine;

  currentCls = currentEnv->GetObjectClass (currentObj);
  mid = currentEnv->GetMethodID (currentCls, "getString1", "()Ljava/lang/String;");
  jchaine = (jstring) currentEnv->CallObjectMethod (currentObj, mid);
  *ident1 = nilstr;
  *ident2 = nilstr;
  if (jchaine != null)
    {
      const char *string1 = ((currentEnv)->GetStringUTFChars (jchaine, 0));
      *ident1 = string2repstr ((char *) string1);
      mid = currentEnv->GetMethodID (currentCls, "getString2", "()Ljava/lang/String;");
      jchaine = (jstring) currentEnv->CallObjectMethod (currentObj, mid);
      if (jchaine != null)
	{
	  const char *string2 = ((currentEnv)->GetStringUTFChars (jchaine, 0));
	  *ident2 = string2repstr ((char *) string2);
	}
    }
}

extern void
C2Java_getIdentsForReplace (ptident * ident1, ptident * ident2)
{
  jclass currentCls;
  jstring jchaine;

  currentCls = currentEnv->GetObjectClass (currentObj);
  mid = currentEnv->GetMethodID (currentCls, "getString1", "()Ljava/lang/String;");
  jchaine = (jstring) currentEnv->CallObjectMethod (currentObj, mid);
  *ident1 = NILIDENT;
  *ident2 = NILIDENT;
  if (jchaine != null)
    {
      const char *string1 = ((currentEnv)->GetStringUTFChars (jchaine, 0));
      *ident1 = string2ident ((char *) string1);
      mid = currentEnv->GetMethodID (currentCls, "getString2", "()Ljava/lang/String;");
      jchaine = (jstring) currentEnv->CallObjectMethod (currentObj, mid);
      if (jchaine != null)
	{
	  const char *string2 = ((currentEnv)->GetStringUTFChars (jchaine, 0));
	  *ident2 = string2ident ((char *) string2);
	}
    }
}

///
extern char *
C2Java_getText (text_win_type w)
{
  jstring jchaine;
  char *result = 0;
  jbyteArray bytes = 0;
  // printf("DEBUG entering C2Java_getText\n"); fflush(stdout);
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getText", "(I)Ljava/lang/String;");
  jchaine = (jstring) currentEnv->CallObjectMethod (currentObj, mid, (jint) w);
  //  const char *result = (currentEnv)->GetStringUTFChars(jchaine, 0);

  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getText1", "(Ljava/lang/String;)[B");
  bytes = (jbyteArray) currentEnv->CallObjectMethod (currentObj, mid, jchaine);
  jint len = currentEnv->GetArrayLength (bytes);
  result = (char *) malloc (len + 1);
  currentEnv->GetByteArrayRegion (bytes, 0, len, (jbyte *) result);
  result[len] = 0;
  currentEnv->DeleteLocalRef (bytes);
  return ((char *) result);
}


///
extern text_win_type
C2Java_getLastActiveTextWin ()
{
  int w;
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getLastActiveTextWin", "()I");
  w = currentEnv->CallIntMethod (currentObj, mid);
  return w;
  // if (w == J_CanvasText) return(C_CanvasText); else return(C_HierarText);
}

// demande d'affichage d'un browser suivant action
/// action peut prendre les valeurs J_load (chargement graphique), J_parse (texte),
extern void
displayinBrowser (int action)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "displayinBrowser", "(I)V");
  currentEnv->CallVoidMethod (currentObj, mid, (jint) action);
}

// demande d'affichage d'un browser suivant action
// action peut prendre les valeurs 
/// J_saveAllRoots (sauvegarde complète graphique), J_save ((sauvegarde courante graphique)...
extern void
displayoutBrowser (int action)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "displayoutBrowser", "(I)V");
  currentEnv->CallVoidMethod (currentObj, mid, (jint) action);
}


///
extern void
C2Java_displayFont ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "displayFont", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}


///
extern int
C2Java_getSizeFontH (int type)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getSizeFontH", "(I)I");
  return (currentEnv->CallIntMethod (currentObj, mid_C2Java_getSizeFontH, (jint) type));
}


///
extern int
C2Java_getSizeFontL (int type)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getSizeFontL", "(I)I");
  return (currentEnv->CallIntMethod (currentObj, mid_C2Java_getSizeFontL, (jint) type));
}


///JTr
extern void
C2Java_appendNewJNode (const char *nom, int noMod, boolean setInAtEnd)
{
  jstring chaine;

  chaine = currentEnv->NewStringUTF (nom);
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "appendNewJNode", "(Ljava/lang/String;IZ)V");
  currentEnv->CallVoidMethod (currentObj, mid, chaine, (jint) noMod, setInAtEnd);
}


/**
  Not used
*/
// extern void insertObject(char *nom, int noProc, int rang)
// {
//     jstring chaine;
//     chaine = currentEnv->NewStringUTF(nom);
//     mid = currentEnv->GetMethodID(currentEnv->GetObjectClass(currentObj),"insertObject","(Ljava/lang/String;II)V");  
//     currentEnv->CallVoidMethod(currentObj, mid, chaine, (jint)noProc);
// }


///
extern void
C2Java_deleteSelectedNode ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "deleteSelectedNode", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}


///
extern void
C2Java_upTempSelectedNode ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "upTempSelectedNode", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}

///
extern void
C2Java_SetCurrentTree (int boxNumber)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "SetCurrentTree", "(I)V");
  currentEnv->CallVoidMethod (currentObj, mid, (jint) boxNumber);
}


///
extern void
C2Java_initSelectedNode (boolean afterCompil)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "initSelectedNode", "(Z)V");
  currentEnv->CallVoidMethod (currentObj, mid, afterCompil);
}



///
extern void
C2Java_displayError (int no)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "displayError", "(I)V");
  currentEnv->CallVoidMethod (currentObj, mid, no);
}

///
extern void
C2Java_displayParsingError (char *str)
{
  jstring js;
  js = currentEnv->NewStringUTF (str);
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "displayParsingError", "(Ljava/lang/String;)V");
  currentEnv->CallVoidMethod (currentObj, mid, js);
}




///
extern void
C2Java_displayCurrentAction (int action)
{
  currentEnv->CallVoidMethod (currentObj, mid_C2Java_displayCurrentAction, action);
}


///
extern void
C2Java_displayCompilerFunct ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "displayCompilerFunct", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}


///
extern void
C2Java_displayAutoCompilerFunct ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "displayAutoCompilerFunct", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}


///
extern void
C2Java_displayCompilerDiag ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "displayCompilerDiag", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}

///
extern int
C2Java_getCompilerFunct ()
{
  return (currentEnv->CallIntMethod (currentObj, mid_C2Java_getCompilerFunctionality));
}


///
extern void
C2Java_setColorCompilerFunct (int func, int col)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "setColorCompilerFunct", "(II)V");
  currentEnv->CallVoidMethod (currentObj, mid, func, col);
}


///
extern void
C2Java_compilerWinClose ()
{
}


/// Postscript print
extern void
C2Java_PostscriptPrint ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "print", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}


///
extern void
C2Java_helpShow ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "helpShow", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}


/// asks to Java the creation of the icons bar window for the library libRoot.
extern void
C2Java_createLibBar (char *filename, ntrees libRoot, char *filenamews, char *libname)
{
  jstring fnws, ln, fn;

  fn = currentEnv->NewStringUTF (filename);
  fnws = currentEnv->NewStringUTF (filenamews);
  ln = currentEnv->NewStringUTF (libname);
  mid =
    currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "createLibBar", "(JLjava/lang/String;Ljava/lang/String;Ljava/lang/String;)V");
  currentEnv->CallVoidMethod (currentObj, mid, (jlong) libRoot, fn, fnws, ln);
}


///
extern ntrees
C2Java_getRefLib ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getRefLib", "()J");
  return ((ntrees) (currentEnv->CallLongMethod (currentObj, mid)));
}

///
extern int
C2Java_getIconIndex ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getIconIndex", "()I");
  return (currentEnv->CallIntMethod (currentObj, mid));
}

///
extern int
C2Java_getHightIcon ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getHightIcon", "()I");
  return (currentEnv->CallIntMethod (currentObj, mid));
}

///
extern int
C2Java_getLenghtIcon ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getLenghtIcon", "()I");
  return (currentEnv->CallIntMethod (currentObj, mid));
}

/** PLG
    @return 0 when operation is canceled or errored
    @return -1 when a file name is provided and the current action is load and current complement is J_before
    @return 1 when a file name is provided
*/
extern int
C2Java_getFileNameFromBrowser (boolean isGraphic, boolean OpenOp, char **fileName)
{
  jstring jchaine;
  int browserOK;

  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "isBrowserOk", "(ZZ)I");
  browserOK = (currentEnv->CallBooleanMethod (currentObj, mid, isGraphic, OpenOp));
  if (browserOK == 0) return 0;
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getBrowserFileName", "()Ljava/lang/String;");
  jchaine = (jstring) currentEnv->CallObjectMethod (currentObj, mid);
  const char *str = ((currentEnv)->GetStringUTFChars (jchaine, 0));
  *fileName = (char *) str;
  return browserOK;
}

///
// extern char * C2Java_getUserCurrentPath()
// {
//   jstring jchaine;
//   mid = currentEnv->GetMethodID(currentEnv->GetObjectClass(currentObj),"getCurrentPath","()Ljava/lang/String;");
//   jchaine = (jstring)currentEnv->CallObjectMethod(currentObj, mid);
//   const char *str = ((currentEnv)->GetStringUTFChars(jchaine, 0));
//   return((char *)str);
// }


///
extern boolean
C2Java_isLoadedLibrary (char *filename)
{
  jstring chaine;
  chaine = currentEnv->NewStringUTF (filename);
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "isLoadedLibrary", "(Ljava/lang/String;)Z");
  return currentEnv->CallBooleanMethod (currentObj, mid, chaine);
}

///
extern void
C2Java_removeSelectedLib (char *filename)
{
  jstring vstr;
  vstr = currentEnv->NewStringUTF (filename);
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "removeSelectedLib", "(Ljava/lang/String;)V");
  currentEnv->CallVoidMethod (currentObj, mid, vstr);
}


///
extern char *
C2Java_getSelectedFilenameLibrary ()
{
  jstring vstr;
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getSelectedFilenameLibrary", "()Ljava/lang/String;");
  vstr = (jstring) currentEnv->CallObjectMethod (currentObj, mid);
  const char *str = ((currentEnv)->GetStringUTFChars (vstr, 0));
  return ((char *) str);
}


///
extern int
C2Java_getIndexLibrary (char *fn)
{
  jstring vstr = currentEnv->NewStringUTF (fn);
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getIndexLibrary", "(Ljava/lang/String;)I");
  return (currentEnv->CallIntMethod (currentObj, mid, vstr));
}

/// The objects of the the vindex library are defined by the tree lib
extern void
C2Java_install_library (ntrees lib, int vindex)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "install_library", "(JI)V");
  currentEnv->CallVoidMethod (currentObj, mid, (jlong) lib, vindex);
}


///
extern boolean
C2Java_getNamesState ()
{
  boolean b;
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getNamesState", "()Z");
  b = currentEnv->CallBooleanMethod (currentObj, mid);
  return (b);
}

///
extern boolean
C2Java_getDetailState ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "getDetailState", "()Z");
  return (currentEnv->CallBooleanMethod (currentObj, mid));
}

///
extern void
C2Java_displayConfirm (int action)
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "displayConfirm", "(I)V");
  currentEnv->CallVoidMethod (currentObj, mid, action);
}


///
extern void
C2Java_quitApplication ()
{
  mid = currentEnv->GetMethodID (currentEnv->GetObjectClass (currentObj), "quitApplication", "()V");
  currentEnv->CallVoidMethod (currentObj, mid);
}
