/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
//_PUBLIC_


/// Constructor
SignalDisplaying::SignalDisplaying (ptDrawing pDes, ptGraphicTrees pGarb, ptGraphicTablesManager pTab, ptCurrentObjectsManager pLoc, ptMessage pMess):pDraw (pDes), the_graphicalTrees (pGarb), the_graphicTablesManager (pTab), pCurObjManager (pLoc),
  the_messageTool
  (pMess)
{
  strcpy (signalDecl, " Signal declaration");
  strcpy (signalExpr, " Signal expression");
  strcpy (signalProfile, " Signal profile");
  strcpy (signalPragmas, " Signal pragmas");
  strcpy (signalExternGraph, " Signal external graph");
}


/**
 Fonction d'initialisation appelee par le programme principal sigmain
*/
void
SignalDisplaying::init_decomp ()
{
  filebuf = new buffer (10, 60);
  affich_nom_port = pk_true;
  force = pk_true;
}


/**
  Positionne le booleen detail:
  lorsqu'il est a faux : l'affichage du texte dans les boites
  se fait avec un holophraste de 3; lorsqu'il est a vrai,
  holophraste egal a 9 (si le texte tient dans la boite).
 */
void
SignalDisplaying::misajour_detail (boolean detail)
{
  force = detail;
}


/**
  Set the boolean affich_nom_port to the value affich (in). This method is
  called when the user clicks on the check button "names".
  if affich_nom_port is pk_true, the port names are displayed in the canvas.
*/
void
SignalDisplaying::misajour_nomport (boolean affich)
{
  affich_nom_port = affich;
}


/**
  Save the value of affich_nom_port in affich_sav and set affich_nom_port to
  true. This method is called when it is necessary to temporarily see the 
  port names in the canvas.
*/
void
SignalDisplaying::sauvegarde_nomport (void)
{
  affich_sav = affich_nom_port;
  affich_nom_port = pk_true;
}


/**
  Set affich_nom_port to the value of affich_sav. This method is called after
  a call of sauvegarde_nomport.
*/
void
SignalDisplaying::restitue_nomport (void)
{
  affich_nom_port = affich_sav;
}


/**
   Decompile dans la fenetre texte ?w l'arbre p precede du phylum adequat.
*/
void
SignalDisplaying::decompTextWin (text_win_type w, ntrees p)
{
  int i, ope, nb_elem;
  boolean toutdec = pk_true;
  ntrees p_fils;

  C2Java_resetText (w);
  // printf("DEBUG entering SignalDisplaying::decompTextWin, w is %i\n", (int)w); fflush(stdout);
  filebuf->setsize (2, C2Java_getTextColumns (w));
  if (p == null)
    {
      // C2Java_setTitle(w, (char*) signalDecl);
      C2Java_textShow (w);
    }
  else
    {
      ope = noper (p);
      if (ope == opldecl)
	{			// only the signals, types, constants declarations are displaying
	  // C2Java_setTitle(w, (char*)signalDecl);
	  nb_elem = ntreearity (p);
	  for (i = 1; i <= nb_elem; i++)
	    {
	      p_fils = nchild (i, p);
	      if (!STree_isModel (p_fils))
		displayText (w, p_fils, ";");
	    }
	  C2Java_textShow (w);
	}
      else
	{
	  switch (ope)
	    {
	    case opnil:	//when ope is opnil, it may be an external graph or a declaration (a port)
	      // if (STree_isSpec(p)) C2Java_setTitle(w, (char*) signalExternGraph); else C2Java_setTitle(w, (char*)signalDecl);
	      break;
	    case opldecs:
	    case opdecsig:
	      // C2Java_setTitle(w, (char*)signalDecl);
	      break;
	    case opspec:
	      // C2Java_setTitle(w, (char*)signalExternGraph);
	      break;
	    default:
	      if (STree_isExprProc (ope))
		{
		  if (ope == oplfeed)
		    {
		      p_fils = nchild (1, p);
		      if (p_fils != null)
			if (nattrntrees (p_fils, GRAPHI_ATTRNUMBER) != null)
			  {
			    toutdec = pk_false;
			    C2Java_textHide (w);
			  }
		    }
		  // if (toutdec) C2Java_setTitle(w, (char*)signalExpr); 
		}
	    }
	  if (ntreearity (p) != 0)
	    displayText (w, p, NULL);
	  if (toutdec)
	    C2Java_textShow (w);
	}
    }
  // printf("DEBUG  exiting SignalDisplaying::decompTextWin, w is %i\n", (int)w); fflush(stdout);
}


/**
  Affichage dans une petite fenêtre du nom d'un port (appelée après
  que l'utilisateur ait associé une expression de processus à une boite
 */
void
SignalDisplaying::editPort (ptrepstr name)
{
  C2Java_showNamePort (repstr2string (name));
}


/**
  Edition dans la fenetre texte du (ou des) pragma(s) associes au  processus ?proc.
  S'il n'y a pas de pragmas, on affiche uniquement le nom du phylum
*/
void
SignalDisplaying::edit_pragmas (ntrees proc)
{
  int nbcol;
  ntrees prag;

  if (proc != null)
    {
      prag = STree_getPragmas (proc);
      nbcol = C2Java_getTextColumns (C_HierarText);
      filebuf->setsize (7, nbcol);
      // C2Java_setTitle(C_HierarText, (char*)signalPragmas);
      C2Java_resetText (C_HierarText);
      displayText (C_HierarText, prag, NULL);
      C2Java_textShow (C_HierarText);
    }
}

///
void
SignalDisplaying::display_treeInfoWindow (ntrees vt)
{
}

///
void
SignalDisplaying::display_CurrentLibSelectedModel ()
{
  ntrees proc;
  ntrees prag, vt, save;

  GetCurrentLibSelectedModel (&proc);
  if (ISNOT_NILTREE (proc))
    {
      save = STree_getDescription (proc);
      STree_setDescription (proc, LL_make_desc_tree (NILTREE, NILTREE));
      display_treeInfoWindow (proc);
      STree_setDescription (proc, save);
      prag = STree_getPragmas (proc);
      ntrees_memoryStoreIndex ();	// For local garbaging
      if (STree_mkSelectedPragma (prag, getPredefinedPragma (PRAGMA_COMMENT), &vt))
	display_treeInfoWindow (vt);
      ntrees_memoryRestoreIndex ();	// PLG ??? local garbaging (done)
    }
}

///
boolean
SignalDisplaying::display_commentInfoWindow (ntrees p_sig)
{
  ntrees proc;
  ntrees prag, vt;
  boolean b;

  if (IS_NILTREE (p_sig))
    return pk_false;
  if (!STree_isBodyTree (p_sig))
    return pk_false;
  proc = STree_getModelInWhich (p_sig);
  if (IS_NILTREE (proc))
    return pk_false;
  prag = STree_getPragmas (proc);
  ntrees_memoryStoreIndex ();	// For local garbaging
  if (!STree_mkSelectedPragma (prag, getPredefinedPragma (PRAGMA_COMMENT), &vt))
    b = pk_false;
  else
    {
      b = pk_true;
      display_treeInfoWindow (vt);
    }
  ntrees_memoryRestoreIndex ();	// PLG ?? local garbaging (done)
  return b;
}

 /**
   Edition in a textual window of the Signal declarations attached to 
   the ports of a box.
   win (in) : the textual window
   bte (in) : the box number
 */
void
SignalDisplaying::edit_profil (text_win_type win, int bte)
{
  ntrees lPorts;

  // C2Java_setTitle(win, (char*)signalProfile);
  lPorts = the_graphicalTrees->getPorts (the_graphicTablesManager->graphObj_get_tree (bte));
  filebuf->setsize (7, C2Java_getTextColumns (win));
  C2Java_resetText (win);
  C2Java_textAppend (win, openParenthesis);
  C2Java_textAppend (win, inputBegin);
  decomp_l_decsig (win, gpkOp_porte, lPorts);
  C2Java_textAppend (win, outputBegin);
  decomp_l_decsig (win, gpkOp_ports, lPorts);
  C2Java_textAppend (win, closeParenthesis);
  C2Java_textShow (win);
}


/**
  Calcule les coordonnees du port ?arb_port pour l'affichage :
  on tient compte du facteur d'echelle lie a la reduction ou l'agrandissement
  de la fenetre graphique,
  Si la boite contenant le port est la boite englobante courante, alors le
  port est affiche en tant que port externe.
  Rend les coordonnees dans !x et !y et dans !ou "gauche" ou "droite" (suivant
  que le port se trouve a gauche ou a droite de la boite).
*/
void
SignalDisplaying::calcul_coord_port (ntrees arb_port, int *x, int *y, int *ou, boolean * isExtern)
{
  ntrees p_boite;
  int form, nbcoord;
  typcoord tcoord[2], vcoord[2];
  float val_reel;
  db_reel fact_ech;

  form = the_graphicalTrees->getCoord (arb_port, &nbcoord, tcoord, 0);
  form = the_graphicalTrees->getCoord (arb_port, &nbcoord, vcoord, 1);
  if (vcoord[0].x < 0)
    *ou = gauche;
  else
    *ou = droite;
  p_boite = the_graphicalTrees->boite_mere (the_graphicalTrees->tree_get_graphObj (arb_port));
  if (p_boite == pCurObjManager->get_bte_cou_engl ())	/* les ports sont alors externes */
    {
      *isExtern = pk_true;
      the_graphicalTrees->getCoord (p_boite, &nbcoord, tcoord, 1);
      val_reel = lg_bte_stand;
      fact_ech.x = val_reel / tcoord[1].x;
      val_reel = ht_bte_stand;
      fact_ech.y = val_reel / tcoord[1].y;
      *x = (int) (vcoord[0].x * fact_ech.x) + abs_mil_cadre;
      *y = (int) (vcoord[0].y * fact_ech.y) + ord_mil_cadre;
    }
  else
    {
      *isExtern = pk_false;
      *x = tcoord[0].x;
      *y = tcoord[0].y;
    }
}


/**
  Cette procedure affiche le nom du port ?arb_port, s'il existe, dans la fenetre 
  graphique a cote du port. Pour cela, elle appelle le decompilateur Signal
  textuel et calcule les coordonnees du point ou sera place la premiere lettre 
  du nom du port.
  Si plusieurs noms sont associes au port, il n'y a pas d'affichage des noms
  dans la fenetre graphique.
*/
void
SignalDisplaying::display_nom_port (ntrees arb_port)
{
  int x, y, i, ope, holo, noligne;
  ntrees arb_sig;
  boolean mult_nom, isExtern;

  arb_sig = nattrntrees (arb_port, SIGNAL_ATTRNUMBER);
  if (arb_sig != null)
    if (noper (arb_sig) != opnil)
      {
	mult_nom = the_graphicalTrees->is_port_multi (arb_port);
	if (!mult_nom)
	  {
	    filebuf->setsize (10, 80);
	    holo = 2;
	    noligne = appel_decompilo_text (STree_getNameOfDeclaration (arb_sig), holo);
	    i = 0;
	    while (stock_chaine[0][i] != '\00')
	      i++;
	    calcul_coord_port (arb_port, &x, &y, &ou, &isExtern);
	    ope = noper (arb_port);
	    if (((ope == gpkOp_porte) && !isExtern) || ((ope == gpkOp_ports) && isExtern))
	      y = y - (taillepave_x / 2);
	    else
	      y = y + car_port[1];
	    if (ou == gauche)
	      x = x - (i * car_port[0]) - taillepatte;
	    else
	      x = x + taillepatte;
	    if (x <= 0)
	      x = taillepatte;
	    envoyer_text (null, x, y, 0, noligne, C_port_police, holo);
	  }
      }
}


/**
  Cette fonction affiche le port ?port avec de nouvelles coordonnees
  ?tcoord.
  ?ty_port : egal a typ_port ou typ_port_ext
  ?ou : gauche ou droite
  ?col : couleur du port
*/
void
SignalDisplaying::display_port_depl (ntrees port, objet_type ty_port, typcoord * tcoord, int ou, int col)
{
  if (noper (port) == gpkOp_porte)
    {
      if (ty_port == typ_port)
	pDraw->drawPort (tcoord[0].x, tcoord[0].y, pk_true, pk_false, ou, col, pk_true);
      else
	pDraw->drawExternPort (tcoord[0].x, tcoord[0].y, pk_true, ou, col);
    }
  else
    {
      if (ty_port == typ_port)
	pDraw->drawPort (tcoord[0].x, tcoord[0].y, pk_false, pk_false, ou, col, pk_false);
      else
	pDraw->drawExternPort (tcoord[0].x, tcoord[0].y, pk_false, ou, col);
    }
}


/**
 * Affiche le port represente par son arbre ?arb_port dans la ?fenetre graphique
 * si ?col vaut 0 : on efface
 * sinon on dessine en couleur suivant la valeur de ?col
 * ?fact_ech_calc : le facteur d'echelle avec lequel on dessine le port,
           vaut (1,1) ou (0,0) normalement, 
	   est inferieur a (1,1) et superieur a (0,0) lorsqu'on visualise des
	   boites incluses dans d'autres boites; dans ce cas, les ports ont
	   leur taille divisee par 2.
*/
void
SignalDisplaying::display_port (ntrees arb_port, int col, db_reel fact_ech_calc)
{
  int nbcoord, ope, form;
  boolean pat, isExtern, isInput;
  typcoord vcoord[2];

  calcul_coord_port (arb_port, &x_port, &y_port, &ou, &isExtern);
  ope = noper (arb_port);
  if ((fact_ech_calc.x == 0. && fact_ech_calc.y == 0.) || (fact_ech_calc.x == 1. && fact_ech_calc.y == 1.))
    {
      isInput = (ope == gpkOp_porte);
      if (col == COL_HALO)
	pDraw->drawHaloPort (x_port, y_port, isInput, ou, col);
      else
	pDraw->drawPort (x_port, y_port, isInput, pk_false, ou, col, pk_true);
    }
  else				/* le fact. d'echelle est > 1 : on divise la taille du port par 2 et on calcule */
    {				/* les coord. par rapport a la position de la bte calculee dans display_rec */
      form = the_graphicalTrees->getCoord (arb_port, &nbcoord, vcoord, 1);
      x_port = (int) (vcoord[0].x * fact_ech_calc.x) + mil_cadre.x;
      y_port = (int) (vcoord[0].y * fact_ech_calc.y) + mil_cadre.y;
      if (ope == gpkOp_porte)
	{
	  pat = (nchild (4, arb_port) == null);
	  pDraw->drawPort (x_port, y_port, pk_true, pk_true, ou, 1, pat);
	}
      else
	pDraw->drawPort (x_port, y_port, pk_false, pk_true, ou, 1, pk_false);
    }
}


/**
  Affiche le port ?p_port avec ses liens si le port est connecté.
  (Il s'agit de ses liens externes).
*/
void
SignalDisplaying::display_port_avec_liens (ntrees p_port, int couleur, db_reel f_ech)
{
  ntrees p_ch, p_lien;
  int ope;

  display_port (p_port, couleur, f_ech);
  p_ch = nchild (3, p_port);
  p_lien = nchild (4, p_port);
  ope = noper (p_port);
  if ((p_ch != null) && (ope == gpkOp_ports))
    display_obj (p_ch, couleur);
  if ((p_lien != null) && (ope == gpkOp_porte))
    {
      p_ch = the_graphicTablesManager->graphObj_get_tree ((int) nintof (p_lien));
      display_obj (p_ch, couleur);
    }
}


/**
  Affiche ou efface la boite ?p_bte seule dans la fenetre graphique avec
  la couleur ?couleur (si la couleur vaut 0, alors on efface).
  Fonction appelee par des modules exterieurs, donc pas de calcul de
  facteur d'echelle (toujours egal a 1).
*/
void
SignalDisplaying::display_bte_seule (ntrees p_bte, int couleur)
{
  int nbcoord, larg, haute;
  typcoord milieu, tcoord[2];

  the_graphicalTrees->getCoord (p_bte, &nbcoord, tcoord, 0);
  milieu.x = tcoord[0].x;
  milieu.y = tcoord[0].y;
  larg = tcoord[1].x;
  haute = tcoord[1].y;
  pDraw->dessiner_rect (milieu, larg, haute, couleur);
}


/**
  Affiche ou efface la boite ?p_bte et ses ports dans la fenetre graphique
  avec la couleur ?couleur (si la couleur vaut 0, alors on efface).
  Fonction appelee par des modules exterieurs, donc pas de calcul de
  facteur d'echelle (toujours egal a 1).
*/
void
SignalDisplaying::display_bte_avec_ports (ntrees p_bte, int couleur)
{
  ntrees p_port;
  db_reel fact_ech;

  display_bte_seule (p_bte, couleur);
  p_port = the_graphicalTrees->getFirstPort (p_bte);
  while (p_port != null)
    {
      fact_ech.x = 1.;
      fact_ech.y = 1.;
      display_port (p_port, couleur, fact_ech);
      p_port = ntrees_getNext (p_port);
    }
}


/**
  Affiche ou efface la boite ?p_bte, ses ports et les liens qui arrivent
  sur ses ports dans la fenetre graphique avec la couleur ?couleur.
  Fonction appelee par des modules exterieurs, donc pas de calcul de
  facteur d'echelle (toujours egal a 1).
  NOT USED FOR THE MOMENT !!!
*/
void
SignalDisplaying::display_bte_avec_ports_liens (ntrees p_bte, int couleur)
{
  ntrees p_port;
  db_reel fact_ech;

  display_bte_seule (p_bte, couleur);
  p_port = the_graphicalTrees->getFirstPort (p_bte);
  while (p_port != null)
    {
      fact_ech.x = 1.;
      fact_ech.y = 1.;
      display_port_avec_liens (p_port, couleur, fact_ech);
      p_port = ntrees_getNext (p_port);
    }
}


/**
 Affiche dans la fenetre graphique l'objet passe en parametre 
 ?g_arb. Si ?couleur vaut 0, alors on efface.
 Cette fonction est appelée pour l'affichage de niveau courant, donc
 pas de calcul de facteur d'echelle (il est égal à 1).
*/
void
SignalDisplaying::display_obj (ntrees g_arb, int couleur)
{
  int form, nbcoord;
  typcoord tcoord[maxcoord];
  db_reel fact_ech;
  typcoord translat;
  boolean port_mult;

  if (ISNOT_NILTREE (g_arb))
    {
      fact_ech.x = fact_ech.y = 1;
      translat.x = decal_cadre_x;
      translat.y = decal_cadre_y;
      if (noper (g_arb) == opchemin)
	{
	  port_mult = the_graphicalTrees->is_port_multi (the_graphicalTrees->get_port_diff (g_arb));
	  display_chemin_rec (g_arb, translat, fact_ech, couleur, 0, 0, port_mult);
	}
      else
	{
	  if (noper (g_arb) == opschema)
	    g_arb = nchild (1, g_arb);
	  form = the_graphicalTrees->getCoord (g_arb, &nbcoord, tcoord, 0);
	  switch (form)
	    {
	    case oprect:
	      display_bte_avec_ports (g_arb, couleur);
	      break;
	    case oppave:
	      display_port (g_arb, couleur, fact_ech);
	      break;
	    }
	}
    }
}


/**
  Cette fonction est appelee après le mapping lorsqu'il y a ambiguité
  sur un routage (exemple: une boîte avec un port de sortie connecté,
  qui est mappée 2 fois, alors il y a ambiguité sur le lien à créer.
  ?l_ambig est une liste de listes de ports : 1ier port le port
  récepteur, ensuite les ports diffuseur qui peuvent l'alimenter.
  Exemple : si on a 2 ports de sortie, numéro 1 et 2 qui peuvent être
  connecté au port 3, et si on a 3 ports de sortie 5,6 et 7 qui 
  peuvent être connectés aux ports 10, alors la liste sera composée
  de (3, 1, 2), (10, 5, 6, 7).
  On affiche une croix à côté de chaque port, suivie d'un numéro
  ?no pour la 1iere sous-liste, no+1 pour la seconde ...
*/
void
SignalDisplaying::display_ports_ambig (ntrees l_ambig, int no)
{
  ntrees p, p_port;
  int nb_ports, x, y, i, l, ou;
  boolean isExtern;

  the_messageTool->error_mess (warmapext);
  p = nchild (1, l_ambig);
  while (ISNOT_NILTREE (p))
    {
      nb_ports = ntreearity (p);
      sprintf (tab, "%d", no);
      l = strlen (tab);
      tab[l] = ' ';
      tab[l + 1] = '\00';
      for (i = 1; i <= nb_ports; i++)
	{
	  p_port = the_graphicTablesManager->graphObj_get_tree ((int) nintof (nchild (i, p)));
	  calcul_coord_port (p_port, &x, &y, &ou, &isExtern);
	  y = y - (taillepave_x / 2);
	  x = x + 2 * taillepave_x;
	  pDraw->dessiner_croix (x, y);
	  pDraw->dessiner_text ((char *) tab, x + 15, y, 2);
	}
      p = ntrees_getNext (p);
      no++;
    }
}


/**
  Affiche en bleu les boites contenues dans le buffer de selection. 
*/
void
SignalDisplaying::display_buff_select (int select_buffer[], int bufSize)
{
  int i, nbcoord, x, y, ou, noBox;
  ntrees pObj;
  typcoord milieu, tcoord[2];
  boolean isExtern;

  i = 0;
  while ((select_buffer[i] != 0) && (i < bufSize))
    {
      pObj = the_graphicTablesManager->graphObj_get_tree (select_buffer[i]);
      if (ISNOT_NILTREE (pObj))
	if (pCurObjManager->is_bte_design (select_buffer[i]))
	  {			// the selected object is a box and is visible
	    the_graphicalTrees->getCoord (pObj, &nbcoord, tcoord, 0);
	    milieu.x = tcoord[0].x;
	    milieu.y = tcoord[0].y;
	    pDraw->dessiner_select (milieu, tcoord[1].x, tcoord[1].y, COL_SEL);
	  }
	else
	  {
	    if ((noper (pObj) == gpkOp_porte) || (noper (pObj) == gpkOp_ports))
	      {			// the selected object is a port
		noBox = the_graphicalTrees->no_boite_mere (pObj);
		if ((pCurObjManager->is_bte_design (noBox)) ||
		    (the_graphicTablesManager->graphObj_get_tree (noBox) == pCurObjManager->get_bte_cou_engl ()))
		  {		// the port is visible and can be displayed
		    calcul_coord_port (pObj, &x, &y, &ou, &isExtern);
		    if (ou == gauche)
		      milieu.x = x - 10;
		    else
		      milieu.x = x + 10;
		    milieu.y = y;
		    pDraw->dessiner_select (milieu, 5, 5, RED);
		  }
	      }
	  }
      i++;
    }
}


/**
 * decompilation de l'arbre courant dans la fenetre graphique
 * ?affich_pur est a vrai si c'est uniquement de l'affichage, a faux
 * s'il faut mettre a jour les 2 listes d'objets designables (lorsqu'on
 * descend ou remonte).
 */
void
SignalDisplaying::display (boolean affich_pur)
{
  ntrees p, p_aux, p_schema, p_cour;

  p_cour = pCurObjManager->get_arb_cou_engl ();
  car_box[0] = C2Java_getSizeFontL (J_box_police);
  car_box[1] = C2Java_getSizeFontH (J_box_police);
  car_port[0] = C2Java_getSizeFontL (J_port_police);
  car_port[1] = C2Java_getSizeFontH (J_port_police);
  p_schema = pCurObjManager->get_bte_cou_engl ();
  if (p_schema != null)
    {
      p = the_graphicalTrees->get_arb_sig (&p_schema);
      if (p != p_cour)		/* on est dans un port */
	{			/* on remonte dans l'arbre uniquement pour reafficher */
	  p_aux = p_cour;
	  pCurObjManager->set_arb_cou_engl (p);
	  display_tree (p, affich_pur);
	  pCurObjManager->set_arb_cou_engl (p_aux);
	}
      else			// on est dans une boite
	display_tree (p, affich_pur);
    }
  else if (p_cour != null)
    if (p_cour == pCurObjManager->getModelInWhich_cou_engl ())	/* on est au niveau du process */
      {
	display_tree (pCurObjManager->get_arb_cou_engl (), affich_pur);
      }
    else
      {				// on est dans des boites internes
	p_aux = pCurObjManager->getModelInWhich_cou_engl ();
	pCurObjManager->set_arb_cou_engl (p_aux);
	display_tree (p_aux, affich_pur);
	pCurObjManager->set_arb_cou_engl (p_cour);
      }
}
