/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
//_PUBLIC_


#define COPY_ATTRIBUTEDSTREE(ATREE) ncopytree(ATREE, pk_true)
/**
  save in the graph_arb object the program Signal root 
*/
void
graphicalTrees::setRoot (ntrees p)
{
  root = p;
}


/**
  give the program Signal root 
*/
ntrees
graphicalTrees::getRoot (void)
{
  return (root);
}


/**
  go up in the hierarchy to give the root of the process Proc, with stopping 
  before the "environment" process.
*/
ntrees
graphicalTrees::getRootProc (ntrees proc)
{
  ntrees prec, p;

  prec = proc;
  p = STree_getModelInWhich (proc);
  while (p != root)
    {
      prec = p;
      p = STree_getModelInWhich (p);
    }
  return (prec);
}


/**
  Returns a copy of the tree arb (in) if destruct (in) is pk_true, otherwise return arb.
 */
ntrees
graphicalTrees::sauver_arbre (ntrees arb, boolean destruct)
{
  if (!destruct)
    return (COPY_ATTRIBUTEDSTREE (arb));
  else
    return (arb);
}


/**
  Returns a copy of the tree arb (in) with changing all the graphical object numbers
  which are in the attributes of the tree to keep unicity of all the numbers.
*/
ntrees
graphicalTrees::dupl_arbre (ntrees arb)
{
  ntrees arb_dupl;

  // printf("DEBUG entering graphicalTrees::dupl_arbre\n"); fflush(stdout);
  arb_dupl = COPY_ATTRIBUTEDSTREE (arb);
  reconstruire_tab (arb_dupl, pk_false, pk_false, 1., 1.);
  // printf("DEBUG  exiting graphicalTrees::dupl_arbre \n"); fflush(stdout);
  return (arb_dupl);
}


/**
  Returns the initialization contained in the declaration tree p_decl (in).
  If there is no initialization, returns NILTREE
  CAUTION : if p_decl is a list of declarations, return the initialization
  of the first declaration.
 */
ntrees
graphicalTrees::init_decl (ntrees p_decl)
{
  ntrees p_init;

  p_init = null;
  if (noper (p_decl) == opldecs)	/* si c'est une liste de declaration, on prend */
    p_decl = nchild (1, p_decl);	/* la premiere */
  if (noper (p_decl) == opdecsig)
    p_decl = nchild (1, nchild (2, p_decl));
  if (noper (p_decl) == opinit)
    p_init = nchild (2, p_decl);
  return (p_init);
}


/**
  * Rend vrai si les deux declarations ?p_decl et ?p_decl_lie
  * representent le meme signal (meme identificateur)
  */
boolean
graphicalTrees::meme_nom (ntrees p_decl, ntrees p_decl_lie)
{
  return (STree_isIdentEq (STree_getNameOfDeclaration (p_decl), STree_getNameOfDeclaration (p_decl_lie)));
}


/**
 * compare les 2 signaux ?sig1 et ?sig2 et regarde s'ils ont le 
 * meme nom et la meme initialisation
 */
boolean
graphicalTrees::meme_sig (ntrees sig1, ntrees sig2)
{
  if (noper (sig1) == opinit)
    if (noper (sig2) != opinit)
      return (pk_false);
    else if (!STree_isIdentEq (nchild (1, sig1), nchild (1, sig2)))
      return (pk_false);
    else
      return (STree_isTreeEq (nchild (2, sig1), nchild (2, sig2)));
  else if (noper (sig2) == opinit)
    return (pk_false);
  else
    return (STree_isIdentEq (sig1, sig2));
}


/**
 Cette fonction rend l'arbre Signal auquel est accroche le sous-arbre
 graphique ?arb_gra
 Attention, ?p_attr n'est pas forcemment directement accroche a l'arbre
 Signal, il faut remonter l'arbre graphique qui est rendu dans arb_gra.
*/
ntrees
graphicalTrees::get_arb_sig (ntrees * arb_gra)
{
  while (nfather (*arb_gra) != null)
    {
      *arb_gra = nfather (*arb_gra);
    }
  return (nattrcarrier (*arb_gra));
}


/**
 Cette fonction rend la boite associee (son arbre) a l'arbre Signal ?arb_sig
*/
ntrees
graphicalTrees::get_arb_boite (ntrees arb_sig)
{
  ntrees arb_bte;

  arb_bte = nattrntrees (arb_sig, GRAPHI_ATTRNUMBER);
  if (arb_bte != NULL)
    if (noper (arb_bte) == opschema)
      arb_bte = nchild (1, arb_bte);
  return (arb_bte);
}


/**
  Rend le numero (unique) de l'objet represente par son arbre ?arb_obj
  Si l'arbre est celui d'un schema (boite avec ports), on rend le numero
  de la boite.
*/
graphObj_type
graphicalTrees::tree_get_graphObj (ntrees arb_obj)
{
  ntrees p;

  // nDEBUG_subtree(0, 4, "entering graphicalTrees::tree_get_graphObj arb_obj is", arb_obj);
  if (noper (arb_obj) == opschema)
    arb_obj = nchild (1, arb_obj);
  // nDEBUG_subtree(0, 4, "     in  graphicalTrees::tree_get_graphObj p is", p);

  return ((graphObj_type) nintof (nchild (1, arb_obj)));
}


/**
  get the box associated with the Signal tree pSig (in); if the box doesn't
  exist, go up in the Signal tree until founding a box. If no box is found,
  returns 0
 */
int
graphicalTrees::getMotherBox (ntrees pSig)
{
  ntrees pBox;

  while ((ISNOT_NILTREE (pSig)) && (IS_NILTREE ((pBox = get_arb_boite (pSig)))))
    pSig = nfather (pSig);
  if ((IS_NILTREE (pSig)) || (IS_NILTREE (pBox)))
    return (0);
  else
    return (tree_get_graphObj (pBox));
}


/**
 Cette fonction rend le numéro de la boite associee a l'arbre Signal ?arb_sig
 S'il n'y a pas de boite, rend 0.
*/
int
graphicalTrees::get_no_boite (ntrees arb_sig)
{
  ntrees arb_bte;
  int no;

  // nDEBUG_subtree(0, 4, "entering graphicalTrees::get_no_boite, arb_sig is", arb_sig);
  arb_bte = get_arb_boite (arb_sig);
  // nDEBUG_subtree(0, 4, "     in  graphicalTrees::get_no_boite, arb_bte is", arb_bte);
  if (arb_bte != NULL)
    no = tree_get_graphObj (arb_bte);
  else
    no = 0;
  return (no);
}


/**
  Rend l'arbre representant la boite mere du port de numero ?no_port
*/
ntrees
graphicalTrees::boite_mere (int no_port)
{
  ntrees p;

  p = ntrees_getPrevious (nfather (the_graphicTablesManager->graphObj_get_tree (no_port)));
  return (p);
}


/**
  Returns the number of the box which contains the port (in).
*/
int
graphicalTrees::no_boite_mere (ntrees port)
{
  return (tree_get_graphObj (ntrees_getPrevious (nfather (port))));
}


/**
  If pMod (in) is a model, returns the number of the box associated with
  its highest level expression; if pMod is a modul, returns its box number.
*/
int
graphicalTrees::no_boite_proc (ntrees pMod)
{
  ntrees p, pSig;

  // nDEBUG_subtree(0, 40, "entering graphicalTrees::no_boite_proc pMod is", pMod);
  if (STree_isModel (pMod))
    {
      pSig = STreeModel_getBody (pMod);
      // nDEBUG_subtree(0, 4, "     in  graphicalTrees::no_boite_proc pSig body is", pSig);
      p = nattrntrees (pSig, GRAPHI_ATTRNUMBER);
      // nDEBUG_subtree(0, 4, "     in  graphicalTrees::no_boite_proc p attribut is", p);
      if (noper (p) == opschema)
	p = nchild (1, p);
    }
  else if (STree_isModule (pMod))
    p = nattrntrees (pMod, GRAPHI_ATTRNUMBER);
  // nDEBUG_subtree(0, 4, "     in  graphicalTrees::no_boite_proc p box is", p);
  return (tree_get_graphObj (p));
}


/** 
   Returns true if the box number no (in) is the box of the hightest level
   expression of a model, pk_false if not.
*/
boolean
graphicalTrees::isBoxProc (int no)
{
  ntrees pBox, pSig;

  pBox = the_graphicTablesManager->graphObj_get_tree (no);
  pSig = get_arb_sig (&pBox);
  return (STree_isBodyTree (pSig));
}

/**
  returns the Signal model (model or modul) associated with the box no (in).
 */
ntrees
graphicalTrees::getModel (int no)
{
  ntrees pBox, pSig;

  pBox = the_graphicTablesManager->graphObj_get_tree (no);
  pSig = get_arb_sig (&pBox);
  if (STree_isModule (pSig))
    return (pSig);
  else
    return (STree_getModelInWhich (pSig));
}


/**
  Returns true if the graphic object obj (in) is a port, false otherwise
*/
boolean
graphicalTrees::isAPort (int obj)
{
  ntrees p;
  int ope;

  p = the_graphicTablesManager->graphObj_get_tree (obj);
  ope = noper (p);
  return ((ope == gpkOp_porte || ope == gpkOp_ports));
}


/**
  Returns the tree represented the list of the ports of the box pBox (in).
  If the box has no ports, returns NILTREE
 */
ntrees
graphicalTrees::getPorts (ntrees pBox)
{
  ntrees p;

  if (noper (pBox) == gpkOp_bte)
    p = nfather (pBox);
  else
    p = pBox;
  if ((p != null) && (noper (p) == opschema)) return (nchild (2, p));
  return NILTREE;
}


/**
 Cette fonction rend le nombre de ports de la boite ?p_bte
*/
int
graphicalTrees::get_nb_ports (ntrees p_bte)
{
  ntrees p;
  int nb;

  nb = 0;
  if (noper (p_bte) == gpkOp_bte)
    p = nfather (p_bte);
  else
    p = p_bte;
  if (p != null)
    if (noper (p) == opschema)
      nb = ntreearity (nchild (2, p));
  return nb;
}


/**
  Rend le premier port de la boite ?p_bte s'il existe sinon rend null
  ?p_bte peut etre soit la boite directement, soit l'arbre au-dessus,
  c'est-a-dire le schema.
*/
ntrees
graphicalTrees::getFirstPort (ntrees p_bte)
{
  ntrees prem_port, l_ports;

  SET_NILTREE (prem_port);
  if (noper (p_bte) == opschema) 
    l_ports = nchild (2, p_bte);
  else if (IS_NILTREE (nfather (p_bte)))
    SET_NILTREE (l_ports);
  else
    l_ports = ntrees_getNext (p_bte);
  if (ISNOT_NILTREE (l_ports))
    if (noper (l_ports) != opnil)
      prem_port = nchild (1, l_ports);
  return prem_port;
}


/**
  Cette fonction rend vrai si la boite ?p_bte possede au moins un port,
  rend faux sinon.
*/
boolean
graphicalTrees::bte_avec_port (ntrees p_bte)
{
  return (get_nb_ports (p_bte) != 0);
}


/**
  Les 2 fonctions qui suivent rendent respectivement le nombre de ports
  d'entree (de sortie) de la boite de numero ?boite.
*/
int
graphicalTrees::nombre_ports_ent (int boite)
{
  int n, i, nb_ports;
  ntrees l_ports;

  n = 0;
  l_ports = ntrees_getNext (the_graphicTablesManager->graphObj_get_tree (boite));
  nb_ports = ntreearity (l_ports);
  if (l_ports != null)
    for (i = 1; i <= nb_ports; i++)
      if (noper (nchild (i, l_ports)) == gpkOp_porte)
	n++;
  return n;
}


///
int
graphicalTrees::nombre_ports_sor (int boite)
{
  int n, i, nb_ports;
  ntrees l_ports;

  n = 0;
  l_ports = ntrees_getNext (the_graphicTablesManager->graphObj_get_tree (boite));
  nb_ports = ntreearity (l_ports);
  if (l_ports != null)
    for (i = 1; i <= nb_ports; i++)
      if (noper (nchild (i, l_ports)) == gpkOp_ports)
	n++;
  return n;
}


/**
  Rend le nombre de ports placé sur le bord gauche de la boite
*/
int
graphicalTrees::getNbLeftPorts (int boite)
{
  int n, i, nb_ports, nbcoord, form;
  ntrees l_ports;
  typcoord tcoord[maxcoord];

  n = 0;
  l_ports = ntrees_getNext (the_graphicTablesManager->graphObj_get_tree (boite));
  nb_ports = ntreearity (l_ports);
  if (l_ports != null)
    for (i = 1; i <= nb_ports; i++)
      {
	form = getCoord (nchild (i, l_ports), &nbcoord, tcoord, 1);
	if (tcoord[0].x < 0)
	  n++;
      }
  return n;
}


/**
  Rend le nombre de ports placé sur le bord droit de la boite
*/
int
graphicalTrees::getNbRightPorts (int boite)
{
  int n, i, nb_ports, nbcoord, form;
  ntrees l_ports;
  typcoord tcoord[maxcoord];

  n = 0;
  l_ports = ntrees_getNext (the_graphicTablesManager->graphObj_get_tree (boite));
  nb_ports = ntreearity (l_ports);
  if (l_ports != null)
    for (i = 1; i <= nb_ports; i++)
      {
	form = getCoord (nchild (i, l_ports), &nbcoord, tcoord, 1);
	if (tcoord[0].x > 0)
	  n++;
      }
  return n;
}


/**
  Returns true if pPort (in) is a port of the box pBox.
 */
boolean
graphicalTrees::isPortInBox (ntrees pPort, ntrees pBox)
{
  ntrees portc;
  boolean result;

  result = pk_false;
  if (pBox != null)
    {
      portc = getFirstPort (pBox);
      while (!result && ISNOT_NILTREE (portc))
	{
	  result = (tree_get_graphObj (pPort) == tree_get_graphObj (portc));
	  portc = ntrees_getNext (portc);
	}
    }
  return result;
}


/**
  Returns true if the port port_ext (in) is a port of the mother box of the box bte (in)
  (in this case, for the box bte, port is an external port). 
  */
boolean
graphicalTrees::isExternPort (ntrees bte, ntrees port)
{
  ntrees signalTree, fatherTree, motherBox, son, arb_gra;
  boolean result;

  result = pk_false;
  arb_gra = bte;
  signalTree = get_arb_sig (&arb_gra);
  if (ISNOT_NILTREE ((fatherTree = nfather (signalTree))))
    if (ISNOT_NILTREE ((motherBox = nattrntrees (fatherTree, GRAPHI_ATTRNUMBER))))
      {
	son = getFirstPort (motherBox);
	while (!result && ISNOT_NILTREE (son))
	  {
	    result = (son == port);
	    son = ntrees_getNext (son);
	  }
      }
  return result;
}


/**
  Rend vrai si le port ?port est connecté (en interne si ?intern à vrai, 
  normal sinon).
*/
boolean
graphicalTrees::is_connected_port (ntrees port, boolean intern)
{
  boolean result, port_recept;

  port_recept = (noper (port) == gpkOp_porte);
  if ((port_recept && !intern) || (!port_recept && intern))
    result = (nchild (4, port) != null);
  else
    result = (nchild (3, port) != null);
  return result;
}



/**
  Return in tcoord and nbcoord (out) the coordinates and their numbers of
  the graphical object p (in).
  p may be a box, a port, a segment or directly the coordinates sub-tree.
  The returned coordinates are relative to the window if rela (in) equals 0,
  are relative to the including box if rela equals 1.
  Note : in the data structure, the coordinates are relative to the 
  including box, except for the ports (relative to the middle point of
  the their box).
  */
int
graphicalTrees::getCoord (ntrees p, int *nbcoord, typcoord * tcoord, int rela)
{
  ntrees pfils;
  int ope, x, y;

  ope = noper (p);
  switch (ope)
    {
    case gpkOp_bte:
    case opschema:
    case gpkOp_porte:
    case gpkOp_ports:
    case gpkOp_segm:
      p = nchild (2, p);
      break;
    }
  if (rela == 0)
    {
      x = decal_cadre_x;
      y = decal_cadre_y;
    }
  else
    {
      x = 0;
      y = 0;
    }
  ope = noper (p);
  switch (ope)
    {
    case oprect:
      *nbcoord = 2;
      pfils = nchild (1, p);
      tcoord[0].x = (int) nintof (nchild (1, pfils)) + x;
      tcoord[0].y = (int) nintof (nchild (2, pfils)) + y;
      pfils = nchild (2, p);
      tcoord[1].x = (int) nintof (nchild (1, pfils));
      tcoord[1].y = (int) nintof (nchild (2, pfils));
      break;
    case opseg:
      *nbcoord = 2;
      pfils = nchild (1, p);
      tcoord[0].x = (int) nintof (nchild (1, pfils)) + x;
      tcoord[0].y = (int) nintof (nchild (2, pfils)) + y;
      pfils = nchild (2, p);
      tcoord[1].x = (int) nintof (nchild (1, pfils)) + x;
      tcoord[1].y = (int) nintof (nchild (2, pfils)) + y;
      break;
    case oppolyline:
      p = nchild (1, p);
      *nbcoord = 0;
      while (p != null)
	{
	  tcoord[*nbcoord].x = (int) nintof (nchild (1, p)) + x;
	  tcoord[*nbcoord].y = (int) nintof (nchild (2, p)) + y;
	  p = ntrees_getNext (p);
	  *nbcoord = *nbcoord + 1;
	};
      break;
    case oppave:
      *nbcoord = 1;
      tcoord[0].x = (int) nintof (nchild (1, p)) + x;
      tcoord[0].y = (int) nintof (nchild (2, p)) + y;
      p = nchild (1, nchild (2, ntrees_getPrevious (nfather (nfather (p)))));
      if (rela == 0)
	{
	  tcoord[0].x = tcoord[0].x + (int) nintof (nchild (1, p));
	  tcoord[0].y = tcoord[0].y + (int) nintof (nchild (2, p));
	}
    };
  return (ope);
}


/**
  set the coordinate point tree pCoord (in) to the value (x, y) (in)
 */
void
graphicalTrees::setCoord (ntrees pCoord, int x, int y)
{
  nmodifvalint (nchild (1, pCoord), x);
  nmodifvalint (nchild (2, pCoord), y);
}



/**
  Cette fonction rend le chemin auquel appartient le segment ?p_seg.
*/
ntrees
graphicalTrees::get_chemin (ntrees p_seg)
{
  return (nfather (nfather (p_seg)));
}


/**
  Cette fonction rend vrai si le chemin ou le segment ?p_chem_seg
  appartient a un lien simple, c'est-a-dire que le port diffuseur 
  du lien n'alimente qu'un seul port recepteur.
*/
boolean
graphicalTrees::chemin_simple (ntrees p_chem_seg)
{
  int ope;
  ntrees p_chem;

  if (noper (p_chem_seg) == opchemin)
    p_chem = p_chem_seg;
  else
    p_chem = get_chemin (p_chem_seg);
  ope = noper (nfather (p_chem));
  return (((ope == gpkOp_ports) || (ope == gpkOp_porte)) && (nchild (2, p_chem) != null));
}


/**
  Cette fonction rend le numero du port diffuseur du lien auquel  
  appartient le segment ou le chemin ?seg_ou_chem.
*/
int
graphicalTrees::get_no_port_diff (ntrees seg_ou_chem)
{
  ntrees chemin;
  int no_port;

  chemin = get_racine_chemin (seg_ou_chem);
  no_port = (int) nintof (ntrees_getIPrevious (2, chemin));
  return (no_port);
}


/**
  Cette fonction rend l'arbre representant le port diffuseur du lien 
  auquel appartient le segment ou le chemin ?seg_ou_chem.
*/
ntrees
graphicalTrees::get_port_diff (ntrees seg_ou_chem)
{
  ntrees chemin;

  chemin = get_racine_chemin (seg_ou_chem);
  return (nfather (chemin));
}


/**
  * creation des s/arbres graphiques lies aux coordonnees :
  *      1 ier parametre = nom de l'operateur (ex opligne, oprect ...)
  *     2 ieme parametre = nombre de coordonnees (ex pour un rectangle oprect,
  *     nbcoord sera egal a 2)
  *      3 ieme parametre = tableau de coordonnees ( de dimension nbcoord)
  */
ntrees
graphicalTrees::createCoordTree (int opform, int nbcoord, typcoord * coord)
{
  ntrees f1, f2, p;
  int i;

  switch (opform)
    {
    case oprect:
      f1 = LL_make_coord (LL_make_abs_atom (coord[0].x), LL_make_ord_atom (coord[0].y));
      f2 = LL_make_coord (LL_make_abs_atom (coord[1].x), LL_make_ord_atom (coord[1].y));
      p = LL_make_rect (f1, f2);
      break;
    case oppave:
      p = LL_make_pave (LL_make_abs_atom (coord[0].x), LL_make_ord_atom (coord[0].y));
      break;
    case oppolyline:
      p = LL_make_polyline_emptyList ();
      for (i = 0; i < nbcoord; i++)
	{
	  f1 = LL_make_coord (LL_make_abs_atom (coord[i].x), LL_make_ord_atom (coord[i].y));
	  p = ntreeList_appendLast (p, f1);
	};
      break;
    case opseg:
      f1 = LL_make_coord (LL_make_abs_atom (coord[0].x), LL_make_ord_atom (coord[0].y));
      f2 = LL_make_coord (LL_make_abs_atom (coord[1].x), LL_make_ord_atom (coord[1].y));
      p = LL_make_seg (f1, f2);
      break;
    };
  return (p);
}


/**
  * creation de la structure de donnees associe a un port.
  * ?bte : boite a laquelle appartient le port
  * ?coord : ses coordonnees en absolu
  * ?en_entree a vrai (a faux) : port d'entree (de sortie)
  * ?rg : on va inserer le port a droite du ?rg ieme port (il sera donc
  * en ?rg+1 ieme position). Si ?rg vaut 0, on l'insère en tête.
  * (les ports sont tries du haut vers le bas et de gauche a droite)
  */
graphObj_type
graphicalTrees::creatept_connex (graphObj_type bte, typcoord coord, boolean en_entree, int rg)
{
  ntrees s2, arb_bte, arb_schema, arb_sig, arb_pt, lPorts;
  graphObj_type pt_design;
  int form, nbcoord, nbPorts;
  typcoord tcoord[1], vcoord[maxcoord], halo[2];

  pt_design = the_graphicTablesManager->newGraphObjIdentity ();
  arb_bte = the_graphicTablesManager->graphObj_get_tree (bte);
  tcoord[0] = coord;
  form = getCoord (arb_bte, &nbcoord, vcoord, 0);
  if (form == oprect)
    {
      tcoord[0].x = tcoord[0].x - vcoord[0].x;
      tcoord[0].y = tcoord[0].y - vcoord[0].y;
    }
  s2 = createCoordTree (oppave, 1, tcoord);
  if (en_entree)
    arb_pt = LL_make_porte (LL_make_design_atom (pt_design), s2, null, null);
  else
    arb_pt = LL_make_ports (LL_make_design_atom (pt_design), s2, null, null);
  arb_schema = nfather (arb_bte);
  if (arb_schema != null)
    {
      lPorts = nchild (2, arb_schema);
      nbPorts = ntreearity (lPorts);
      if (lPorts != null && nbPorts != 0)
	if (rg == 0)
	  ntreeList_insertFirst (arb_pt, lPorts);
	else if (rg > nbPorts)
	  ntreeList_appendLast (lPorts, arb_pt);
	else
	  ninsertright (rg, arb_pt, lPorts);
      else
	nchangchild (arb_schema, LL_make_lports_list (arb_pt), 2);
    }
  else
    {
      arb_sig = nattrcarrier (arb_bte);
      arb_schema = LL_make_schema (arb_bte, LL_make_lports_list (arb_pt), null);
      nmakeattrntrees (arb_sig, GRAPHI_ATTRNUMBER, arb_schema);
    };
  calcul_halo (arb_pt, halo);
  the_graphicTablesManager->inser_obj (pt_design, arb_pt, halo);
  return (pt_design);
}


/** 
   Creation of a new segment with its coordinates tcoord (in), its diffusor port pt_neg (in).
   Returns the new tree and the new number in noSeg (out).
*/
ntrees
graphicalTrees::createSeg (typcoord * tcoord, int pt_neg, graphObj_type * noSeg)
{
  ntrees pSeg, pCoord;
  typcoord halo[2];
  int direc_seg;

  *noSeg = the_graphicTablesManager->newGraphObjIdentity ();
  pCoord = createCoordTree (opseg, 2, tcoord);
  pSeg = LL_make_segm (LL_make_design_atom (*noSeg), pCoord);
  calcul_halo (pSeg, halo);
  if (tcoord[0].x == tcoord[1].x)
    direc_seg = vertic;
  else
    direc_seg = horiz;
  the_graphicTablesManager->inser_seg (*noSeg, direc_seg, pt_neg, pSeg, halo);
  return (pSeg);
}


/**
  Cette procedure cree un arbre boite associe qui sera associe a ?arb_sig
  de ?nbcoord coordonnees ?tcoord et insere l'objet dans la table.
  rend le numero de la boite creee.
*/
graphObj_type
graphicalTrees::createbte (ntrees arb_sig, typcoord * tcoord, int nbcoord)
{
  ntrees p_bte;
  graphObj_type bte_design;
  typcoord halo[2];

  bte_design = the_graphicTablesManager->newGraphObjIdentity ();
  p_bte = LL_make_bte (LL_make_design_atom (bte_design), createCoordTree (oprect, nbcoord, tcoord), null);
  nmakeattrntrees (arb_sig, GRAPHI_ATTRNUMBER, p_bte);
  calcul_halo (p_bte, halo);
  the_graphicTablesManager->inser_obj (bte_design, p_bte, halo);
  return (bte_design);
}

/**
  Create the box associated with the high level expression of the 
  model arb_sig (in).
  Return the number of the box.
*/
int
graphicalTrees::createbte_proc (ntrees arb_sig)
{
  ntrees p_loc, p_exp;
  int noBox;
  typcoord tcoord_exp[2];

  setDefaultCoord (tcoord_exp);
  p_exp = STreeModel_getBody (arb_sig);
  noBox = createbte (p_exp, tcoord_exp, 2);	/* boite expression de proc */
  p_loc = STreeModel_getLocalDeclarations (arb_sig);
  if (IS_NILTREE (p_loc))
    STree_mkLocalDeclaration (arb_sig);
  return (noBox);
}


/**
  Creation of an empty module whose name is ?c_name ans insertion in
  the program hierarchy, under ?father.
  Be carefull, ?father is a process.
  This method is called when the user creates a module in the JTree window. 
  A box is associated with the modul, only used to reference the modul
  in the Java world.
*/
ntrees
graphicalTrees::SignalModuleCreate (ntrees father, char *c_name)
{
  ntrees name, newModule;
  int noBox;
  typcoord tcoord[2];

  name = LL_make_nom_atom (string2repstr (c_name));
  newModule = STree_mkSkeletonModule (name);
  ntreeList_appendLast (STreeModel_getLocalDeclarations (father), newModule);
  setDefaultCoord (tcoord);
  noBox = createbte (newModule, tcoord, 2);
  return (newModule);
}


/**
  Creation of a process whose name is ?c_name ans insertion in
  the program hierarchy, under ?father.
  ?father may be a process or a module.
  Method called when the use creates a process in the JTree window. 
  Return the new process.
*/
ntrees
graphicalTrees::SignalProcCreate (ntrees father, char *c_name, int category, int index)
{
  ptrepstr str;
  ntrees name, newProc;
  int noProcBox;

  str = string2repstr (c_name);
  name = LL_make_nom_atom (str);
  if (category == C_creerproc)
    newProc = STree_mkEmptyProcess (name);
  else if (category == C_creernode)
    newProc = STree_mkEmptyNode (name);
  else if (category == C_creerfunc)
    newProc = STree_mkEmptyFunction (name);
  else if (category == C_creeraction)
    newProc = STree_mkEmptyAction (name);
  else if (category == C_creerautomaton)
    newProc = STree_mkEmptyAutomaton (name);

  noProcBox = createbte_proc (newProc);
  ninsertright (index, newProc, STreeModel_getLocalDeclarations (father));
  return (newProc);
}


///
int
graphicalTrees::createData (typcoord * tcoord, int nbcoord)
{
  ntrees pData;
  int no;
  typcoord halo[2];

  no = the_graphicTablesManager->newGraphObjIdentity ();
  pData = LL_make_data (LL_make_design_atom (no), createCoordTree (oprect, nbcoord, tcoord));
  calcul_halo (pData, halo);
  the_graphicTablesManager->inser_obj (no, pData, halo);
  return (no);
}
