
/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
///
_PRIVATE_
#define COPY_ATTRIBUTEDSTREE(ATREE) ncopytree(ATREE, pk_true)
/*
  Cette procedure detruit l'arbre SIGNAL associe a une boite detruite
  par l'utilisateur.
*/
  void
builder::detach_arb_sig (ntrees p_sig)
{
  // nDEBUG_subtree(3, 2, "entering  builder::detach_arb_sig, p_sig is ", p_sig);
  if (STree_isModel (p_sig) || STree_isModule (p_sig))
    ntreesDestroy (p_sig);
}

/*
  Cette procedure detruit l'arbre SIGNAL associe a une boite detruite
  par l'utilisateur.
*/
void
builder::detruire_arb_sig (ntrees p_sig)
{
  int i;
  ntrees pere, p;

  // nDEBUG_subtree(3, 2, "entering graph_text::detruire_arb_sig psig is ", p_sig);
  if (!STree_isModel (p_sig) && !STree_isModule (p_sig))
    {
      pere = the_graphicalTrees->remonter_expr_sig (oplfeed, p_sig, &i);
      // nDEBUG_subtree(3, 6, "     in  graph_text::detruire_arb_sig pere is ", pere);
      p = ndestroyListElem (pere, i);
      // nDEBUG_subtree(3, 6, " exiting graph_text::detruire_arb_sig pere is ", pere);
    }
}


/**
  Detruit la structure de donnï¿œes associï¿œe ï¿œ un port et les liens qui
  arrivent ou partent de ce port.
  ?typ_port vaut "typ_port" ou "typ_port_ext"
*/
void
builder::detruire_port (int no_port, int typ_obj)
{
  ntrees p_port;

  p_port = the_graphicTablesManager->graphObj_get_tree (no_port);
  detruire_liens (p_port, typ_obj);
  the_graphicTablesManager->enlever_obj (no_port);
  ntreesDestroy (p_port);
}


/**
  Cette procedure modifie les coordonnees d'une boite 
  elle est agrandie ou diminue proportionnellement a la taille 
  de la boite externe de numero (?bte) utilisee pour les
  encapsulage (action "glue") et decapsulage (action "break").
  ?encapsulee a vrai la boite est agrandie (encapsulage)
  ?arb_gra_int = arbre representant la boite a modifiee.
*/
void
builder::recalcul_bte (int bte, ntrees arb_gra_int, boolean encapsulee)
{
  typcoord bextcoord[2], bintcoord[2], obintcoord[2], halo[2], centre_bte, translat;
  int nbcoord, largeur, hauteur;
  ntrees arb_gra_ext;
  float fact_ech_x, fact_ech_y, val_reel;

  the_graphicalTrees->getCoord (arb_gra_int, &nbcoord, obintcoord, 0);
  arb_gra_ext = the_graphicTablesManager->graphObj_get_tree (bte);
  the_graphicalTrees->getCoord (arb_gra_ext, &nbcoord, bextcoord, 0);
  centre_bte.x = bextcoord[0].x;
  centre_bte.y = bextcoord[0].y;
  largeur = bextcoord[1].x;
  hauteur = bextcoord[1].y;
  if (encapsulee)
    {
      calcul_fact_ech (arb_gra_ext, &fact_ech_x, &fact_ech_y);
      translat.x = centre_bte.x - largeur;
      translat.y = centre_bte.y - hauteur;
      bintcoord[0].x = (int) ((obintcoord[0].x - translat.x) * fact_ech_x) + decal_cadre_x;
      bintcoord[0].y = (int) ((obintcoord[0].y - translat.y) * fact_ech_y) + decal_cadre_y;
    }
  else
    {
      val_reel = largeur;
      fact_ech_x = val_reel / lg_bte_stand;
      val_reel = hauteur;
      fact_ech_y = val_reel / ht_bte_stand;
      translat.x = centre_bte.x - largeur;
      translat.y = centre_bte.y - hauteur;
      bintcoord[0].x = (int) ((obintcoord[0].x - decal_cadre_x) * fact_ech_x) + translat.x;
      bintcoord[0].y = (int) ((obintcoord[0].y - decal_cadre_y) * fact_ech_y) + translat.y;
    }
  bintcoord[1].x = (int) (obintcoord[1].x * fact_ech_x);
  bintcoord[1].y = (int) (obintcoord[1].y * fact_ech_y);
  the_graphicalTrees->calcul_nouv_halo (typ_bte, bintcoord, halo);
  modif_coord (the_graphicalTrees->tree_get_graphObj (arb_gra_int), typ_bte, halo, 2, bintcoord, 0);
  the_graphicalTrees->modif_port_bte (arb_gra_int, obintcoord, bintcoord, pk_true);
}



/**
  Fonction appelee pour l'encapsulation : 
  rend vrai s'il est possible de creer un port pour la boite de numero
  ?bte correspondant au segment ?arb_seg qui coupe la boite ?bte, 
  de port diffuseur ?arb_port_diff.  
  ?ent: vrai s'il s'agit d'un port d'entree.
  Les ports sont crees du cote ou coupe le segment avec si possible
  l'ordonnee du segment, sinon la plus proche possible.
  Au nouveau port cree, on lui accroche 2 attributs:
      - un attribut de type "ALGO_ATTRNUMBER" constitue d'un arbre a 2 fils, le 
      premier fils etant le numero du port diffuseur (soit no_port_map,
      soit no_port_lie) et le second fils la liste de ses ports lies. 
      - un attribut de type  signal contenant la declaration (si elle existe)
      du port diffuseur.
  rend faux si le port ne peut etre cree.
*/
boolean
builder::createport_seg_coupe (int bte, ntrees arb_port_diff, ntrees arb_seg, boolean ent)
{
  int nbcoord, cote_droit, cote_gauche, hauteur, largeur, no_port_bte, no_port_diff, b_inf, b_sup, sens, rg, cote_haut, cote_bas;
  typcoord bcoord[2], scoord[2], nvscoord, centre_bte;
  ntrees arb_no_port_diff, l_no_ports_lies, info_ports_encaps, p_decl, ch;
  boolean b;

  // printf("DEBUG entering builder::createport_seg_coupe\n"); fflush(stdout);
  the_graphicalTrees->getCoord (the_graphicTablesManager->graphObj_get_tree (bte), &nbcoord, bcoord, 0);
  centre_bte.x = bcoord[0].x;
  centre_bte.y = bcoord[0].y;
  largeur = bcoord[1].x;
  hauteur = bcoord[1].y;
  cote_gauche = centre_bte.x - largeur;
  cote_droit = centre_bte.x + largeur;
  cote_haut = centre_bte.y - hauteur;
  cote_bas = centre_bte.y + hauteur;
  b_inf = centre_bte.y - hauteur;
  b_sup = centre_bte.y + hauteur;
  the_graphicalTrees->getCoord (nchild (2, arb_seg), &nbcoord, scoord, 0);
  if (scoord[0].y == scoord[1].y)
    {
      if ((scoord[0].x <= cote_gauche && scoord[1].x >= cote_gauche) || (scoord[1].x <= cote_gauche && scoord[0].x >= cote_gauche))
	nvscoord.x = cote_gauche;
      else
	nvscoord.x = cote_droit;
      nvscoord.y = scoord[0].y;
    }
  else
    {
      if (scoord[0].x <= centre_bte.x)
	nvscoord.x = cote_gauche;
      else
	nvscoord.x = cote_droit;
      if ((scoord[0].y <= cote_bas && scoord[1].y >= cote_bas) || (scoord[1].y <= cote_bas && scoord[0].y >= cote_bas))
	nvscoord.y = b_sup - taillepave_y;
      else
	nvscoord.y = b_inf + taillepave_y;
    }
  if (nvscoord.x == cote_gauche)
    sens = gauche;
  else
    sens = droite;
  if (the_graphicalTrees->place_trou_port (bte, sens, b_inf, b_sup, &nvscoord.y, &rg))
    {
      no_port_bte = the_graphicalTrees->creatept_connex (bte, nvscoord, ent, rg);
      no_port_diff = the_graphicalTrees->tree_get_graphObj (arb_port_diff);
      arb_no_port_diff = LL_make_port_atom (no_port_diff);
      b = the_graphicalTrees->isExternPort (the_graphicTablesManager->graphObj_get_tree (bte), arb_port_diff);
      l_no_ports_lies = the_graphicalTrees->get_ports_lies (no_port_diff, b, &ch, &b);
      info_ports_encaps = LL_make_encaps (arb_no_port_diff, l_no_ports_lies);
      nmakeattrntrees (the_graphicTablesManager->graphObj_get_tree (no_port_bte), ALGO_ATTRNUMBER, info_ports_encaps);
      p_decl = nattrntrees (arb_port_diff, SIGNAL_ATTRNUMBER);
      if (p_decl != null)
	nmakeattrntrees (the_graphicTablesManager->graphObj_get_tree (no_port_bte), SIGNAL_ATTRNUMBER, COPY_ATTRIBUTEDSTREE (p_decl));
      // printf("DEBUG  exiting builder::createport_seg_coupe 1\n"); fflush(stdout);
      return (pk_true);
    }
  // printf("DEBUG  exiting builder::createport_seg_coupe 2\n"); fflush(stdout);
  return (pk_false);
}



/**
  Cette fonction cree autant de ports affectes a la boite de numero
  ?bte que de chemins dont au moins un segment coupe la boite.
  D'autres parts, elle insere dans la liste ?l_n_p_d_e les numeros des
  ports diffuseurs encapsules qui ne sont connectes qu'avec des boites
  encapsulees.
  ?l_n_p_d_e (a l'appel de la fonction, la liste est vide).
  Cette fonction rend faux si une creation de port a echouee
 ?l_bte_enc :  liste des numeros des boites encapsulees
*/
boolean
builder::createports_encaps (int bte, ntrees l_bte_enc, ntrees l_n_p_d_e)
{
  ntrees arb_no_seg, arb_seg, arb_port, arb_port_diff, l_n_seg_c, arb_no_bte, arb_bte;
  int no_bte_diff, n_p_d_e;
  boolean en_entree, fait;

  l_n_seg_c = LL_make_lsegs_emptyList ();
  fait = pk_true;
  cherch_segs_coupe_bte (bte, l_n_seg_c);
  arb_no_seg = nchild (1, l_n_seg_c);
  while (arb_no_seg != null && fait)
    {
      arb_seg = the_graphicTablesManager->acces_segment ((int) nintof (arb_no_seg));
      arb_port_diff = the_graphicalTrees->get_port_diff (arb_seg);
      if (the_graphicalTrees->existe_port_lie_non_encaps (bte, arb_port_diff, l_bte_enc))
	{
	  no_bte_diff = the_graphicalTrees->no_boite_mere (arb_port_diff);
	  en_entree = !the_graphicalTrees->is_in_list (no_bte_diff, l_bte_enc);
	  fait = createport_seg_coupe (bte, arb_port_diff, arb_seg, en_entree);
	}
      arb_no_seg = ntrees_getNext (arb_no_seg);
    }
  if (fait)
    {
      arb_no_bte = nchild (1, l_bte_enc);
      while (arb_no_bte != null)
	{
	  arb_bte = the_graphicTablesManager->graphObj_get_tree ((int) nintof (arb_no_bte));
	  if (nfather (arb_bte) != null)
	    {
	      arb_port = nchild (1, ntrees_getNext (arb_bte));
	      while (arb_port != null)
		{
		  if (noper (arb_port) == gpkOp_ports && nchild (3, arb_port) != null &&
		      !the_graphicalTrees->existe_port_lie_non_encaps (bte, arb_port, l_bte_enc))
		    {
		      n_p_d_e = the_graphicalTrees->tree_get_graphObj (arb_port);
		      ntreeList_insertFirst (LL_make_port_atom (n_p_d_e), l_n_p_d_e);
		    }
		  arb_port = ntrees_getNext (arb_port);
		}
	    }
	  arb_no_bte = ntrees_getNext (arb_no_bte);
	}
    }
  return (fait);
}


/**
  Cette fonction teste si on peut decapsuler la boite bte (in)
  Rend faux si la taille des boites est insuffisante ou si l'ecartement des ports 
  est insuffisant, et positionne dans noErrMess (out) le numï¿œro de message 
  ï¿œ afficher, sinon rend true.
*/
boolean
builder::test_decaps_bte (int bte, int *noErrMess)
{
  ntrees arb_gra, arb_sig, arb_sig_int;
  int no_bte, nbcoord, mask, x, y, d_h;
  boolean ok;
  typcoord bcoord[2], bougetail;

  ok = pk_true;
  bougetail.x = 1;
  bougetail.y = 1;
  arb_gra = the_graphicTablesManager->graphObj_get_tree (bte);
  arb_sig = the_graphicalTrees->get_arb_sig (&arb_gra);
  arb_sig_int = the_graphicalTrees->desc_expr_sig (nchild (1, arb_sig));
  while (arb_sig_int != null && ok)
    {
      arb_gra = the_graphicalTrees->get_arb_boite (arb_sig_int);
      no_bte = the_graphicalTrees->tree_get_graphObj (arb_gra);
      the_graphicalTrees->annuler_halo_seg_bte (no_bte, pk_false, pk_false, &bougetail, pk_true);
      recalcul_bte (bte, arb_gra, pk_false);
      the_graphicalTrees->getCoord (arb_gra, &nbcoord, bcoord, 0);
      mask = 1022;
      x = bcoord[0].x + bcoord[1].x;
      y = bcoord[0].y + bcoord[1].y;
      d_h = bcoord[1].y;
      if (ecart_port (arb_gra, mask, x, y, d_h, pk_true))
	arb_sig_int = ntrees_getNext (arb_sig_int);
      else
	{
	  *noErrMess = decapsimpecpt;
	  ok = pk_false;
	}
    }
  return (ok);
}

/**
  Cette procedure insere dans la liste des ports diffuseurs et
   leurs ports lies (?l_p_d_p_l) les informations concernant
   le port (?port) d'entree ou sortie selon (?en_entree) appartenant 
   a la boite a decapsulee (?bte)
   on cherche le port diffuseur du signal qui passe par port et 
   les ports lies au port en interne et en externe
   on cree un arbre binaire contenant le numero du port diffuseur et
   la listes de tous les ports lies que l'on insere dans  l_p_d_p_l
   on traite uniquement si le port est lie en interne et en externe.
   ?l_bte_dec = liste des numeros des boites internes
*/
void
builder::inser_l_info_pt_decaps (int bte, int port, boolean en_entree, ntrees l_p_d_p_l, ntrees l_bte_dec)
{
  ntrees l_n_p_l_e, l_n_p_l_i, l_n_p_l_ei, arb_port_diff, arb_bte, ch, arb_no_port_lie, info_ports_rout;
  int port_diff;
  boolean b;

  l_n_p_l_e = the_graphicalTrees->get_ports_lies (port, pk_false, &ch, &b);
  if (l_n_p_l_e != null && !nisemptyList (l_n_p_l_e))
    {
      if (en_entree)
	{
	  port_diff = (int) nintof (nchild (1, l_n_p_l_e));
	  arb_bte = the_graphicTablesManager->graphObj_get_tree (bte);
	  arb_port_diff = the_graphicTablesManager->graphObj_get_tree (port_diff);
	  if (the_graphicalTrees->isExternPort (arb_bte, arb_port_diff))
	    l_n_p_l_e = the_graphicalTrees->get_ports_lies (port_diff, pk_true, &ch, &b);
	  else
	    l_n_p_l_e = the_graphicalTrees->get_ports_lies (port_diff, pk_false, &ch, &b);
	  the_graphicalTrees->oter_port_lie_non_decaps (l_n_p_l_e, l_bte_dec);
	}
      l_n_p_l_i = the_graphicalTrees->get_ports_lies (port, pk_true, &ch, &b);
      if (l_n_p_l_i != null && !nisemptyList (l_n_p_l_i))
	{
	  if (!en_entree)
	    {
	      port_diff = (int) nintof (nchild (1, l_n_p_l_i));
	      arb_bte = the_graphicTablesManager->graphObj_get_tree (bte);
	      arb_port_diff = the_graphicTablesManager->graphObj_get_tree (port_diff);
	      l_n_p_l_i = the_graphicalTrees->get_ports_lies (port_diff, pk_false, &ch, &b);
	      the_graphicalTrees->oter_port_lie_bte (bte, l_n_p_l_i);
	    }
	  l_n_p_l_ei = l_n_p_l_e;
	  arb_no_port_lie = nchild (1, l_n_p_l_i);
	  while (arb_no_port_lie != null)
	    {
	      ntreesDestroy (arb_no_port_lie);
	      ntreeList_appendLast (l_n_p_l_ei, arb_no_port_lie);
	      arb_no_port_lie = nchild (1, l_n_p_l_i);
	    }
	  info_ports_rout = LL_make_encaps (LL_make_port_atom (port_diff), l_n_p_l_ei);
	  ntreeList_appendLast (l_p_d_p_l, info_ports_rout);
	}
    }
}


/**
  Cette procedure insere dans la liste des ports diffuseurs et
   leurs ports lies (?l_p_d_p_l) les informations concernant
   le port diffuseur (?port_diff) si le port n'est pas deja present
   dans la liste.
   on cree un arbre binaire contenant le numero du port diffuseur et
   la liste des ports lies que l'on insere dans  l_p_d_p_l
   on detruit le chemin partant du port.
   on traite les ports diffuseurs internes a la boite a decapsulee
   qui n'ont pas de liens externes.
   bte = numero de la boites a decapsulee
*/
void
builder::inser_l_info_pti_decaps (int bte, int port_diff, ntrees l_p_d_p_l)
{
  ntrees info_ports_rout, ch, l_n_p_l_d;
  boolean trouver, b;

  trouver = pk_false;
  if (!nisemptyList (l_p_d_p_l))
    {
      info_ports_rout = nchild (1, l_p_d_p_l);
      while (info_ports_rout != null && !trouver)
	{
	  if (the_graphicalTrees->tree_get_graphObj (info_ports_rout) == port_diff)
	    trouver = pk_true;
	  else
	    info_ports_rout = ntrees_getNext (info_ports_rout);
	}
    }
  if (!trouver)
    {
      l_n_p_l_d = the_graphicalTrees->get_ports_lies (port_diff, pk_false, &ch, &b);
      the_graphicalTrees->oter_port_lie_bte (bte, l_n_p_l_d);
      if (!nisemptyList (l_n_p_l_d))
	{
	  info_ports_rout = LL_make_encaps (LL_make_port_atom (port_diff), l_n_p_l_d);
	  ntreeList_appendLast (l_p_d_p_l, info_ports_rout);
	}
    }
  detruire_chemin (nchild (3, the_graphicTablesManager->graphObj_get_tree (port_diff)), pk_true);
}


/**
  Cette fonction cree la liste des ports diffuseurs et leurs ports
  lies correspondant aux ports a relier lors d'un decapsulage
  on parcourt l'ensemble des ports de la boite a decapsulee (?bte)
  puis la liste des numeros des boites internes (?l_bte_dec)
*/
ntrees
builder::creer_l_info_bte_decaps (int bte, ntrees l_bte_dec)
{
  ntrees l_p_d_p_l, arb_gra, arb_port, arb_no_bte, arb_bte;
  int port;
  boolean en_entree;

  l_p_d_p_l = LL_make_lnports_emptyList ();
  /*liste des ports diffuseurs et leurs ports lies */
  arb_gra = the_graphicTablesManager->graphObj_get_tree (bte);
  if (nfather (arb_gra) != null)
    {
      arb_gra = nfather (arb_gra);
      arb_port = nchild (1, nchild (2, arb_gra));
      while (arb_port != null)
	{
	  port = the_graphicalTrees->tree_get_graphObj (arb_port);
	  if (noper (arb_port) == gpkOp_porte)
	    en_entree = pk_true;
	  else
	    en_entree = pk_false;
	  inser_l_info_pt_decaps (bte, port, en_entree, l_p_d_p_l, l_bte_dec);
	  arb_port = ntrees_getNext (arb_port);
	}
    }
  arb_no_bte = nchild (1, l_bte_dec);
  while (arb_no_bte != null)
    {
      arb_bte = the_graphicTablesManager->graphObj_get_tree ((int) nintof (arb_no_bte));
      if (nfather (arb_bte) != null)
	{
	  arb_port = nchild (1, ntrees_getNext (arb_bte));
	  while (arb_port != null)
	    {
	      port = the_graphicalTrees->tree_get_graphObj (arb_port);
	      if (noper (arb_port) == gpkOp_ports && nchild (3, arb_port) != null)
		inser_l_info_pti_decaps (bte, port, l_p_d_p_l);
	      arb_port = ntrees_getNext (arb_port);
	    }
	}
      arb_no_bte = ntrees_getNext (arb_no_bte);
    }
  return (l_p_d_p_l);
}


/**
  Cette fonction effectue le routage entre le port diffuseur et
  les ports a relier lorsque l'utilisateur a fait un "break".
  ?l_p_d_p_l la liste des ports diffuseurs et leurs ports lies
  composee d'arbres binaires dont le premier fils represente le
  numero du port diffuseur et le deuxieme la liste des numeros
  des ports a relier. 
  ?l_bte_dec la liste des numeros des boites internes.
  Rend faux si le routage n'est pas possible.
*/
boolean
builder::rout_ports_btes_decaps (ntrees l_p_d_p_l, ntrees l_bte_dec)
{
  ntrees arb_port_diff, arb_info_rout, l_no_ports_lies;
  boolean cont;
  int bte_niv;

  cont = pk_true;
  bte_niv = (int) nintof (nchild (1, l_bte_dec));
  arb_info_rout = nchild (1, l_p_d_p_l);
  while (arb_info_rout != null && cont)
    {
      arb_port_diff = the_graphicTablesManager->graphObj_get_tree (the_graphicalTrees->tree_get_graphObj (arb_info_rout));
      l_no_ports_lies = nchild (2, arb_info_rout);
      cont = rout_port_diff_ports_lies (arb_port_diff, l_no_ports_lies, bte_niv);
      arb_info_rout = ntrees_getNext (arb_info_rout);
    }
  return (cont);
}


/**
  Rend vrai s'il est possible d'effectuer le routage entre le port 
  ?arb_port_diff  et la liste de numero de ports ?l_no_ports_lies 
  ?bte_niv a 0 si aucun des ports n'est externe sinon represente le numero de
   la boite par rapport a laquelle on teste si les ports sont externes.
*/
boolean
builder::rout_port_diff_ports_lies (ntrees arb_port_diff, ntrees l_no_ports_lies, int bte_niv)
{
  ntrees arb_no_port_lie, arb_port_lie;
  boolean cont, p_ext_a, p_ext_b;

  cont = pk_true;
  arb_no_port_lie = nchild (1, l_no_ports_lies);
  while (arb_no_port_lie != null && cont)
    {
      arb_port_lie = the_graphicTablesManager->graphObj_get_tree ((int) nintof (arb_no_port_lie));
      if (bte_niv != 0)
	{
	  p_ext_a = the_graphicalTrees->isExternPort (the_graphicTablesManager->graphObj_get_tree (bte_niv), arb_port_lie);
	  p_ext_b = the_graphicalTrees->isExternPort (the_graphicTablesManager->graphObj_get_tree (bte_niv), arb_port_diff);
	}
      else
	{
	  p_ext_a = pk_false;
	  p_ext_b = pk_false;
	}
      cont = rout_port_port (arb_port_lie, arb_port_diff, p_ext_a, p_ext_b);
      arb_no_port_lie = ntrees_getNext (arb_no_port_lie);
    }
  return (cont);
}

/**
  Creation of the ports for the model ?pProc, from the signal declaration 
  list ?lSignal which is the list of inputs of ?pProc (if isInput is true), 
  which is the list of outputs of ?pProc (if isInput is pk_false).
  ?p_bte is the expression box of the process.
  The input (output) ports are created on the left (right) side of the box and
  are anotated with their Signal declaration.
*/
void
builder::createInOrOutInterfacePorts (ntrees pProc, ntrees p_bte, ntrees lSignal, boolean isInput)
{
  ntrees signalIdent, decattr;
  typcoord tcoord[2], pcoord;
  int rg, noport, no_pt, nbport, pas, nbcoord, bsup, binf;

  // printf("DEBUG entering builder::createInOrOutInterfacePorts\n"); fflush(stdout);
  the_graphicalTrees->getCoord (p_bte, &nbcoord, tcoord, 0);
  bsup = tcoord[0].y + tcoord[1].y;
  binf = tcoord[0].y - tcoord[1].y;
  nbport = the_graphicalTrees->nb_elem_list_decl (lSignal);
  pas = (tcoord[1].y * 2) / (nbport + 1);	// space between 2 ports
  if (pas < taillepave_y)
    pas = taillepave_y;		// minimum space between 2 ports 
  if (isInput)
    {
      rg = -1;
      pcoord.x = tcoord[0].x - tcoord[1].x;	// left side of the box
    }
  else
    {
      rg = STree_getNbInputs (pProc) - 1;
      pcoord.x = tcoord[0].x + tcoord[1].x;	// right side of the box
    }
  noport = 1;
  signalIdent = the_graphicalTrees->first_list_dec (lSignal, &decattr);
  while (signalIdent != null)
    {
      pcoord.y = binf + (pas * noport);
      if (pcoord.y <= (bsup - (taillepave_y / 2)))
	{
	  rg++;
	  noport++;
	  no_pt = the_graphicalTrees->creatept_connex (the_graphicalTrees->tree_get_graphObj (p_bte), pcoord, isInput, rg);
	  the_graphicalTrees->assoc_decl_port (the_graphicTablesManager->graphObj_get_tree (no_pt), decattr);
	  signalIdent = the_graphicalTrees->next_list_dec (signalIdent, &decattr);
	}
      else			// error when there is not enought place to draw all the ports 
	{
	  signalIdent = null;
	  the_messageTool->error_mess (crportimp);
	}
    }
  // printf("DEBUG  exiting builder::createInOrOutInterfacePorts\n"); fflush(stdout);
}

/**
  Creation of the ports from the interface of the model ?p_proc  
  whose expression box is ?p_bte.
  Method called when the user imports a textual file.
*/
void
builder::mkInterfacePorts (ntrees pModel, ntrees p_bte)
{
  ntrees lSig;

  lSig = STree_getInputs (pModel);
  createInOrOutInterfacePorts (pModel, p_bte, lSig, pk_true);
  lSig = STree_getOutputs (pModel);
  createInOrOutInterfacePorts (pModel, p_bte, lSig, pk_false);
}


/**
  Cette procedure est appelee par la fonction gere_appel_proc qui est
  appelee lorsque l'utilisateur construit une boite contenant un appel.
  Si le processus appele existe et que la boite contenant l'appel n'a
  pas de ports alors les ports sont automatiquement crees avec leur noms 
  associes.
  Cette procedure cree les ports, recopie les declarations associees
  On met les ports d'entree a gauche et les ports de sortie a droite.
  ?p_bte : boite contenant l'appel de processus
  ?p_proc : declaration du processus appele
*/
void
builder::createports_bte_appel (ntrees p_bte, ntrees p_proc)
{
  ntrees p_port, l_ports, decl_port;
  typcoord tcoord[2], pcoord;
  boolean non_fini;
  int no_boite, noport, x, y, no_pt, rg, nb_ent, nb_sor, pas, nbcoord, bsup, binf;

  // printf("DEBUG entering builder::createports_bte_appel\n"); fflush(stdout);
  no_boite = the_graphicalTrees->no_boite_proc (p_proc);
  nb_ent = the_graphicalTrees->nombre_ports_ent (no_boite);	/* on recupere le nombre des ports d'entree */
  nb_sor = the_graphicalTrees->nombre_ports_sor (no_boite);	/* et de sortie du processus appele */
  if ((nb_ent + nb_sor) != 0)
    {
      the_graphicalTrees->getCoord (p_bte, &nbcoord, tcoord, 0);
      bsup = tcoord[0].y + tcoord[1].y;
      binf = tcoord[0].y - tcoord[1].y;
      l_ports = ntrees_getNext (the_graphicTablesManager->graphObj_get_tree (no_boite));
      /* on s'occupe des entrees */
      x = tcoord[0].x - tcoord[1].x;
      pas = (tcoord[1].y * 2) / (nb_ent + 1);	/* calcul de l'espacement entre les ports */
      if (pas < taillepave_y)
	pas = taillepave_y;	/* espacement the_graphicalTrees->minimum */
      noport = 1;
      rg = -1;
      p_port = nchild (1, l_ports);
      non_fini = pk_true;
      while (non_fini && noport <= nb_ent)
	{
	  while ((p_port != null) && (noper (p_port) == gpkOp_ports))
	    p_port = ntrees_getNext (p_port);
	  if (p_port != null)
	    decl_port = nattrntrees (p_port, SIGNAL_ATTRNUMBER);
	  pcoord.x = x;
	  y = binf + (pas * noport);
	  if (y <= (bsup - (taillepave_y / 2)))
	    {
	      pcoord.y = y;
	      rg = noport - 1;
	      no_pt = the_graphicalTrees->creatept_connex (the_graphicalTrees->tree_get_graphObj (p_bte), pcoord, pk_true, rg);
	      the_graphicalTrees->assoc_decl_port (the_graphicTablesManager->graphObj_get_tree (no_pt), COPY_ATTRIBUTEDSTREE (decl_port));
	      noport++;
	      p_port = ntrees_getNext (p_port);
	    }
	  else			/* cas ou on n'a pas assez de place pour mettre tous les ports */
	    {
	      non_fini = pk_false;
	      the_messageTool->error_mess (crportimp);
	    }
	}
      /* on s'occupe des sorties */
      pas = (tcoord[1].y * 2) / (nb_sor + 1);
      x = tcoord[0].x + tcoord[1].x;
      if (pas < taillepave_y)
	pas = taillepave_y;
      noport = 1;
      p_port = nchild (1, l_ports);
      non_fini = pk_true;
      while (non_fini && noport <= nb_sor)
	{
	  while ((p_port != null) && (noper (p_port) == gpkOp_porte))
	    p_port = ntrees_getNext (p_port);
	  if (p_port != null)
	    decl_port = nattrntrees (p_port, SIGNAL_ATTRNUMBER);
	  pcoord.x = x;
	  y = binf + (pas * noport);
	  if (y <= (bsup - (taillepave_y / 2)))
	    {
	      rg++;
	      pcoord.y = y;
	      no_pt = the_graphicalTrees->creatept_connex (the_graphicalTrees->tree_get_graphObj (p_bte), pcoord, pk_false, rg);
	      the_graphicalTrees->assoc_decl_port (the_graphicTablesManager->graphObj_get_tree (no_pt), COPY_ATTRIBUTEDSTREE (decl_port));
	      noport++;
	      p_port = ntrees_getNext (p_port);
	    }
	  else
	    {
	      non_fini = pk_false;
	      the_messageTool->error_mess (crportimp);
	    }
	}
    }
  // printf("DEBUG  exiting builder::createports_bte_appel\n"); fflush(stdout);
}

/**
  Cette fonction est appelï¿œe lorsqu'on fait un changement de taille, un
  "paste" ou un "break" (decapsuler).
  Elle rend vrai si l'espacement minimal entre les ports est respecte, 
  faux sinon ou si la taille de la boite n'a pas changï¿œe.
   ?demi_h : demi hauteur de la boite
   ?bte : numero de la boite
   ?mask : permet d'obtenir une hauteur paire
   ?xc,yc: coordonnees du centre de la boite
   ?force a vrai les ports se deplacent proportionnelement au changement de 
   taille sinon les ports gardent leurs anciennes coordonnees absolues.
 */
boolean
builder::ecart_port (ntrees bte, int mask, int x_c, int y_c, int demi_h, boolean force)
{
  int i, h, hauteur, binf, bsup, newy, nbcoord;
  ntrees filsc;
  typcoord oldcoord[2], pcoord[1], inter_p;
  boolean result;

  h = abs (demi_h) & mask;
  if (demi_h < 0)
    h = -h;
  hauteur = abs (h);
  binf = (inter_pave / 2) - hauteur;
  bsup = hauteur - (inter_pave / 2);
  the_graphicalTrees->getCoord (bte, &nbcoord, oldcoord, 0);
  filsc = the_graphicalTrees->getFirstPort (bte);
  i = 1;
  result = (hauteur >= (inter_pave / 2));
  while (result && filsc != null)
    {
      the_graphicalTrees->getCoord (filsc, &nbcoord, pcoord, 1);
      if (force)
	if (hauteur == oldcoord[1].y)
	  newy = pcoord[0].y;
	else
	  {
	    newy = (hauteur * pcoord[0].y) / oldcoord[1].y;
	    if (newy < 0)
	      newy--;
	    else
	      newy++;
	  }
      else
	newy = pcoord[0].y + oldcoord[0].y - y_c;
      if ((newy >= binf) && (newy <= bsup))
	{
	  if (i == 1)
	    {
	      inter_p.x = pcoord[0].x;
	      inter_p.y = newy;
	    }
	  else if (((inter_p.x * pcoord[0].x) > 0) && (abs (inter_p.y - newy) < inter_pave - 1))
	    {
	      result = pk_false;
	    }
	  else
	    {
	      inter_p.x = pcoord[0].x;
	      inter_p.y = newy;
	    }
	}
      else
	result = pk_false;
      filsc = ntrees_getNext (filsc);
      i++;
    }
  return (result);
}


/**
   Cette fonction verifie si le port ?noport peut etre deplace en
   ?(x_ret,y_ret).
   y_ret peut etre ajuste (a cause de l'espacement the_graphicalTrees->minimum entre 2 ports)
   Cette fonction est appelee a la suite d'un deplacement d'un segment.
   !rg : la fonction rend le nouveau rang du port (les ports sont
    ordonnes de haut en bas).
*/
boolean
builder::deplac_port (int noport, int *x_ret, int *y_ret, int *rg)
{
  int ou, nbcoord, form, abs_mil, ord_mil, haute, larg, y_ext_b, y_ext_h;
  ntrees g_arb;
  typcoord tcoord[maxcoord];
  boolean trouve;

  trouve = pk_true;
  g_arb = the_graphicalTrees->boite_mere (noport);
  form = the_graphicalTrees->getCoord (g_arb, &nbcoord, tcoord, 0);
  if (form == oprect)
    {
      larg = tcoord[1].x;
      haute = tcoord[1].y;
      abs_mil = tcoord[0].x;
      ord_mil = tcoord[0].y;
      if (*x_ret < abs_mil - larg + (2 * taillepave_x) &&
	  *x_ret > abs_mil - larg - 2 * (taillepave_x) && *y_ret < ord_mil + haute && *y_ret > ord_mil - haute)
	{
	  *x_ret = abs_mil - larg;
	  ou = gauche;
	}
      else
	if (*x_ret < abs_mil + larg + 2 * (taillepave_x) &&
	    *x_ret > abs_mil + larg - 2 * (taillepave_x) && *y_ret < ord_mil + haute && *y_ret > ord_mil - haute)
	{
	  *x_ret = abs_mil + larg;
	  ou = droite;
	}
      else
	trouve = pk_false;
    };
  if (!trouve)
    return (pk_false);
  else
    {
      y_ext_b = ord_mil + haute;
      y_ext_h = ord_mil - haute;
      return (the_graphicalTrees->insert_port (g_arb, noport, ou, y_ret, y_ext_b, y_ext_h, rg));
    }
}
