/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
///
_PUBLIC_
/* Constructor
 */
graph_text::graph_text (ptGraphicTrees pGarb, ptGraphicTablesManager pTab):
the_graphicalTrees (pGarb), the_graphicTablesManager (pTab)
{
}


/**
  Completes the Signal tree sourceSubtree (in) in order to compile it or to
  obtain its textual representation by :
       adding the affectations which corresponds to the links (new identifiers 
       may be created if necessary), 
       adding local declarations,
       cleaning the empty lists (parameters, local signals...).
  The treatment is done in 2 phases :
       an inheritance phase of the ports names and types from the hight level box to 
       the most internal boxes,
       a synthesis phase of the ports names and types from the most internal boxes to the up level.
  The method is recursively called for each sub-process. sourceSubtree is a process or a module.
  */
ntrees
graph_text::completeTree (ntrees sourceSubtree)
{
  ntrees proc_loc, pBox, p_expr, lDecl, child, p_sig;
  int nb, i;

  // nDEBUG_subtree(4, 8, "entering graph_text::completeTree sourceSubtree is", sourceSubtree);
  if (STree_isModule (sourceSubtree))
    {
      lDecl = STreeModel_getLocalDeclarations (sourceSubtree);
      nb = ntreearity (lDecl);
      if (nb == 0)
	{
	  treatEmptyModule (sourceSubtree);
	  // nDEBUG_subtree(4, 8, " exiting graph_text::completeTree 1  sourceSubtree is", sourceSubtree);
	  return sourceSubtree;
	}
      for (i = 1; i <= nb; i++)
	{
	  proc_loc = nchild (i, lDecl);
	  if (STree_isModel (proc_loc))
	    (void) completeTree (proc_loc);
	}
      // nDEBUG_subtree(4, 8, " exiting graph_text::completeTree 2  sourceSubtree is", sourceSubtree);
      return sourceSubtree;
    }

  proc_encours = sourceSubtree;
  pBox = nattrntrees (STreeModel_getBody (sourceSubtree), GRAPHI_ATTRNUMBER);
  // nDEBUG_subtree(4, 8, "     in  graph_text::completeTree 1 pbox is", pBox);
  the_graphicalTrees->verif_interface (sourceSubtree, pBox);
  // nDEBUG_subtree(4, 8, "     in  graph_text::completeTree 2  sourceSubtree is", sourceSubtree);
  if (isEmptyBody (sourceSubtree))
    {
      p_sig = treatExternalProc (sourceSubtree);
      // nDEBUG_subtree(4, 8, " exiting graph_text::completeTree 3  sourceSubtree is", sourceSubtree);
      // nDEBUG_subtree(4, 2, " exiting graph_text::completeTree 3 p_sig is", p_sig);
      return p_sig;
    }

  if ((lDecl = STreeModel_getLocalDeclarations (sourceSubtree)) == null)
    lDecl = STree_mkLocalDeclaration (sourceSubtree);
  else
    {
      proc_loc = nchild (1, lDecl);
      while (proc_loc != null)
	{
	  if (STree_isModel (proc_loc))
	    proc_loc = completeTree (proc_loc);
	  proc_loc = ntrees_getNext (proc_loc);
	}
    }
  proc_cour = sourceSubtree;
  // nDEBUG_subtree(4, 8, "     in  graph_text::completeTree 3 sourceSubtree  sourceSubtree is", sourceSubtree);
  p_expr = STreeModel_getBody (sourceSubtree);
  pBox = nattrntrees (p_expr, GRAPHI_ATTRNUMBER);
  if (pBox != null)
    {
      // nDEBUG_subtree(4, 8, "     in  graph_text::completeTree 4 p_box", pBox);
      boite_proc_encours = the_graphicalTrees->no_boite_proc (proc_cour);
      // printf("DEBUG      in  graph_text::completeTree 4 boite_proc_encours is %i\n", boite_proc_encours); fflush(stdout);
      inheritance (pBox);
      // nDEBUG_subtree(4, 8, "     in  graph_text::completeTree 5 sourceSubtree  sourceSubtree is", sourceSubtree);
      synthesis (p_expr);
      // nDEBUG_subtree(4, 8, "     in  graph_text::completeTree 6 sourceSubtree  sourceSubtree is", sourceSubtree);
    }
  treatEmptyListParam (sourceSubtree);
  // nDEBUG_subtree(4, 8, " exiting graph_text::completeTree * sourceSubtree is  sourceSubtree is", sourceSubtree);
  return sourceSubtree;
}


/**
  Method called before the compilation of p_proc_comp (in) and necessary when p_proc_comp
  is a sub-process; in this case, if p_proc_comp calls processes that are not local
  to it, the method will insert them in its local declarations.
  Searches in the p_proc_comp expression and in its sub-processes all the 
  process calls and for each call, verifies that the called process is defined and 
  visible. If it is not visible, the method goes up in the processes hierarchy and if 
  it finds it, inserts its declaration in the local declarations of p_proc_comp.
*/
void
graph_text::prepareLocalCompilation (ntrees p_proc_comp)
{
  boolean fini = pk_false;
  while (!fini)
    {
      fini = pk_true;
      recPrepareLocalCompilation (p_proc_comp, p_proc_comp, &fini);
    }
}


/**
  Adds the mark "compilMark_ATTRNUMBER" to the sub-tree p (in). 
  All sub-trees created in this module are marked
  */
void
graph_text::addMark (ntrees p)
{
  nmakeattrint (p, compilMark_ATTRNUMBER, compilMarkVal);	/* PLG VERY STRANGE compilMarkVal is a constant equal to 1 */
}


/**
  cleans the tree p (in) by deleting all the sub-tree which have the
  mark "compilMark_ATTRNUMBER" and by replacing the external processes by
  processes with an empty body.
  This method is called after a compiling or after a textual saving.
  */
void
graph_text::cleanMarkedTree (ntrees p)
{
  boolean b;
  ntrees lDecl, p_expr;

  if (STree_isModel (p))
    {
      if (STree_isExternModel (p))
	cleanExternalProc (p);
      else
	{
	  lDecl = STreeModel_getLocalDeclarations (p);
	  cleanMarkedDeclTree (lDecl);
	  p_expr = STreeModel_getBody (p);
	  b = cleanMarkedExprTree (p_expr);
	}
    }
  else				// p is a module
    {
      // setNameModelStr(p);

      lDecl = STreeModel_getLocalDeclarations (p);
      if (ntreearity (lDecl) == 0)
	cleanEmptyModule (p);
      else
	cleanMarkedDeclTree (lDecl);
    }
}
