/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
///
_PRIVATE_
#define COPY_ATTRIBUTEDSTREE(ATREE) ncopytree(ATREE, pk_true)
/*
  delete the mark "compilMark_ATTRNUMBER"rep
*/
  void
graph_text::removeMark (ntrees p)
{
  ndestroyattr (p, compilMark_ATTRNUMBER);
}


/**
  returns true if the tree p (in) is marked with the "compilMark_ATTRNUMBER", pk_false in the other case.
*/
boolean
graph_text::isMarked (ntrees p)
{
  return (nattrexist (p, compilMark_ATTRNUMBER));
}


/**
  delete the tree p (in) which has the "compilMark_ATTRNUMBER".
*/
void
graph_text::removeMarkedTree (ntrees p)
{
  removeMark (p);		// necessary otherwise problem in the tree memory managment
  ntreesDestroy (p);
}


/**
  delete all the elements of the liste (in) which are marked with the compilMark_ATTRNUMBER
*/
void
graph_text::cleanMarkedSignalList (ntrees liste)
{
  ntrees p, prec;

  p = nchild (1, liste);
  prec = p;
  while (ISNOT_NILTREE (p))
    {
      if (isMarked (p))
	{
	  removeMarkedTree (p);
	  if (p == prec)
	    {			// the list head has been destoyed
	      p = nchild (1, liste);
	      prec = p;
	    }
	  else
	    p = ntrees_getNext (prec);
	}
      else
	{
	  prec = p;
	  p = ntrees_getNext (p);
	}
    }
}


/**
  deletes all the signals of the declaration p_decl (in) which are marked.
*/
void
graph_text::cleanMarkedSignalDecl (ntrees p_decl)
{
  ntrees liste;

  liste = nchild (2, p_decl);
  cleanMarkedSignalList (liste);
}


/**
  changes the external model p (in) to a model with an empty expression
  (which must be its representation during editing).
*/
void
graph_text::cleanExternalProc (ntrees p)
{
  ntrees p_desc, p_bte, p_expr;

  p_desc = nchild (3, p);
  nchangchild (p_desc, STree_mkEmptyConfinedExpr (), 2);
  p_bte = nattrntrees (p_desc, GRAPHI_ATTRNUMBER);
  ndestroyattr (p_desc, GRAPHI_ATTRNUMBER);
  p_expr = STreeModel_getBody (p);
  if (ISNOT_NILTREE (p_bte))
    nmakeattrntrees (p_expr, GRAPHI_ATTRNUMBER, p_bte);
}


/**
  cleans the declaration l_decl (in), i.e. deletes each sub-tree which is marked. 
  This tree may contain signals, types, constants or processes declarations.
 */
void
graph_text::cleanMarkedDeclTree (ntrees l_decl)
{
  ntrees p, prec;

  if ((l_decl != NULL) && (ntreearity (l_decl) != 0))
    {
      p = nchild (1, l_decl);
      prec = p;
      while (ISNOT_NILTREE (p))
	{
	  if (!isMarked (p))
	    {
	      if (STree_isModel (p))
		cleanMarkedTree (p);
	      else if (STree_isSignalDecl (p))
		cleanMarkedSignalDecl (p);
	      prec = p;
	      p = ntrees_getNext (p);
	    }
	  else
	    {
	      removeMarkedTree (p);
	      if (p == prec)
		{		// on vient d'enlever la tête de liste
		  p = nchild (1, l_decl);
		  prec = p;
		}
	      else
		p = ntrees_getNext (prec);
	    }
	}
    }
}


/**
  recursively cleans the marked expression tree p_expr (in), i.e. deletes each 
  sub-tree which is marked.
*/
boolean
graph_text::cleanMarkedExprTree (ntrees p_expr)
{
  boolean isModified;
  int ope, i, nb;
  ntrees fils, l_sig;

  ope = noper (p_expr);
  switch (ope)
    {
    case opobli:
    case oppconf:
      isModified = cleanMarkedExprTree (nchild (1, p_expr));
      cleanMarkedSignalList (nchild (2, p_expr));
      l_sig = nchild (2, p_expr);
      if (nisemptyList (l_sig) || isMarked (p_expr))
	{
	  fils = nchild (1, p_expr);
	  STree_replaceSon (p_expr, fils, 0);
	  isModified = pk_true;
	}
      break;
    case opdefsig:
    case opappe:
    case opinst:
    case oplequhor:
      isModified = isMarked (p_expr);
      if (isModified)
	removeMarkedTree (p_expr);
      break;
    case oplfeed:
      isModified = pk_true;
      while (isModified)
	{
	  nb = ntreearity (p_expr);
	  i = 1;
	  isModified = pk_false;
	  while ((i <= nb) && !isModified)
	    {
	      isModified = cleanMarkedExprTree (nchild (i, p_expr));
	      i++;
	    }
	  if (i > nb)
	    isModified = pk_false;	/* it is finished */
	}
      if ((ntreearity (p_expr) == 1) && isMarked (p_expr))
	{
	  STree_replaceSon (p_expr, nchild (1, p_expr), 0);
	  isModified = pk_true;
	}
      else
	isModified = pk_false;
      break;
    default:
      isModified = pk_false;
      break;
    }
  return (isModified);		// returns true if the tree has been modified
}


 /**
   inserts the signal whose declaration is dec_sig (in) in the local
   declarations of the model "proc_cou" if dec_sig does not exist in its local 
   signals or in its outputs. The signal identifier is inserted with the type
   ptyp (in) if the type of dec_sig is implicit or if dec_sig has no typ.
   If isInserInit (in) is pk_true, the initialisation contained in dec_sig is
   inserted too; if false, it is not (because of delays and windows).
 */
void
graph_text::inser_signal (ntrees dec_sig, ntrees ptyp, boolean isInsertInit)
{
  ntrees l_locales, l_sorties, dec_ins, nom_ins, typ_ins, init_ins, p_typ_decl, init;
  int ope;

  if (dec_sig != null)
    {
      l_locales = STreeModel_getLocalDeclarations (proc_cour);
      l_sorties = STree_getOutputs (proc_cour);
      if (!STree_isInDeclList (l_locales, dec_sig) && !STree_isInDeclList (l_sorties, dec_sig))
	{
	  nom_ins = COPY_ATTRIBUTEDSTREE (STree_getNameOfDeclaration (dec_sig));
	  ope = STreeDecl_getType (dec_sig, &p_typ_decl);
	  init = the_graphicalTrees->init_decl (dec_sig);
	  if ((ope == opimplicit) || (p_typ_decl == null))
	    typ_ins = COPY_ATTRIBUTEDSTREE (ptyp);
	  else
	    typ_ins = COPY_ATTRIBUTEDSTREE (p_typ_decl);
	  if (isInsertInit)
	    init_ins = COPY_ATTRIBUTEDSTREE (init);
	  else
	    init_ins = null;
	  dec_ins = STree_mkSignalDeclaration (nom_ins, typ_ins, init_ins);
	  addMark (dec_ins);
	  the_graphicalTrees->inser_l_dec_sig (l_locales, dec_ins);
	}
    }
}


/**
  Gives the first port connected to the port "port" (in). If port is an input
  port, the method searches in the boxes with the same level or goes up in
  the boxes hierarchy. If the port is an output port, the method searchs in
  the internal boxes of the box which contains the port.
  If "port" is an interface port (which corresponds to an input or an output of
  a model, the method returns the port itself.
  If "port" is an output port of a terminal box (which contains text), 
  the method returns the port itself.
  */
ntrees
graph_text::racine_signal (ntrees port)
{
  boolean b;
  ntrees result, l_ports_lies, portlie, ch;

  if (noper (port) == gpkOp_porte)
    /* the external ports are searched */
    if ((l_ports_lies = the_graphicalTrees->get_ports_lies (the_graphicalTrees->tree_get_graphObj (port), pk_false, &ch, &b)) != null)
      {
	portlie = the_graphicTablesManager->graphObj_get_tree ((int) nintof (nchild (1, l_ports_lies)));
	result = racine_signal (portlie);
      }
    else
      result = port;
  else
    /* the internal ports are searched */
  if ((l_ports_lies = the_graphicalTrees->get_ports_lies (the_graphicalTrees->tree_get_graphObj (port), pk_true, &ch, &b)) != null)
    {
      portlie = the_graphicTablesManager->graphObj_get_tree ((int) nintof (nchild (1, l_ports_lies)));
      if (is_connexion_directe (the_graphicalTrees->tree_get_graphObj (port), the_graphicalTrees->tree_get_graphObj (portlie)))
	result = port;
      else
	result = racine_signal (portlie);
    }
  else
    result = port;
  return (result);
}


/**
  Returns true if the ports noport_diff (in) and noport_recept (in) belong to
  the same box and are directely connected in this box, otherwise returns false.
 */
boolean
graph_text::is_connexion_directe (int noport_diff, int noport_recept)
{
  ntrees port_diff, port_recept;

  port_diff = the_graphicTablesManager->graphObj_get_tree (noport_diff);
  port_recept = the_graphicTablesManager->graphObj_get_tree (noport_recept);
  return ((noper (port_diff) == gpkOp_porte) && (the_graphicalTrees->boite_mere (noport_diff) == the_graphicalTrees->boite_mere (noport_recept)));
}


/**
  Changes the type of the signal declaration p_decl2 (in) by this of p_decl1 (in) 
  if the type of p_decl2 is implicit.
 */
void
graph_text::synthese_type (ntrees p_decl1, ntrees * p_decl2)
{
  ntrees p_typ1, p_typ2, p1, p2;
  int ope1, ope2, typ1, typ2;

  if ((p_decl1 != null) && (p_decl2 != null))
    ope2 = noper (*p_decl2);
  if (ope2 == opdecsig)
    p2 = *p_decl2;
  else if (ntreearity (*p_decl2) == 1)
    p2 = nchild (1, *p_decl2);
  else
    SET_NILTREE (p2);
  if (ISNOT_NILTREE (p2))
    {
      p_typ2 = nchild (1, p2);
      typ2 = noper (p_typ2);
    }
  ope1 = noper (p_decl1);
  if (ope1 == opdecsig)
    p1 = p_decl1;
  else if (ntreearity (p_decl1) == 1)
    p1 = nchild (1, p_decl1);
  else
    SET_NILTREE (p1);
  if (ISNOT_NILTREE (p1))
    {
      p_typ1 = nchild (1, p1);
      typ1 = noper (p_typ1);
    }
  if ((ISNOT_NILTREE (p1)) && (ISNOT_NILTREE (p2)))
    if (typ2 == opimplicit)
      {
	if (typ1 != opimplicit)
	  nchangchild (p2, COPY_ATTRIBUTEDSTREE (p_typ1), 1);
      }
    else if (typ1 == opimplicit)
      nchangchild (p1, COPY_ATTRIBUTEDSTREE (p_typ2), 1);
}


/**
  Returns true if the port "port" (in) is a port of the interface of the
  current process.
*/
boolean
graph_text::is_port_interface (ntrees port)
{
  return (the_graphicalTrees->no_boite_mere (port) == boite_proc_encours);
}


 /**
   Returns true if, from the diffusor port "port" (in), it is possible to
   reach an interface port (of the current process), returns false otherwise.
  */
boolean
graph_text::is_signal_interface (ntrees port)
{
  boolean result, b;
  ntrees l_ports_lies, n_port_lie, portlie, ch;

  result = pk_false;
  if (noper (port) == gpkOp_porte)
    {
      /* we search the internal ports connected to "port" */
      if ((l_ports_lies = the_graphicalTrees->get_ports_lies (the_graphicalTrees->tree_get_graphObj (port), pk_true, &ch, &b)) != null)
	{
	  n_port_lie = nchild (1, l_ports_lies);
	  while (n_port_lie != null && !result)
	    {
	      portlie = the_graphicTablesManager->graphObj_get_tree ((int) nintof (n_port_lie));
	      if (is_port_interface (portlie))
		result = pk_true;
	      else
		result = is_signal_interface (portlie);
	      n_port_lie = ntrees_getNext (n_port_lie);
	    }
	}
    }
  else
    /* we search the external ports connected to "port" */
  if ((l_ports_lies = the_graphicalTrees->get_ports_lies (the_graphicalTrees->tree_get_graphObj (port), pk_false, &ch, &b)) != null)
    {
      n_port_lie = nchild (1, l_ports_lies);
      while (n_port_lie != null && !result)
	{
	  portlie = the_graphicTablesManager->graphObj_get_tree ((int) nintof (n_port_lie));
	  if (is_port_interface (portlie))
	    result = pk_true;
	  else
	    result = is_signal_interface (portlie);
	  n_port_lie = ntrees_getNext (n_port_lie);
	}
    }
  return (result);
}


/**
  returns the diffusor internal port of the port "port" (in), unless
  it is possible to go up to the process interface from this port;
  in this case, returns NILTREE.
  If "port" is an output port of a terminal box, returns the port itself.
  */
ntrees
graph_text::get_diffuseur_intern (ntrees port)
{
  ntrees rac_sig, result;

  result = null;
  rac_sig = racine_signal (port);
  if (!is_port_interface (rac_sig))
    if (!is_signal_interface (rac_sig))
      result = rac_sig;
  return (result);
}


/**
  Returns true if the signal name p_ch (in) is not in the list of names 
  namesList (global list), and in this case inserts it; otherwise
  returns false.
*/
boolean
graph_text::isUnique (ntrees p_ch)
{
  boolean uniq;
  ntrees p_ch_c, p_ch_i;

  p_ch_i = COPY_ATTRIBUTEDSTREE (p_ch);
  uniq = pk_true;
  if (ntreearity (namesList) != 0)
    {
      p_ch_c = nchild (1, namesList);
      while (p_ch_c != null && uniq)
	{
	  if (STree_isIdentEq (p_ch_c, p_ch_i))
	    uniq = pk_false;
	  p_ch_c = ntrees_getNext (p_ch_c);
	}
    }
  if (uniq)
    namesList = ntreeList_insertFirst (p_ch_i, namesList);
  return (uniq);
}


/**
  returns a new signal declaration built from decl (in). The name of the
  new declaration is unique, for that, the declaration name of decl is
  concatenated with the counter namesCount.
  */
ntrees
graph_text::createNewName (ntrees decl)
{
  ntrees pnom, declu, fils, newnom;
  declu = COPY_ATTRIBUTEDSTREE (decl);
  pnom = STree_getNameOfDeclaration (declu);

  if (noper (pnom) != opnom)  // no reference now!!!
    newnom = makeIdent_ident_int (pnom, namesCount);
  else
    newnom = makeIdent_repstr_int (nstringof (pnom), namesCount);

  if (noper (declu) == opdecsig)
    {
      fils = nchild (1, nchild (2, declu));
      if (noper (fils) == opinit)
	nchangchild (fils, newnom, 1);
      else
	nchangchild (nchild (2, declu), newnom, 1);
    }
  namesCount++;
  if (!isUnique (STree_getNameOfDeclaration (declu)))
    declu = createNewName (decl);
  return (declu);
}


/**
  Returns the next sub-tree from p_sig (in) which is not an unary tree
  (with the masking operator "/" or with the oppconf operator) or which has
  a graphical attribute (a box); for that, goes down in the tree hierarchy.
 */
ntrees
graph_text::avaler_unaire (ntrees p_sig)
{
  if (STree_isUnaryExprProc (noper (p_sig)) && nattrntrees (p_sig, GRAPHI_ATTRNUMBER) == null)
    {
      p_sig = nchild (1, p_sig);
      while (STree_isUnaryExprProc (noper (p_sig)) && nattrntrees (p_sig, GRAPHI_ATTRNUMBER) == null)
	p_sig = nchild (1, p_sig);
    }
  return (p_sig);
}


/**
  Returns the next sub-tree from p_text (in) to treat; for that, goes up until
  a composition tree (operator oplfeed "|") and takes the next son.
*/
ntrees
graph_text::next_p_text (ntrees p_text)
{
  int i;
  ntrees p;

  // nDEBUG_subtree(3, 4, "entering graph_text::next_p_text p_text tree is ", p_text);
  p = p_text;
  while (pk_true)
    {
      // nDEBUG_subtree(10, 3, "     in  graph_text::next_p_text W p tree is ", p);
      p = the_graphicalTrees->remonter_expr_sig (oplfeed, p, &i);
      // nDEBUG_subtree(10, 3, "     in  graph_text::next_p_text W-1 p tree is ", p);
      if (noper (p) == oplfeed)
	{
	  if (ntreearity (p) > i)
	    {
	      p = nchild (i + 1, p);
	      // nDEBUG_subtree(3, 4, " exiting graph_text::next_p_text W-2T-1T p tree is ", p);
	      return p;
	    }
	}
      else
	{
	  // nDEBUG_subtree(3, 4, " exiting graph_text::next_p_text W-2F p tree is ", NILTREE);
	  return NILTREE;
	}
    }
}


/**
  Returns true if the model p_proc (in) has an empty body, false otherwise.
  The body is empty if :
     there is no local declarations and,
     the process expression is empty or is a composition with no sons.
*/
boolean
graph_text::isEmptyBody (ntrees p_proc)
{
  ntrees p_expr, p_decl;
  boolean isEmpty;

  isEmpty = pk_false;
  p_expr = STreeModel_getBody (p_proc);
  if (STree_isList (p_expr))
    {
      if (nisemptyList (p_expr))
	{
	  p_decl = STreeModel_getLocalDeclarations (p_proc);
	  if ((p_decl != null) && (noper (p_decl) == opldecl))
	    isEmpty = nisemptyList (p_decl);
	}
    }
  else if ((noper (p_expr) == opnil) && (noper (nfather (p_expr)) == oppconf))
    {
      p_decl = STreeModel_getLocalDeclarations (p_proc);
      if ((p_decl != null) && (noper (p_decl) == opldecl))
	isEmpty = nisemptyList (p_decl);
    }
  return (isEmpty);
}


/**
  Replaces the list of parameters of the process p_proc (in) by NILTREE if the list is empty.
*/
void
graph_text::treatEmptyListParam (ntrees p_proc)
{
  ntrees decl_param;

  if (STree_isModel (p_proc))
    {
      decl_param = STree_getParameters (p_proc);
      if ((decl_param == null) || ntreearity (decl_param) == 0)
	STree_deleteParameters (p_proc);
    }
}


 /**
   Changes the process expression of the external process p_proc (in) 
   by NILTREE and changes the empty list of parameters by NILTREE.
 */
ntrees
graph_text::treatExternalProc (ntrees p_proc)
{
  ntrees p_desc, p_bte;

  p_desc = nchild (3, p_proc);
  p_bte = nattrntrees (STreeModel_getBody (p_proc), GRAPHI_ATTRNUMBER);
  if (ISNOT_NILTREE (p_bte))
    nmakeattrntrees (p_desc, GRAPHI_ATTRNUMBER, p_bte);
  nchangchild (p_desc, STree_mkEmptyTree (), 2);
  treatEmptyListParam (p_proc);
  return (p_proc);
}


/**
  Changes the empty declaration list of the empty modul p (in) by an empty description of module.
  (cf Signal grammar.)
 */
void
graph_text::treatEmptyModule (ntrees p)
{
  nchangchild (nchild (2, p), NILTREE, 2);
}


/**
  the list of model declarations of the empty module p (in) which is represented 
  by NILTREE is changed by an empty list.
*/
void
graph_text::cleanEmptyModule (ntrees p)
{
  nchangchild (nchild (2, p), LL_make_ldecl_emptyList (), 2);	// change the NILTREE by an empty list
}


 /**
   This method searches if the signal contained in the declaration p_decl (in)
   has been renamed (i.e. appears in an assignment built by this module,
   for example, a_1 := a means that the signal a has been renamed to a_1, to
   avoid name conflicts).
   Returns a new declaration with the new name if the signal has been renamed,
   returns NILTREE if not.
   if isInput (in) is pk_true, the signal is searched in the left side of the 
   assignment, if isInput is false, it is searched in the right side.
   The search is done by going up in the tree, from p_text (in).
 */
ntrees
graph_text::getNewSignalName (boolean isInput, ntrees p_decl, ntrees p_text)
{
  ntrees arb, father, result, typsign, arb_typ_dec, nom, p_renom;
  boolean isFound;
  int i, nb, typ_dec;

  isFound = pk_false;
  result = null;
  // nDEBUG_subtree(3, 4, "entering graph_text::getNewSignalName p_decl is ", p_decl);
  // nDEBUG_subtree(3, 4, "entering graph_text::getNewSignalName p_text is ", p_text);
  typ_dec = STreeDecl_getType (p_decl, &arb_typ_dec);
  if (arb_typ_dec != null)
    typsign = COPY_ATTRIBUTEDSTREE (arb_typ_dec);
  else
    typsign = null;
  father = the_graphicalTrees->remonter_expr_sig (oplfeed, p_text, &i);
  // nDEBUG_subtree(3, 4, "     in  graph_text::getNewSignalName 3 father is ", father);

  if ((noper (father) == oplfeed) && (nattrntrees (father, GRAPHI_ATTRNUMBER) == null))
    {
      nb = ntreearity (father);
      i = 1;
      while (!isFound && (i <= nb))
	{
	  p_renom = nchild (i, father);
	  if ((noper (p_renom) == opdefsig) && (nattrntrees (p_renom, GRAPHI_ATTRNUMBER) == null))
	    {			/* on est sur une affectation rajoutee pour un lien */
	      if (isInput)
		nom = nchild (1, p_renom);
	      else
		nom = nchild (2, p_renom);
	      isFound = STree_isIdentEq (nom, STree_getNameOfDeclaration (p_decl));
	    }
	  i++;
	}
      if (isFound)
	{
	  if (isInput)
	    arb = COPY_ATTRIBUTEDSTREE (nchild (2, p_renom));
	  else
	    arb = COPY_ATTRIBUTEDSTREE (nchild (1, p_renom));
	  result = STree_mkSignalDeclaration (arb, typsign, null);
	  addMark (result);
	}
    }
  // nDEBUG_subtree(3, 4, " exiting graph_text::getNewSignalName result is ", result);
  return (result);
}


/**
  Creates, inserts and returns a new composition tree (operator "|") on top 
  of p_sig (in). This composition will be necessary to add assignments
  at the same level of p_sig (which corresponds to the links with different
  names on the ports or when it is necessary to rename signals).
  If the composition has already been created, returns it.
*/
ntrees
graph_text::createComposition (ntrees p_sig)
{
  ntrees p, father;
  int position;

  // nDEBUG_subtree(4, 2, "entering graph_text::createComposition p_sig is", p_sig);
  father = nfather (p_sig);
  // nDEBUG_subtree(4, 3, "     in  graph_text::createComposition father is", father);
  position = nposit (p_sig);
  // printf("DEBUG     in  graph_text::createComposition position is %i\n", position); fflush(stdout);
  if (noper (father) == oplfeed)
    {
      if (position > 1)
	p = nchild (1, father);
      else
	p = nchild (2, father);
      // nDEBUG_subtree(4, 3, "     in  graph_text::createComposition 2T-1 p is", p);
      if ((p == null) || (noper (p) != opdefsig) || (nattrntrees (p, GRAPHI_ATTRNUMBER) != null))
	{
	  /* building of the composition and insertion */
	  p = LL_make_lfeed_list (p_sig);
	  // nDEBUG_subtree(4, 3, "     in  graph_text::createComposition 2T-2T-1 p is", p);
	  addMark (p);
	  // nDEBUG_subtree(4, 3, "     in  graph_text::createComposition 2T-2T-2 p is", p);
	  nchangchild (father, p, position);
	  // nDEBUG_subtree(4, 3, " exiting graph_text::createComposition 2T-2T-3 p is", p);
	  // nDEBUG_subtree(4, 3, " exiting graph_text::createComposition 2T-2T-3 father is", father);
	  return (p);
	}
      // nDEBUG_subtree(4, 3, " exiting graph_text::createComposition 2T-2F p is", father);
      return father;
    }
  p = LL_make_lfeed_list (p_sig);	/* building of the composition */
  // nDEBUG_subtree(4, 3, "     in  graph_text::createComposition 2F-1 p is", p);
  addMark (p);
  // nDEBUG_subtree(4, 3, "     in  graph_text::createComposition 2F-2 p is", p);
  nchangchild (father, p, position);
  // nDEBUG_subtree(4, 3, " exiting graph_text::createComposition 2F-3 p is", p);
  // nDEBUG_subtree(4, 3, " exiting graph_text::createComposition 2F-3 father is", father);
  return (p);
}


/**
  Treats the direct links between an input port (with its declaration p_decl (in)
  and an output port (with its declaration declOut (in) inside a box
  For example, if S is the name of the output port and E the name of the input port,
  the assignment S := E is created.
  The assignment is inserted at the same level as p_text (in)
 */
void
graph_text::treatDirectConnection (ntrees p_text, ntrees declOut, ntrees declIn)
{
  ntrees p_aff, p_sig, nom1, nom2;

  nom1 = STree_getNameOfDeclaration (declIn);
  nom2 = STree_getNameOfDeclaration (declOut);
  p_sig = createComposition (p_text);
  p_aff = LL_make_defsig_tree (COPY_ATTRIBUTEDSTREE (nom2), COPY_ATTRIBUTEDSTREE (nom1));
  addMark (p_aff);
  ntreeList_appendLast (p_sig, p_aff);
}


/**
  * cette procedure masque le signal contenu dans la declaration ?p_decl
  * ?p_text : l'arbre Signal a partir duquel on va acrrocher le masquage.
  * Si ?intern est a vrai, on accroche le masquage juste au-dessus de
  * ?p_text ; si ?intern est a faux, on remonte jusqu'a trouver une
  * composition et on accroche le masquage au-dessus.
  */
void
graph_text::masquage (boolean intern, ntrees p_text, ntrees p_decl)
{
  ntrees p_sig, p, p_nom, l_sig, p_typ_decl, son;
  int i, ope;
  boolean isFound;

  p_sig = p_text;
  // nDEBUG_subtree(3, 4, "entering graph_text::masquage p_text p_text is ", p_text);
  // nDEBUG_subtree(3, 4, "entering graph_text::masquage p_text p_decl is ", p_decl);
  p_nom = COPY_ATTRIBUTEDSTREE (STree_getNameOfDeclaration (p_decl));
  if (!intern)
    {
      // nDEBUG_subtree(3, 4, "     in  graph_text::masquage 2T p_sig is ", p_sig);
      p_sig = the_graphicalTrees->remonter_expr_sig (oplfeed, p_sig, &i);
      // nDEBUG_subtree(3, 4, "     in  graph_text::masquage 2T-1 p_sig is ", p_sig);
      if (nattrntrees (p_sig, GRAPHI_ATTRNUMBER) == null)
	{			/* cas ou on a rajoute un
				   niveau de composition pour les renommages de signaux, on remonte
				   encore d'un niveau. */
	  // nDEBUG_subtree(3, 4, "     in  graph_text::masquage 2T-2T p_sig is ", p_sig);
	  p_sig = the_graphicalTrees->remonter_expr_sig (oplfeed, p_sig, &i);
	  // nDEBUG_subtree(3, 4, "     in  graph_text::masquage 2T-2T-1 p_sig is ", p_sig);
	}
    }
  // nDEBUG_subtree(3, 4, "     in  graph_text::masquage 3 p_sig is ", p_sig);
  p_sig = the_graphicalTrees->remonter_expr_sig (opobli, p_sig, &i);
  // nDEBUG_subtree(3, 4, "     in  graph_text::masquage 4 p_sig is ", p_sig);
  if (noper (p_sig) == opobli)
    {
      l_sig = nchild (2, p_sig);	/*on regarde si le masquage n'est pas deja fait */
      p = nchild (1, l_sig);
      isFound = pk_false;
      while (p != null && !isFound)
	{
	  isFound = STree_isIdentEq (p, p_nom);
	  p = ntrees_getNext (p);
	}
      if (!isFound)
	p = ntreeList_insertFirst (p_nom, l_sig);
    }
  else
    {
      if ((noper (p_sig) == oplfeed) && (nattrntrees (p_sig, GRAPHI_ATTRNUMBER) == null))
	{
	  /* cas ou on a rajoute un niveau de composition pour les renommages 
	     de signaux, on remonte encore d'un niveau. */
	  // nDEBUG_subtree(3, 4, "     in  graph_text::masquage 2T-5F p_sig is ", p_sig);
	  p_sig = the_graphicalTrees->remonter_expr_sig (oplfeed, p_sig, &i);
	  // nDEBUG_subtree(3, 4, "     in  graph_text::masquage 2T-5F-1 p_sig is ", p_sig);
	}
      l_sig = LL_make_lident_list (p_nom);
      addMark (l_sig);
      son = nchild (i, p_sig);
      ope = noper (son);
      if ((ope != oplfeed) && (ope != opobli) && (ope != oppconf))
	{
	  son = LL_make_lfeed_list (son);
	  addMark (son);
	}
      p = LL_make_obli_tree (son, l_sig);
      addMark (p);
      nchangchild (p_sig, p, i);
    }
  ope = STreeDecl_getType (p_decl, &p_typ_decl);
  inser_signal (p_decl, p_typ_decl, pk_true);
  addMark (p_typ_decl);
  // nDEBUG_subtree(3, 4, " exiting graph_text::masquage * p_decl is ", p_decl);
}


/**
  Cette procedure est appelee par la procedure de renommage qui engendre
  une affectation pour masquer le signal de la déclaration ?p_decl.
  S'il s'agit d'un changement de nom en entree ,on masque le signal de 
  la partie gauche de l'affectation , s'il s'agit d'un changement de 
  nom en sortie, on masque le signal de la partie droite.
  Pour  masquer, on crée un processus confiné (s'il n'existe déjà pas)
  et on insère le signal dans la partie where.
*/
void
graph_text::confinage (ntrees p_sig, ntrees p_decl)
{
  ntrees p, l_decl;
  int i;

  // nDEBUG_subtree(4, 10, "entering graph_text::confinage p_sig is", p_sig);
  // nDEBUG_subtree(4, 10, "entering graph_text::confinage p_decl is", p_decl);
  p_sig = the_graphicalTrees->remonter_expr_sig (oppconf, p_sig, &i);
  // nDEBUG_subtree(4, 10, "     in  graph_text::confinage 1 p_sig is", p_sig);
  if (noper (p_decl) == opldecs)
    p_decl = nchild (1, p_decl);
  // nDEBUG_subtree(4, 10, "     in  graph_text::confinage 2 p_decl is", p_decl);
  p_decl = COPY_ATTRIBUTEDSTREE (p_decl);
  // nDEBUG_subtree(4, 10, "     in  graph_text::confinage 3 p_decl is", p_decl);
  addMark (p_decl);
  // nDEBUG_subtree(4, 10, "     in  graph_text::confinage 4 p_decl is", p_decl);
  if (noper (p_sig) != oppconf)
    {				/* on crée le processus confiné */
      // nDEBUG_subtree(4, 10, "     in  graph_text::confinage 5T p_decl is", p_decl);
      l_decl = LL_make_ldecs_emptyList ();
      addMark (l_decl);
      p = LL_make_pconf_tree (nchild (i, p_sig), l_decl);
      addMark (p);
      nchangchild (p_sig, p, i);
    }
  else
    p = p_sig;
  // nDEBUG_subtree(4, 10, "     in  graph_text::confinage 6 p_decl is", p_decl);
  the_graphicalTrees->inser_l_dec_sig (nchild (2, p), p_decl);
}


 /**
  * cette procedure renomme le signal s1 passe en parametre dans
  * ?p_decl par le signal s2 contenu dans la declaration ?p_decl_lie en 
  * construisant l'affectation s1:=s2 si ?isInput est vrai
  *                            s2:=s1 si ?isInput est faux et en 
  * l'inserant devant ?p_text si ?isInput est vrai, derriere sinon.
  * ?p_text : l'arbre signal a partir duquel on va acrrocher le renommage
  * si ?p_typ est non nul, c'est le type de p_decl (et dans ce cas, p_decl
  * contient uniquement le nom du signal (concerne les ports multiples).
  */
void
graph_text::renommage (boolean isInput, ntrees p_text, ntrees p_decl, ntrees p_typ, ntrees p_decl_lie)
{
  ntrees p, p_aff, p_sig, nom1, nom2, typ_sig;
  int ope;

  // nDEBUG_subtree(4, 10, "entering graph_text::renommge p_text is", p_text);
  // nDEBUG_subtree(4, 10, "entering graph_text::renommge p_decl is", p_decl);
  // nDEBUG_subtree(4, 10, "entering graph_text::renommge p_typ is", p_typ);
  nom1 = STree_getNameOfDeclaration (p_decl);
  nom2 = STree_getNameOfDeclaration (p_decl_lie);
  synthese_type (p_decl, &p_decl_lie);
  // nDEBUG_subtree(4, 10, "     in  graph_text::renommge 1 p_decl_lie is", p_decl_lie);
  p_sig = createComposition (p_text);
  // nDEBUG_subtree(4, 10, "     in  graph_text::renommge 2 p_sig is", p_sig);
  if (isInput)
    {
      // nDEBUG_subtree(4, 10, "     in  graph_text::renommge 3T p_sig is", p_sig);
      p_aff = LL_make_defsig_tree (COPY_ATTRIBUTEDSTREE (nom1), COPY_ATTRIBUTEDSTREE (nom2));
      // nDEBUG_subtree(4, 10, "     in  graph_text::renommge 3T-1 p_aff is", p_aff);
      addMark (p_aff);
      // nDEBUG_subtree(4, 10, "     in  graph_text::renommge 3T-2 p_aff is", p_aff);
      ntreeList_insertFirst (p_aff, p_sig);
      // nDEBUG_subtree(4, 10, "     in  graph_text::renommge 3T-3 p_aff is", p_aff);
      if (p_typ == null)
	{
	  ope = STreeDecl_getType (p_decl, &typ_sig);
	  p = p_decl;
	}
      else
	{			/* on reconstruit un s/arbre declaration de signal a voir si necessaire ? */
	  typ_sig = p_typ;
	  /*          p = LL_make_decsig_tree( COPY_ATTRIBUTEDSTREE(p_typ), 
	     LL_make_lsign_list( COPY_ATTRIBUTEDSTREE(p_decl))); */
	}
      // nDEBUG_subtree(4, 10, "     in  graph_text::renommge 3T-4 p_sig is", p_sig);
      confinage (p_sig, p_decl);
      // nDEBUG_subtree(4, 10, "     in  graph_text::renommge 3T-5 p_sig is", p_sig);
    }
  else
    {
      // nDEBUG_subtree(4, 10, "     in  graph_text::renommge 3F p_sig is", p_sig);
      p_aff = LL_make_defsig_tree (COPY_ATTRIBUTEDSTREE (nom2), COPY_ATTRIBUTEDSTREE (nom1));
      addMark (p_aff);
      ntreeList_appendLast (p_sig, p_aff);
      confinage (p_sig, p_decl);
    }
}


/**===========================================================================
  Cette fonction s'occupe des renommages des ports auxquels sont accrochees
  plusieurs declarations de signaux.
  ?isInput a vrai, le port auquel est accroche ?p_decl est un port d'entree
  Ce port a une declaration multiple et ?p_decl est un des ses declarations.
  ?l_decl_lie est la declaration acrroche a un des ports lies du port en
  cours de traitement; ?l_decl_lie peut etre une declaration simple ou multiple.
  ?p_text : l'arbre Signal auquel on va rajouter le renommage si necessaire.
  2 cas se presentent :
      -  p_decl contient le nom veritablement rentre par l'utilisateur,
         par exemple X :
	       le nom X est present dans l_decl_lie, on ne fait rien (meme
	       nom, donc pas de renommage)
	       le nom X n'est pas present, on ne fait rien (ce port lie n'est
	       pas concerne par le signal X
      - p_decl contient un renommage de X, par exemple X_1 (car X est utilise
      ailleurs dans un autre port)
              le nom X n'est pas present, on ne fait rien 
	      le nom X est present dans l_decl_lie, on fait un renommage
  ===========================================================================*/
void
graph_text::verif_renommage_port_multi (boolean isInput, ntrees p_text, ntrees l_decl_lie, ntrees p_decl)
{
  int i, j, nb, nb_sig;
  boolean isFound, is_renom_necessaire;
  ntrees liste_princ, liste, nom, nom_lie, suf, suf_lie, typ_lie;

  isFound = pk_false;
  is_renom_necessaire = pk_false;
  nom = STree_getNameOfDeclaration (p_decl);
  if (noper (l_decl_lie) == opldecs)
    {
      nb = ntreearity (l_decl_lie);
      liste_princ = nchild (1, l_decl_lie);
    }
  else
    {
      nb = 1;
      liste_princ = l_decl_lie;
    }
  i = 1;
  while ((i <= nb) && !isFound)
    {
      liste = nchild (2, liste_princ);
      typ_lie = nchild (1, liste_princ);
      nb_sig = ntreearity (liste);
      j = 1;
      while ((j <= nb_sig) && !isFound)
	{
	  nom_lie = STree_getNameOfDeclaration (nchild (j, liste));
	  isFound = STree_isIdentEq (nom, nom_lie);
	  if (!isFound)
	    if (noper (nom) == opident)	/* on recherche si les 2 noms ont meme
					   suffixe */
	      {
		suf = nchild (1, nom);
		if (noper (nom_lie) == opident)
		  suf_lie = nchild (1, nom_lie);
		else
		  suf_lie = nom_lie;
		isFound = (nstringof (suf) == nstringof (suf_lie));
		is_renom_necessaire = isFound;
	      }
	  j++;
	}
      i++;
      if (nb > 1)
	liste_princ = ntrees_getNext (liste_princ);
    }
  if (isFound && is_renom_necessaire)
    renommage (isInput, p_text, nom_lie, typ_lie, p_decl);
}


/**------------------------------------------------------------------------
  Fonction appelée par la fonction boite_texte.
  Rend vrai si la boîte correspondant à ?p_sig et toutes les btes de même
  niveau n'ont aucun port. Dans ce cas, on considère que ce sont des btes
  décoratives et donc on fera comme si elles n'existaient pas (comme s'il
  n'y avait que du texte).
  -----------------------------------------------------------------------*/
boolean
graph_text::contient_btes_deco (ntrees p_sig)
{
  boolean res, b;
  int nb, i;
  ntrees p_bte, p;

  res = pk_false;
  if (ISNOT_NILTREE (p_sig))
    {
      p = nfather (p_sig);
      if (noper (p) == oplfeed)
	{
	  nb = ntreearity (p);
	  b = pk_false;
	  for (i = 1; i <= nb; i++)
	    {
	      p_bte = nattrntrees (nchild (i, p), GRAPHI_ATTRNUMBER);
	      if (ISNOT_NILTREE (p_bte))
		if (the_graphicalTrees->bte_avec_port (p_bte))
		  b = pk_true;
	    }
	  res = !b;		/* il y a au moins 1 boite avec des ports */
	}
    }
  return (res);
}


/**
  * Cette fonction descend dans l'arbre ?p_sig (en descendant gauche-droit)
  * et rend le sous_arbre Signal le plus profond qui est attribue 
  * graphiquement (a qui est accroche une boite). En bref, rend l'arbre 
  * Signal qui correspond a la premiere boite la plus interne.
  * S'il n'y a pas de boite interne, rend l'arbre ?p_sig lui-meme.
  * Attention, du fait qu'on rajoute des affectations pour effectuer les
  * renommages, a un meme niveau, certains arbres Signal peuvent etre 
  * attribues graphiquement et d'autres pas (on saute ces affectations).
  */
ntrees
graph_text::boite_texte (ntrees p_sig)
{
  ntrees p_gra, p_aux, p_retour, father;
  int i;

  // nDEBUG_subtree(4, 10, "entering graph_text::boite_texte p_sig is", p_sig);
  p_sig = avaler_unaire (p_sig);
  p_gra = nattrntrees (p_sig, GRAPHI_ATTRNUMBER);
  while (p_gra != null && p_sig != null && noper (p_sig) == oplfeed)
    {
      father = p_sig;		// Modif 27 juin 2000
      p_sig = nchild (1, p_sig);
      if (p_sig != null)
	{
	  p_sig = avaler_unaire (p_sig);
	  p_gra = nattrntrees (p_sig, GRAPHI_ATTRNUMBER);
	}
    }
  if (p_sig == null)
    p_retour = father;
  else
    {
      p_aux = p_sig;
      /* on teste s'il s'agit d'un niveau rajoute
         pour les affectations dues aux renommages de signaux */
      if ((noper (p_sig) == oplfeed) && (nattrntrees (p_sig, GRAPHI_ATTRNUMBER) == null))
	{
	  p_sig = nchild (1, p_sig);
	  while ((p_sig != null) && (noper (p_sig) == opdefsig) && nattrntrees (p_sig, GRAPHI_ATTRNUMBER) == null)
	    p_sig = ntrees_getNext (p_sig);
	  if ((p_sig == null) || (nattrntrees (p_sig, GRAPHI_ATTRNUMBER) == null))
	    p_sig = p_aux;
	}
      if ((p_sig != null) && (nattrntrees (p_sig, GRAPHI_ATTRNUMBER) == null))
	{
	  // nDEBUG_subtree(4, 10, "     in graph_text::boite_texte p_sig is", p_sig);
	  p_retour = the_graphicalTrees->remonter_expr_sig (oplfeed, p_sig, &i);
	  // nDEBUG_subtree(4, 10, "     in graph_text::boite_texte p_retour is", p_retour);
	}
      else
	p_retour = p_sig;
    }
  if (contient_btes_deco (p_retour))	/* si toutes les btes de même niveau que  p_retour */
    p_retour = nfather (p_retour);	/* n'ont aucun port, on remonte d'un cran */
  // nDEBUG_subtree(4, 10, " exiting graph_text::boite_texte p_retour is", p_retour);
  return (p_retour);
}


/**
  Cette fonction rend l'arbre Signal qui correspond a la boite
  courante. ?p_sig peut etre :
    un masquage suivi d' (ou) une composition rajoutee pour les
  renommages des signaux; dans ce cas, on recherche le premier
  fils qui a un attribut graphique (les affectations qu'on rajoute
  n'ont pas d'attributs).
*/
ntrees
graph_text::boite_meme_niv (ntrees p_sig)
{
  ntrees p;
  int ope;

  ope = noper (p_sig);
  if (((ope == opobli) || (ope == oppconf)) && (nattrntrees (p_sig, GRAPHI_ATTRNUMBER) == null))
    p_sig = nchild (1, p_sig);
  if ((noper (p_sig) == oplfeed) && (nattrntrees (p_sig, GRAPHI_ATTRNUMBER) == null))
    {				/* cas ou on est dans un sous-arbre ou on a rajoute des affectations 
				   pour les renommages, on les saute */
      p = nchild (1, p_sig);
      while ((p != null) && (nattrntrees (p, GRAPHI_ATTRNUMBER) == null))
	p = ntrees_getNext (p);
    }
  else
    p = the_graphicalTrees->desc_expr_sig (p_sig);
  return (p);
}


/**
  Copies the initialisation of the signal declaration p_decl1 (in) if it exists in the
  declaration p_decl2 (in)
 */
void
graph_text::synthese_init (ntrees p_decl1, ntrees p_decl2)
{
  ntrees nom1, nom2, init, p;

  if (p_decl1 != null)
    if ((noper (p_decl1) == opdecsig) && (noper (p_decl2) == opdecsig))
      {
	nom1 = nchild (1, nchild (2, p_decl1));
	if (noper (nom1) == opinit)
	  {
	    nom2 = nchild (1, nchild (2, p_decl2));
	    p = COPY_ATTRIBUTEDSTREE (nchild (2, nom1));
	    if (noper (nom2) == opinit)
	      nchangchild (nom2, p, 2);
	    else
	      {
		init = LL_make_init_tree (nom2, p);
		nchangchild (nchild (2, p_decl2), init, 1);
	      }
	  }
      }
}


/**
  Returns true if the number of port (in) belongs to the list l_ports_lies, 
  otherwise returns false. 
  This list is a list of port numbers and is created during the synthesis.
*/
boolean
graph_text::is_in_list_ports (ntrees port, ntrees linkedPortsList)
{
  int i, nb, no_port;
  boolean notFound;

  no_port = the_graphicalTrees->tree_get_graphObj (port);
  nb = ntreearity (linkedPortsList);
  i = 1;
  notFound = pk_true;
  while (i <= nb && notFound)
    {
      notFound = (no_port != (int) nintof (nchild (i, linkedPortsList)));
      i++;
    }
  return (!notFound);
}


/**
  Cette fonction rend vrai si le nom ?p_nom qui correspond à un nom de 
  port de sortie d'une boite interne n'existe pas comme nom d'une 
  entrée du processus auquel appartient le port. (si on a le meme nom 
  pour une entree et une sortie locale, il faudra faire un renommage).
  On ne regarde pas avec les sorties du processus car une sortie locale
  est forcemment masquée.
*/
boolean
graph_text::is_nom_port_unique_interf (ntrees p_nom)
{
  boolean uniq;
  ntrees p_port, p_decl;

  p_port = the_graphicalTrees->getFirstPort (the_graphicTablesManager->graphObj_get_tree (boite_proc_encours));
  uniq = pk_true;
  while ((ISNOT_NILTREE (p_port)) && uniq)
    {
      if (noper (p_port) == gpkOp_porte)	/* on ne regarde que les ports d'entree */
	{
	  p_decl = nattrntrees (p_port, SIGNAL_ATTRNUMBER);
	  if (p_decl != null)
	    if (noper (p_decl) == opdecsig)
	      uniq = !the_graphicalTrees->meme_nom (p_nom, p_decl);
	    else
	      uniq = !STree_isInDeclList (p_decl, p_nom);
	}
      p_port = ntrees_getNext (p_port);
    }
  return (uniq);
}


/**
  Returns true if the name p_nom (in) of the port p_port (in) is the only one
  for all the ports of all the boxes of same level as the box which contains
  p_port, except for the ports directly connected to p_port.
  p_sig (in) is the signal tree from which the box (which contains p_port) is attached.
*/
boolean
graph_text::isPortNameUnique (ntrees p_port, ntrees p_nom, ntrees p_sig)
{
  ntrees p_text, p_gra, l_ports, p_decl2, port, chem, linkedPortsList;
  boolean unique, boucl;

  unique = pk_true;
  p_text = boite_texte (p_sig);
  linkedPortsList = the_graphicalTrees->get_ports_lies (the_graphicalTrees->tree_get_graphObj (p_port), pk_false, &chem, &boucl);
  while ((p_text != null) && !STree_isBodyTree (p_text) && unique)
    {
      p_gra = nattrntrees (p_text, GRAPHI_ATTRNUMBER);
      if (p_gra != null)
	if (noper (p_gra) == opschema)
	  {
	    l_ports = nchild (2, p_gra);
	    if (l_ports != null)
	      {
		port = nchild (1, l_ports);
		while ((port != null) && unique)
		  {
		    if (p_port != port)	/*on ne compare pas p_port avec lui-meme */
		      {
			if (!is_in_list_ports (port, linkedPortsList))
			  {
			    p_decl2 = nattrntrees (port, SIGNAL_ATTRNUMBER);
			    if (p_decl2 != null)
			      if (noper (p_decl2) == opdecsig)
				unique = !the_graphicalTrees->meme_nom (p_nom, p_decl2);
			      else
				unique = !STree_isInDeclList (p_decl2, p_nom);
			  }
		      }
		    port = ntrees_getNext (port);
		  }
	      }
	  }
      p_text = next_p_text (p_text);
      if (p_text != null)
	p_text = boite_meme_niv (p_text);
    }
  return (unique);
}


/**
  * cette procedure est appelee par la procedure d'inheritance et de 
  * synthese; elle  propage recursivement la declaration ?p_decl du 
  * port ?noport comme attribut a tous les ports lies au port.
  * elle engendre les renommages et masquages necessaires.
  */
void
graph_text::propagatePortsNames (int noport, ntrees * p_decl)
{
  ntrees port, linkedPortsList, chem, p_sig;
  int niveau;
  boolean port_ext, b;
  // printf("DEBUG entering graph_text::propagatePortsNames, noport is %i\n", noport); fflush(stdout);
  // nDEBUG_subtree(4, 10, "entering graph_text::propagatePortsNames p_decl is", *p_decl);
  port = the_graphicTablesManager->graphObj_get_tree (noport);
  // nDEBUG_subtree(4, 10, "     in  graph_text::propagatePortsNames 1 port is", port);
  /* on descend dans les ports internes */
  if (noper (port) == gpkOp_porte && ((linkedPortsList = the_graphicalTrees->get_ports_lies (noport, pk_true, &chem, &b)) != null))
    {
      // nDEBUG_subtree(4, 10, "     in  graph_text::propagatePortsNames 2T p_decl is", *p_decl);
      // nDEBUG_subtree(4, 10, "     in  graph_text::propagatePortsNames 2T linkedPortsList is", linkedPortsList);
      // nDEBUG_subtree(4, 10, "     in  graph_text::propagatePortsNames 2T chem is", chem);
      propager_loc (noport, p_decl, linkedPortsList, &port_ext, &niveau);
      // nDEBUG_subtree(4, 10, "     in  graph_text::propagatePortsNames 2T-1 p_decl is", *p_decl);
    }
  // nDEBUG_subtree(4, 10, "     in  graph_text::propagatePortsNames 3 p_decl is", *p_decl);
  // nDEBUG_subtree(4, 10, "     in  graph_text::propagatePortsNames 3 port is", port);
  /* on remonte par les ports externes */
  if (noper (port) == gpkOp_ports && ((linkedPortsList = the_graphicalTrees->get_ports_lies (noport, pk_false, &chem, &b)) != null))
    {
      port_ext = pk_false;
      niveau = 0;
      propager_loc (noport, p_decl, linkedPortsList, &port_ext, &niveau);
      if (!port_ext)
	{
	  /* le port de sortie n'a pas de port de sortie externe lie */
	  p_sig = the_graphicalTrees->get_arb_sig (&port);
	  if (niveau == 0)
	    masquage (pk_true, p_sig, *p_decl);
	  else
	    masquage (pk_false, p_sig, *p_decl);
	}
    }
}


/**=======================================================================
  Procedure appelee par la procedure ci-dessus "propagatePortsNames" pour
  propager la declaration ?*p_decl du port ?noport a ses ports lies
  ?linkedPortsList.
  Rappelle recursivement "propagatePortsNames".
  ATTENTION, la gestion des connexions directes ne marchent pas
  complètement, A REVOIR!
  ======================================================================*/
void
graph_text::propager_loc (int noport, ntrees * p_decl, ntrees linkedPortsList, boolean * port_ext, int *niveau)
{
  ntrees p_sig, p_port_lie, arb_port_lie, p_text, p_decl_lie, p;
  boolean isInput;
  int noport_lie;

  // printf("DEBUG entering graph_text::propager_loc, noport is %i\n", noport); fflush(stdout);
  // nDEBUG_subtree(4, 10, "entering graph_text::propager_loc p_decl is", *p_decl);
  // nDEBUG_subtree(4, 10, "entering graph_text::propager_loc linkedPortsList is", linkedPortsList);
  p_port_lie = nchild (1, linkedPortsList);
  // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 1 p_port_lie is", p_port_lie);
  while (p_port_lie != null)
    {
      noport_lie = (int) nintof (p_port_lie);
      arb_port_lie = the_graphicTablesManager->graphObj_get_tree (noport_lie);
      // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-2 arb_port_lie is", arb_port_lie);
      isInput = (noper (arb_port_lie) == gpkOp_porte);
      p_decl_lie = nattrntrees (arb_port_lie, SIGNAL_ATTRNUMBER);
      if (is_connexion_directe (noport, noport_lie))
	{			/* il s'agit d'une connexion directe E-->S d'une meme boite */
	  // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3T p_decl_lie is", p_decl_lie);
	  p = arb_port_lie;
	  p_sig = the_graphicalTrees->get_arb_sig (&p);
	  // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3T-1 p_sig is", p_sig);
	  if (!the_graphicalTrees->meme_nom (*p_decl, p_decl_lie))
	    {
	      // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3T-2T p_decl is", *p_decl);
	      treatDirectConnection (p_sig, p_decl_lie, *p_decl);
	      // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3T-2T-1 p_decl is", *p_decl);
	    }
	}
      else
	{
	  // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F p_decl_lie is", p_decl_lie);
	  // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F arb_port_lie is", arb_port_lie);
	  if (noper (arb_port_lie) == gpkOp_porte)
	    *niveau = 1;
	  else
	    *port_ext = pk_true;
	  if ((p_decl_lie == null) || noper (p_decl_lie) == opnil)
	    {
	      // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F-2T arb_port_lie is", arb_port_lie);
	      nmakeattrntrees (arb_port_lie, SIGNAL_ATTRNUMBER, *p_decl);
	    }
	  else
	    {
	      // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F-2F arb_port_lie is", arb_port_lie);
	      synthese_type (p_decl_lie, p_decl);
	      p = arb_port_lie;
	      p_sig = the_graphicalTrees->get_arb_sig (&p);
	      p_text = boite_texte (p_sig);
	      // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F-2F-3 p_sig is", p_sig);
	      if (p_sig == p_text)
		{		/* port d'une boite terminale (contenant du texte) */
		  // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F-2F-3T p_sig is", p_sig);
		  if (!the_graphicalTrees->is_port_multi (racine_signal (arb_port_lie)))
		    {
		      // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F-2F-3T-1T p_sig is", p_sig);
		      if (!the_graphicalTrees->meme_nom (*p_decl, p_decl_lie))
			{
			  // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F-2F-3T-1T-1T p_sig is", p_sig);
			  renommage (isInput, p_text, p_decl_lie, null, *p_decl);
			}
		    }
		  else
		    verif_renommage_port_multi (isInput, p_text, p_decl_lie, *p_decl);
		  // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F-2F-3T-2 p_sig is", p_sig);
		}
	      else		/* port appartenant a une boite intermediaire ou a la boite processus */
		{
		  // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F-2F-3F p_sig is", p_sig);
		  nmakeattrntrees (arb_port_lie, SIGNAL_ATTRNUMBER, *p_decl);
		}
	      // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F-2F-4 p_decl is", *p_decl);
	      propagatePortsNames (noport_lie, p_decl);
	      // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3F-2F-5 p_decl is", *p_decl);
	    }
	}
      p_port_lie = ntrees_getNext (p_port_lie);
      // nDEBUG_subtree(4, 10, "     in  graph_text::propager_loc 2W-3 p_port_lie is", p_port_lie);
    }
}


 /**
  * Cette procedure fait l'heritage des noms et des types des ports.
  * Elle prend en parametre la boite expression de processus la plus
  * englobante (?motherBox).
  * A partir des ports d'entree et de sortie de cette boite,
  * elle propage le nom et type de ces ports a tous les ports lies.
  */
void
graph_text::inheritance (ntrees motherBox)
{
  ntrees l_ports, p_port, pdecl, pdecl_r, pdecl_rac, pdecl_aux, pdecl_elem, p_sig, p_rac, p_text_rac, p_sig_rac, arb_typ, p;
  boolean isInput, non_renommee;
  int typ;

  // nDEBUG_subtree(4, 100, "entering graph_text::inheritance motherBox is", motherBox);
  non_renommee = pk_true;
  namesList = LL_make_lident_emptyList ();
  if (noper (motherBox) == opschema)
    // nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-1 motherBox is", motherBox);
    if ((l_ports = nchild (2, motherBox)) != null)
      {
	// nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-1 l_ports is", l_ports);
	namesCount = 1;
	p_sig = the_graphicalTrees->get_arb_sig (&motherBox);
	// nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-2 p_sig is", p_sig);
	p_port = nchild (1, l_ports);
	// nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-3 p_ports is", p_port);
	while (p_port != null)
	  {
	    // nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W p_port is", p_port);
	    if ((pdecl = nattrntrees (p_port, SIGNAL_ATTRNUMBER)) != null)
	      {
		// nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T pdecl is", pdecl);
		isInput = (noper (p_port) == gpkOp_porte);
		if (isInput)
		  p_rac = p_port;
		else
		  p_rac = racine_signal (p_port);
		pdecl_aux = pdecl;
		// nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T-2 p_rac is", p_rac);
		while (pdecl_aux != null)
		  {
		    // nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T-3W pdecl_aux is", pdecl_aux);
		    pdecl_elem = the_graphicalTrees->get_elem_decl (&pdecl_aux, pdecl);
		    // nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T-3W-1 pdecl_aux is", pdecl_aux);
		    if (isUnique (STree_getNameOfDeclaration (pdecl_elem)))
		      pdecl_r = pdecl_elem;
		    else
		      {		/* renommage au niveau le plus haut */
			// nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T-3W-2T pdecl_elem is", pdecl_elem);
			pdecl_r = createNewName (pdecl_elem);
			renommage (isInput, p_sig, pdecl_r, null, pdecl_elem);
			typ = STreeDecl_getType (pdecl_r, &arb_typ);
			inser_signal (pdecl_r, arb_typ, pk_true);
		      }
		    // nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T-3W-3 pdecl_r is", pdecl_r);
		    if (!isInput)
		      {		/* on traite le port de sortie diffuseur du signal de sortie */
			// nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T-3W-4T pdecl_r is", pdecl_r);
			pdecl_rac = nattrntrees (p_rac, SIGNAL_ATTRNUMBER);
			// nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T-3W-4T-1 pdecl_rac is", pdecl_rac);
			if ((pdecl_rac == null) || (noper (pdecl_rac) == opnil))
			  nmakeattrntrees (p_rac, SIGNAL_ATTRNUMBER, pdecl_r);
			else
			  {
			    // nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T-3W-4T-2F pdecl_rac is", pdecl_rac);
			    p = p_rac;
			    p_sig_rac = the_graphicalTrees->get_arb_sig (&p);
			    p_text_rac = boite_texte (p_sig_rac);
			    if (p_text_rac == p_sig_rac)
			      if (!the_graphicalTrees->is_port_multi (p_rac))
				{
				  if (!the_graphicalTrees->meme_nom (pdecl_rac, pdecl_r))
				    {
				      renommage (isInput, p_sig_rac, pdecl_rac, null, pdecl_r);
				      non_renommee = pk_false;
				    }
				}
			      else
				verif_renommage_port_multi (isInput, p_sig_rac, pdecl_r, pdecl_rac);
			    else
			      nmakeattrntrees (p_rac, SIGNAL_ATTRNUMBER, pdecl_r);
			  }
		      }
		    // nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T-3W-4 pdecl_r is", pdecl_r);
		    propagatePortsNames (the_graphicalTrees->tree_get_graphObj (p_rac), &pdecl_r);
		    // nDEBUG_subtree(4, 4, "     in  graph_text::inheritance 1T-2T-4W-2T-3W-5 pdecl_r is", pdecl_r);
		  }
	      }
	    p_port = ntrees_getNext (p_port);
	  }
      }
  // nDEBUG_subtree(4, 100, " exiting graph_text::inheritance motherBox is", motherBox);
}


/**
  * Cette fonction effectue la synthese des noms de ports.
  * ?p_exp est une expression de processus qui correspond a une boite 
  * interne du plus bas niveau. On considere
  * que les ports associes a cette boite sont nommes.
  * Si on trouve un signal non renomme et qui n'est pas issu d'un
  * signal de l'interface (atteint par heritage), alors
  * on remonte au port diffuseur du signal puis on effectue a
  * nouveau une phase d'heritage.
  */
void
graph_text::synthesis (ntrees p_exp)
{
  ntrees p_text, p_gra, l_ports, port, ch, arb_typ, pdecl_aux, cop_typ, pdecl_ext_r, p_p_ext;
  ntrees pdecl;			/* declarations de signaux associees a un port */
  ntrees pdecl_elem;
  boolean b;
  int typ;

  p_text = boite_texte (p_exp);
  while ((p_text != null) && !STree_isBodyTree (p_text))
    {
      p_gra = nattrntrees (p_text, GRAPHI_ATTRNUMBER);
      if (p_gra != null)
	if (noper (p_gra) == opschema)
	  if ((l_ports = nchild (2, p_gra)) != null)
	    {
	      port = nchild (1, l_ports);
	      while (port != null)
		{
		  pdecl = nattrntrees (port, SIGNAL_ATTRNUMBER);
		  if ((noper (port) == gpkOp_ports) && (pdecl != null))
		    {		/* On ne regarde que les ports de sortie puisque a partir d'eux, on va atteindre les 
				   ports d'entree non atteints pendant la phase d'heritage */
		      p_p_ext = get_diffuseur_intern (port);
		      pdecl_aux = pdecl;
		      while (pdecl_aux != null)
			{
			  pdecl_elem = the_graphicalTrees->get_elem_decl (&pdecl_aux, pdecl);
			  if ((getNewSignalName (pk_false, pdecl_elem, p_text) == null) || !is_nom_port_unique_interf (pdecl_elem))
			    {
			      if (p_p_ext != null)
				if (the_graphicalTrees->get_ports_lies (the_graphicalTrees->tree_get_graphObj (port), pk_false, &ch, &b) == null)
				  {
				    masquage (pk_true, p_text, pdecl_elem);
				  }
				else	/* le port de sortie est diffuseur interne */
				  {
				    if (isPortNameUnique (port, pdecl_elem, p_exp) && is_nom_port_unique_interf (pdecl_elem))
				      pdecl_ext_r = pdecl_elem;
				    else
				      {
					pdecl_ext_r = createNewName (pdecl_elem);
					renommage (pk_false, p_text, pdecl_elem, null, pdecl_ext_r);
				      }
				    /* synthese de declaration */
				    propagatePortsNames (the_graphicalTrees->tree_get_graphObj (p_p_ext), &pdecl_ext_r);
				    typ = STreeDecl_getType (pdecl_ext_r, &arb_typ);
				    cop_typ = COPY_ATTRIBUTEDSTREE (arb_typ);
				    addMark (cop_typ);
				    inser_signal (pdecl_elem, cop_typ, pk_false);
				  }
			    }
			}
		    }
		  port = ntrees_getNext (port);
		}
	    }
      p_text = next_p_text (p_text);
      if (p_text != null)
	p_text = boite_texte (p_text);
    }
}


/**
  Searches all the processes calls in the expression of p_proc (in) and looks if the
  called process is declared under p_proc_comp (in). If it is not, the method
  searches it on top of p_proc_comp and if the declaration exists, inserts it
  under p_proc_comp.
  Returns true if no process has been inserted, pk_false othwise.
*/
boolean
graph_text::searchProcCallAndInsert (ntrees p_proc_comp, ntrees p_proc)
{
  int ope;
  boolean isFinished, isInserted;
  ntrees p_expr, p, p_sav;

  isInserted = pk_false;
  p_expr = STreeModel_getBody (p_proc);
  isFinished = pk_false;
  if ((p_expr != null) && (noper (p_expr) != opnil))
    while (!isFinished)
      {
	while (p_expr != null)
	  {
	    p_sav = p_expr;
	    ope = noper (p_expr);
	    if (LL_is_atomOp (ope))
	      p_expr = null;
	    else
	      {
		switch (ope)
		  {
		  case opdefsig:
		    p_expr = nchild (2, p_expr);
		    break;
		  case opmoti:
		    p_expr = nchild (3, p_expr);
		    break;
		  case opappe:
		  case opinst:
		    p_expr = STree_getNameCalledModel (p_expr);
		    p = the_graphicalTrees->inser_decl_proc (p_proc_comp, p_proc, p_expr, &isInserted);
		    if ((ISNOT_NILTREE (p)) && isInserted)
		      addMark (p);
		    p_expr = null;
		    break;
		  default:
		    p_expr = nchild (1, p_expr);
		    break;
		  }
	      }
	  }
	p_expr = ntrees_getNext (p_sav);
	while (p_expr == null)
	  {
	    p_sav = nfather (p_sav);
	    p_expr = ntrees_getNext (p_sav);
	  }
	isFinished = (STree_isLocalDeclList (p_expr) || STree_isBodyTree (p_expr));
      }
  return (!isInserted);
}


/**
  Recursive method called by PrepareLocalCompilation : searches in the p_proc_comp (in) 
  expression and recursively in its sub-processes all the process calls and for each 
  call, verifies that the called process is defined and visible.
  If it is not visible, the method goes up in the processes hierarchy and if it finds it, 
  inserts its declaration in the local declarations of p_proc_comp.
  First call of recPrepareLocalCompilation : p_proc_comp and p_proc are the same process,
  then, p_proc is a local process, p_proc_comp doesn't change.
*/
void
graph_text::recPrepareLocalCompilation (ntrees p_proc_comp, ntrees p_proc, boolean * isFinished)
{
  ntrees l_proc_loc, proc_loc;
  boolean res;

  l_proc_loc = STreeModel_getLocalDeclarations (p_proc);
  if (ISNOT_NILTREE (l_proc_loc) && (ntreearity (l_proc_loc) != 0))
    {
      proc_loc = nchild (1, l_proc_loc);
      while (proc_loc != null)
	{
	  if (STree_isModel (proc_loc))
	    recPrepareLocalCompilation (p_proc_comp, proc_loc, isFinished);
	  proc_loc = ntrees_getNext (proc_loc);
	}
    }
  res = searchProcCallAndInsert (p_proc_comp, p_proc);
  *isFinished = *isFinished && res;
}
