/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
//_PUBLIC_


/**
  Constructor
*/
graphicalTrees::graphicalTrees (ptGraphicTablesManager pTab):the_graphicTablesManager (pTab)
{
}


/**
  Initialisations
*/
void
graphicalTrees::initGraphicTrees (void)
{
  chainevide = string2repstr (" ");
}


/**
  Returns the integer which represents the empty string in the strings memory.
*/
int
graphicalTrees::chaine_vide (void)
{
  return (chainevide);
}


 /**
  * cette procedure avale en descente les operateurs signal de validation
  * d'obliteration, de renomage et de rebouclage, jusqu'a trouver un
  * sous-arbre Signal attribue d'un arbre graphique.
  */
ntrees
graphicalTrees::desc_expr_sig (ntrees p)
{
  ntrees arb_gra;

  if (p != null)
    arb_gra = nattrntrees (p, GRAPHI_ATTRNUMBER);
  while ((arb_gra == null) && (p != null) && (STree_isUnaryExprProc (noper (p))))
    {
      p = nchild (1, p);
      if (p != null)
	arb_gra = nattrntrees (p, GRAPHI_ATTRNUMBER);
    }
  return (p);
}


/**
  * Cette procedure remonte a partir de l'arbre ?p_sig les operateurs 
  * Signal de validation, d'obliteration, de renomage et de rebouclage, 
  * jusqu'a trouver un operateur Signal egal a ?ope (on ne remonte que
  * jusqu'a la premiere expression de composition rencontree ou jusqu'au 
  * corps s'il n'y a pas de composition.
  * Rend l'arbre d'operateur ope s'il existe sinon la composition
  * !i : rang de p_sig dans la liste de composition
  */
ntrees
graphicalTrees::remonter_expr_sig (int ope, ntrees p_sig, int *i)
{
  ntrees prec, pfils;

  // nDEBUG_subtree(4, 4, "entering graphicalTrees::remonter_expr_sig p_sig is", p_sig);
  prec = p_sig;
  *i = 1;
  if (p_sig != null)
    {
      p_sig = nfather (p_sig);
      while ((noper (p_sig) != oplfeed) && (noper (p_sig) != ope) && !STree_isBodyTree (p_sig))
	{
	  prec = p_sig;
	  p_sig = nfather (p_sig);
	}
      if (noper (p_sig) == oplfeed)
	{
	  pfils = nchild (1, p_sig);
	  while ((pfils != prec) && (!IS_NILTREE (pfils)))
	    {
	      (*i)++;
	      pfils = ntrees_getNext (pfils);
	    }
	}
    }
  // nDEBUG_subtree(4, 2, " exiting graphicalTrees::remonter_expr_sig p_sig is", p_sig);
  return (p_sig);
}



/**
  Function called to build the hierarchy of processes from ?p in the JTree
  window. If ?isTop is pk_true, ?p is inserted under the root, else 
     if setInAtEnd ?p is inserted under the selected process else before the selected process
  (function called after the loading of a graphical or textual Signal file
  and after the insertion of a duplicated model).
*/
void
graphicalTrees::create_modelInJTree (ntrees p, boolean isTop, bool setInAtEnd)
{
  // printf("DEBUG entering graphicalTrees::create_modelInJTree\n"); fflush(stdout);
  C2Java_initSelectedNode (isTop);
  add_modelInJtree (p, setInAtEnd);
  // printf("DEBUG  exiting graphicalTrees::create_modelInJTree\n"); fflush(stdout);
}

/**
  * insere l'arbre signal ?p comme fils de l'arbre ?pere
  *  Cette fonction est appelee lors d'une duplication de boite.
  */
void
graphicalTrees::inser_arb_sig (ntrees p, ntrees * pere)
{
  ntrees p_gra, p_fils;

  p_gra = nattrntrees (p, GRAPHI_ATTRNUMBER);
  if (p_gra != null)
    {
      if (STree_isLocalDeclList (*pere))	/* on insere un processus */
	{
	  STree_addDeclarationToList (*pere, p);
	  create_modelInJTree (p, pk_false, pk_true);	/* affichage hierarchie processus */
	}
      else
	/* avant d'inserer la boite dupliquee (dans une boite expression de 
	   processus, on teste si la boite courante contient
	   du texte ou du graphique. Si elle contient du texte, on l'ecrase */
      if (noper (*pere) == oplfeed)
	{
	  if ((p_fils = nchild (1, *pere)) != null)
	    if (nattrntrees (desc_expr_sig (p_fils), GRAPHI_ATTRNUMBER) == null)
	      while (p_fils != *pere)
		p_fils = ndestroyListElem (*pere, 1);
	  ntreeList_appendLast (*pere, p);
	}
      else
	{
	  p_gra = nattrntrees (*pere, GRAPHI_ATTRNUMBER);
	  p_fils = LL_make_lfeed_list (p);
	  nmakeattrntrees (p_fils, GRAPHI_ATTRNUMBER, p_gra);
	  STree_replaceSon (*pere, p_fils, 0);
	  *pere = p_fils;
	}
    }
}


/**
  Cette fonction est appelee pour parcourir la liste de declarations 
  de signaux ?liste_decl accrochees a un port, rend a chaque fois la 
  declaration courante et positionne la declaration suivante dans *elem_suiv.
  1ier appel de get_elem_decl : *elem_suiv et liste_decl identiques.
  Exemple: 
  si ?liste_decl vaut integer a,b; real c; boolean b1, b2
  le premier appel rendra le sous-arbre integer a
  le deuxieme appel rendra              integer b
  le troisieme                          real c    etc,etc
*/
ntrees
graphicalTrees::get_elem_decl (ntrees * elem_suiv, ntrees liste_decl)
{
  int rang, nb_liste;
  ntrees p_elem, p_typ;

  if (*elem_suiv == liste_decl)	/* 1 ier appel de la fonction get_elem_decl */
    if (noper (liste_decl) == opdecsig)
      {
	*elem_suiv = null;
	return (liste_decl);
      }
    else
      {
	p_elem = nchild (1, nchild (2, nchild (1, liste_decl)));
	p_typ = nchild (1, nchild (1, liste_decl));
      }
  else
    {
      p_elem = *elem_suiv;
      p_typ = ntrees_getPrevious (nfather (p_elem));
    }
  /* calcul du prochain element *elem_suiv */
  *elem_suiv = ntrees_getNext (p_elem);
  if (*elem_suiv == null)
    {				/* on passe a la sou-liste suivante (s'il y en a une) */
      nb_liste = ntreearity (liste_decl);
      rang = nposit (nfather (nfather (p_elem)));
      if (rang < nb_liste)
	*elem_suiv = nchild (1, nchild (2, nchild (rang + 1, liste_decl)));
    }
  return (STree_mkSignalDeclaration (COPY_ATTRIBUTEDSTREE (p_elem), COPY_ATTRIBUTEDSTREE (p_typ), null));
}


/**
 * Cette procedure verifie que les entrees ou sorties d'un processus 
 * correspondent bien aux declarations associees aux ports du processus, 
 * et si non, met l'interface a jour.
 * ?p_proc : arbre processus
 * ?enentree : a vrai, alors on s'occupe des entrees
 *             a faux, alors on s'occupe des sorties
 * ?l_ports : liste des ports du processus
 * Les declarations sont comparees une par une et des qu'il y a une
 * difference entre un signal de l'interface et la declaration accrochee
 * au port courant, on remplace.
 */
void
graphicalTrees::verif_decl_sig (ntrees p_proc, boolean enentree, ntrees l_ports)
{
  int i, ope_typ_list, ope_typ_port;
  int ind_list_inte, nb_list_inte;	/*indice courant dans la liste de declarations
					   et le nombre de declarations (interface) */
  int ind_list_sig, nb_sig;	/* indice courant dans la liste de signaux courante
				   et nombre de signaux dans la liste */
  ntrees l_sig_inte, p_sig_inte, p_port, decl_port, p, p_sig, list_sig, typ_list,
    p_typ_port, fils, decl_elem, p_decl_suiv, nom_port, nom_sig, p_port_insert;

  // nDEBUG_subtree(4, 100, "entering graphicalTrees::verif_decl_sig p_proc is", p_proc);
  // nDEBUG_subtree(4, 6, "entering graphicalTrees::verif_decl_sig l_ports is", l_ports);
  i = 1;
  if (enentree)
    l_sig_inte = STree_getInputs (p_proc);
  else
    l_sig_inte = STree_getOutputs (p_proc);
  p_port = nchild (1, l_ports);
  p_decl_suiv = null;
  if ((l_sig_inte == null) || (noper (l_sig_inte) == opnil))
    {
      l_sig_inte = LL_make_ldecs_emptyList ();
      STree_replaceInputsOutputsModel (p_proc, l_sig_inte, enentree);
    }
  p_sig_inte = nchild (1, l_sig_inte);
  if (p_sig_inte != null)
    {
      ind_list_inte = 1;
      nb_list_inte = ntreearity (l_sig_inte);
      ope_typ_list = STreeDecl_getType (p_sig_inte, &typ_list);	/* type de la liste de signaux courante */
      list_sig = nchild (2, p_sig_inte);
      p_sig = nchild (1, list_sig);
      ind_list_sig = 1;
      nb_sig = ntreearity (list_sig);
      while ((p_port != null) && (p_sig != null))
	{
	  if (((noper (p_port) == gpkOp_porte) && enentree) || ((noper (p_port) == gpkOp_ports) && !enentree))
	    {
	      decl_port = nattrntrees (p_port, SIGNAL_ATTRNUMBER);
	      decl_elem = decl_port;
	      p_decl_suiv = decl_port;
	      while ((p_decl_suiv != null) && (p_sig != null))
		{
		  decl_elem = get_elem_decl (&p_decl_suiv, decl_port);
		  if (decl_elem != null)
		    {
		      ope_typ_port = STreeDecl_getType (decl_elem, &p_typ_port);
		      if (STree_isTreeEq (p_typ_port, typ_list))
			{
			  nom_sig = p_sig;
			  nom_port = STree_getDeclarationWithoutType (decl_elem);	// modif 13 nov 2001
			  if (!meme_sig (nom_port, nom_sig))
			    STree_replaceSon (list_sig, COPY_ATTRIBUTEDSTREE (nom_port), ind_list_sig);
			  if (ind_list_sig == nb_sig)
			    {	/*la sous-liste est terminee */
			      if (ind_list_inte == nb_list_inte)	/* plus de liste dans l'interface */
				p_sig_inte = null;
			      else
				{
				  ind_list_inte++;
				  p_sig_inte = nchild (ind_list_inte, l_sig_inte);
				  if (p_sig_inte != null)
				    {
				      ope_typ_list = STreeDecl_getType (p_sig_inte, &typ_list);
				      list_sig = nchild (2, p_sig_inte);
				      nb_sig = ntreearity (list_sig);
				      ind_list_sig = 1;
				    }
				}
			    }
			  else
			    ind_list_sig++;
			  if (p_sig_inte != null)
			    p_sig = nchild (ind_list_sig, list_sig);
			  else
			    p_sig = null;
			}
		      else
			{	/* cas ou ce n'est pas le meme type */
			  p_port_insert = COPY_ATTRIBUTEDSTREE (decl_elem);
			  if (nb_sig == 1)
			    STree_replaceSon (l_sig_inte, p_port_insert, ind_list_inte);
			  else
			    {
			      p = ndestroyListElem (list_sig, ind_list_sig);
			      nb_list_inte++;
			      if (ind_list_sig == 1)	/* le signal a remplacer par le port 
							   est en tete de liste */
				ninsertleft (1, p_port_insert, l_sig_inte);
			      else if (ind_list_sig == nb_sig)
				{	/*le signal a remplacer est en fin de liste */
				  ninsertright (ind_list_inte, p_port_insert, l_sig_inte);
				  ind_list_inte++;
				}
			      else
				{
				  p = COPY_ATTRIBUTEDSTREE (p_sig_inte);
				  ninsertright (ind_list_inte, p, l_sig_inte);
				  for (i = 1; i < ind_list_sig; i++)
				    fils = ndestroyListElem (nchild (2, p), 1);
				  for (i = ind_list_sig; i < nb_sig; i++)
				    fils = ndestroyListElem (list_sig, ind_list_sig);
				  ninsertright (ind_list_inte, p_port_insert, l_sig_inte);
				  ind_list_inte++;
				  nb_list_inte++;
				}
			    }
			  ind_list_inte++;
			  p_sig_inte = nchild (ind_list_inte, l_sig_inte);
			  if (p_sig_inte != null)
			    {
			      ope_typ_list = STreeDecl_getType (p_sig_inte, &typ_list);
			      list_sig = nchild (2, p_sig_inte);
			      nb_sig = ntreearity (list_sig);
			      p_sig = nchild (1, list_sig);
			      ind_list_sig = 1;
			    }
			  else
			    p_sig = null;
			}
		    }
		}
	    }
	  p_port = ntrees_getNext (p_port);
	}			/* while ((p_port != null) && (p_sig != null)){...}  */
    }				/* if (p_sig_inte  != null) {...} */
  /* on insere toutes les declarations relatives au dernier port traite
     qui n'ont pas encore ete inserees dans l'interface.  */
  while (p_decl_suiv != null)
    {
      decl_elem = get_elem_decl (&p_decl_suiv, decl_port);
      ntreeList_appendLast (l_sig_inte, decl_elem);
    }
  while (p_port != null)
    {
      if (((noper (p_port) == gpkOp_porte) && enentree) || ((noper (p_port) == gpkOp_ports) && !enentree))
	if ((decl_port = nattrntrees (p_port, SIGNAL_ATTRNUMBER)) != null)
	  {
	    p_decl_suiv = decl_port;
	    while (p_decl_suiv != null)
	      {
		decl_elem = get_elem_decl (&p_decl_suiv, decl_port);
		ntreeList_appendLast (l_sig_inte, COPY_ATTRIBUTEDSTREE (decl_elem));
	      }
	  }
      p_port = ntrees_getNext (p_port);
    }
  /* on detruit toutes les declarations dans l'interface qui ne correspondent
     pas a des ports */
  if (p_sig_inte != null)
    for (i = ind_list_inte; i <= nb_list_inte; i++)
      p = ndestroyListElem (l_sig_inte, ind_list_inte);
  if (ntreearity (l_sig_inte) == 0)
    STree_replaceInputsOutputsModel (p_proc, STree_mkEmptyTree (), enentree);
  // nDEBUG_subtree(4, 20, " exiting graphicalTrees::verif_decl_sig * p_proc is", p_proc);
}


/**
   Verification de la coherence entre l'interface d'un processus ?p_proc
   et les declarations associees aux ports de la boite ?p_gra
   S'il n'a pas la meme chose, l'interface est modifiee en fonction
   des declarations associees aux ports.
   Cette procedure fait appel a la fonction verif_decl_sig (ci-dessus)
*/
void
graphicalTrees::verif_interface (ntrees p_proc, ntrees p_gra)
{
  ntrees p_inte, l_ports;
  int nbports;

  if (ISNOT_NILTREE (p_gra))
    {
      if (noper (p_gra) == opschema)
	{
	  l_ports = nchild (2, p_gra);
	  nbports = ntreearity (l_ports);
	}
      else
	nbports = 0;
      p_inte = STree_getInterface (p_proc);
      if (nbports == 0)
	{			/* cas ou le processus n'a pas de port */
	  STree_replaceInputsOutputsModel (p_proc, STree_mkEmptyTree (), pk_true);
	  STree_replaceInputsOutputsModel (p_proc, STree_mkEmptyTree (), pk_false);
	}
      else
	{
	  verif_decl_sig (p_proc, pk_true, l_ports);
	  verif_decl_sig (p_proc, pk_false, l_ports);
	}
    }
}


/**
 *  Recherche si le modèle de nom ?p_nom existe dans la liste de
 * processus ?list_proc; s'il existe, rend vrai et rend la declaration
 * du processus dans ?p_decl, sinon rend faux.
 * ?p_nom : arbre d'operateur opnom ou opident (correspond au premier fils d'un arbre processus).
 * ATTENTION : la liste ?list_proc peut contenir des declarations de 
 * signaux, types, processus, fonctions.
 */
boolean
graphicalTrees::isModelInList (ntrees p_nom, ntrees list_proc, ntrees * p_decl)
{
  int i, nbproc;
  boolean trouver;
  ntrees proc;

  trouver = pk_false;
  if (list_proc != null)
    {
      nbproc = ntreearity (list_proc);
      i = 1;
      while (!trouver && (i <= nbproc))
	{
	  proc = nchild (i, list_proc);
	  if (STree_isModel (proc))
	    {
	      trouver = eq_idents (p_nom, STreeModelModule_getName (proc));
	      *p_decl = proc;
	    }
	  i++;
	}
    }
  return (trouver);
}


/**
  Cette fonction recherche la declaration de processus correspondant 
  a l'appel ?p_appel en partant du processus ?p_proc_dep et en allant 
  jusqu'au processus ?p_proc_arr.
  Rend la declaration si elle existe.
*/
ntrees
graphicalTrees::chercher_decl_proc (ntrees p_appel, ntrees p_proc_dep, ntrees p_proc_arr)
{
  int nb, i, rang;
  ntrees list_proc_cou, p_decl, p_proc;
  boolean trouve;

  p_proc = p_proc_dep;
  trouve = pk_false;
  while ((p_proc != p_proc_arr) && !trouve && STree_isModel (p_proc))
    {
      list_proc_cou = nfather (p_proc);
      rang = nposit (p_proc);
      i = 1;
      while ((i < rang) && !trouve)
	{
	  p_decl = nchild (i, list_proc_cou);
	  if (STree_isModel (p_decl))
	    trouve = STree_isIdentEq (p_appel, STreeModelModule_getName (p_decl));
	  i++;
	}
      if (!trouve)
	{
	  nb = ntreearity (list_proc_cou);
	  i = rang + 1;
	  while ((i <= nb) && !trouve)
	    {
	      p_decl = nchild (i, list_proc_cou);
	      if (STree_isModel (p_decl))
		trouve = STree_isIdentEq (p_appel, STreeModelModule_getName (p_decl));
	      i++;
	    }
	}
      p_proc = STree_getModelFather (p_proc);
    }
  if (trouve)
    return (p_decl);
  return (null);
}


/**
  Recherche la declaration de processus correspondant a l'appel ?p_appel
  et l'insere si necessaire dans les déclarations locales de ?p_proc_comp.
  ?p_appel     : nom du processus appelé
  ?p_proc_comp : processus a compiler
  ? p_proc     : soit p_proc_comp, soit un processus local a ?p_proc_comp
  Rend une copie de la déclaration trouvée ou sinon NILTREE.
  Rend dans !is_inser : vrai si on a inséré la déclaration du modèle, faux sinon.
*/
ntrees
graphicalTrees::inser_decl_proc (ntrees p_proc_comp, ntrees p_proc, ntrees p_appel, boolean * is_inser)
{
  ntrees proc_a_inser, l_decl_loc;
  boolean trouve;

  SET_NILTREE (proc_a_inser);
  *is_inser = pk_false;
  if (!STree_isIdentEq (nchild (1, p_proc), p_appel))	/* en cas d'appel recursif, on ne fait rien */
    {
      trouve = pk_false;
      l_decl_loc = STreeModel_getLocalDeclarations (p_proc);
      if (l_decl_loc != null)
	{			/* on regarde d'abord dans la liste des processus locaux */
	  trouve = isModelInList (p_appel, l_decl_loc, &proc_a_inser);
	}
      if (!trouve)
	{
	  proc_a_inser = chercher_decl_proc (p_appel, p_proc, p_proc_comp);
	  if (IS_NILTREE (proc_a_inser))
	    {
	      proc_a_inser = chercher_decl_proc (p_appel, p_proc_comp, root);
	      if (proc_a_inser != null)
		{
		  proc_a_inser = dupl_arbre (proc_a_inser);
		  STree_addDeclarationToList (STreeModel_getLocalDeclarations (p_proc_comp), proc_a_inser);
		  *is_inser = pk_true;
		}
	    }
	}
    }
  return (proc_a_inser);
}





/**
  Rend vrai si les 2 boites ?bte1 et ?bte2 sont directement connectees
  entre elles, rend faux sinon.
  ?bte1 et ?bte2 sont des boîtes de même niveau ou ?bte1 est une boîte
  interne de ?bte2.
  Donc, ATTENTION à l'ordre des paramètres lorsqu'on appelle la fonction.
*/
boolean
graphicalTrees::is_btes_connectees (int bte1, int bte2)
{
  ntrees p_bte, p_port, p_port_lie, l_ports_lies, chem;
  boolean is_connect, b;
  int no_port;

  p_bte = the_graphicTablesManager->graphObj_get_tree (bte1);
  p_port = getFirstPort (p_bte);
  is_connect = pk_false;
  while ((ISNOT_NILTREE (p_port)) && !is_connect)
    {
      no_port = tree_get_graphObj (p_port);
      l_ports_lies = get_ports_lies (no_port, pk_false, &chem, &b);
      if (ISNOT_NILTREE (l_ports_lies))
	{
	  p_port_lie = nchild (1, l_ports_lies);
	  while (ISNOT_NILTREE (p_port_lie) && !is_connect)
	    {
	      is_connect = (tree_get_graphObj (boite_mere ((int) nintof (p_port_lie))) == bte2);
	      p_port_lie = ntrees_getNext (p_port_lie);
	    }
	}
      p_port = ntrees_getNext (p_port);
    }
  return (is_connect);
}




/**
  Cette fonction rend le type (Signal) de la déclaration accrochée
  au port ?arbre_port
  Si le port n'a pas de declaration associee, rend 0.
  S'il s'agit d'un port multiple, rend 0 si les ports ont des types
  différents, sinon rend le type commun.
*/
int
graphicalTrees::get_typesignal_port (ntrees arbre_port)
{
  ntrees pdecl;
  int type_port, type_cou, n;
  boolean meme_type;

  type_port = 0;
  pdecl = nattrntrees (arbre_port, SIGNAL_ATTRNUMBER);
  if (pdecl != null)
    if (noper (pdecl) == opdecsig)
      type_port = noper (nchild (1, pdecl));
    else
      {
	n = ntreearity (pdecl);
	if (n == 1)
	  {
	    pdecl = nchild (1, pdecl);
	    if (pdecl != null)
	      if (noper (pdecl) == opdecsig)
		type_port = noper (nchild (1, pdecl));
	  }
	else
	  {
	    pdecl = nchild (1, pdecl);
	    type_port = noper (nchild (1, pdecl));
	    meme_type = pk_true;
	    pdecl = ntrees_getNext (pdecl);
	    while ((ISNOT_NILTREE (pdecl)) && meme_type)
	      {
		type_cou = noper (nchild (1, pdecl));
		meme_type = ((type_port == type_cou)
			     || (type_port == opevent) && (type_cou == oplogical) || (type_port == oplogical) && (type_cou == opevent));
		pdecl = ntrees_getNext (pdecl);
	      }
	    if (!meme_type)
	      type_port = 0;
	  }
      }
  return (type_port);
}


/**
  Cette fonction rend vrai si au port ?p_port sont accroches plusieurs noms de 
  signaux, rend faux sinon.
  Cette fonction est utilisee pour savoir si le lien qui part de ce port est un
  un lien simple ou un bus.
*/
boolean
graphicalTrees::is_port_multi (ntrees p_port)
{
  ntrees p_decl, p_fils;
  boolean res;

  res = pk_false;
  p_decl = nattrntrees (p_port, SIGNAL_ATTRNUMBER);
  if (p_decl != null)
    if (noper (p_decl) == opldecs)
      if (ntreearity (p_decl) > 1)
	res = pk_true;
      else if (ntreearity (p_decl) == 1)
	{
	  p_fils = nchild (1, p_decl);
	  res = (noper (p_fils) == opdecsig) && (ntreearity (nchild (2, p_fils)) > 1);
	}
  return (res);
}




/**
  Vérifie si le signal de déclaration ?p_decl existe en tant que port de 
  la boite ?p_bte, port d'entrée si ?en_entree à vrai, port de sortie sinon;
  Rend le port trouvé, sinon rend NILTREE.
*/
ntrees
graphicalTrees::getPortWithDecl (ntrees p_decl, ntrees p_bte, boolean en_entree)
{
  boolean trouve;
  int ope;
  ntrees p_decl_port, port;

  trouve = pk_false;
  port = getFirstPort (p_bte);
  while ((port != null) && !trouve)
    {
      ope = noper (port);
      if ((en_entree && ope == gpkOp_porte) || (!en_entree && ope == gpkOp_ports))
	{
	  p_decl_port = nattrntrees (port, SIGNAL_ATTRNUMBER);
	  if (p_decl_port != null)
	    if (noper (p_decl_port) == opdecsig)
	      trouve = meme_nom (p_decl_port, p_decl);
	    else
	      trouve = STree_isInDeclList (p_decl_port, p_decl);
	}
      if (!trouve)
	port = ntrees_getNext (port);
    }
  if (trouve) return port;
  return NILTREE;
}


/**
  Fonction appelée par la fonction "add_new_links" (module builder,
  pour la compilation d'une architecture).
  Recherche les ports d'entrée, non connectés et qui ont même nom que le 
  port ?port.
  ?port est forcemment un port de sortie.
  ?p_sig denote une composition; on va faire la recherche dans toutes les
  boites qui sont sous ?p_sig.
  rend la liste des numéros des ports trouvés.
*/
ntrees
graphicalTrees::get_ports_same_name (ntrees p_sig, ntrees port)
{
  ntrees l_nom_ports, decl, p, p_decl, p_bte;
  int nb, i;
  boolean ok;

  l_nom_ports = LL_make_lports_emptyList ();
  decl = nattrntrees (port, SIGNAL_ATTRNUMBER);
  nb = ntreearity (p_sig);
  for (i = 1; i <= nb; i++)
    {
      p_bte = nattrntrees (nchild (i, p_sig), SIGNAL_ATTRNUMBER);
      if (ISNOT_NILTREE (p_bte))
	{
	  p = getFirstPort (p_bte);
	  ok = pk_false;
	  while (ISNOT_NILTREE (p))
	    {
	      if ((noper (p) == gpkOp_porte) && (!is_connected_port (p, pk_false)))
		{
		  p_decl = nattrntrees (p, SIGNAL_ATTRNUMBER);
		  if (ISNOT_NILTREE (p_decl))
		    if (noper (p_decl) == opdecsig)
		      ok = meme_nom (p_decl, decl);
		    else
		      ok = STree_isInDeclList (p_decl, decl);
		  if (ok)
		    ntreeList_insertFirst (LL_make_port_atom (tree_get_graphObj (p)), l_nom_ports);
		}
	      p = ntrees_getNext (p);
	    }
	}
    }
  return (l_nom_ports);
}


/**
  Verifie s'il est possible d'inserer un port. Rend "vrai" si oui, faux
   sinon.
   ?p = arbre representant la boite.
   ?port = si port est a 0, c'est un nouveau port a inserer
          si port est different de 0, il s'agit du deplacement d'un port
          deja existant (il ne faut pas comparer ses nouvelles coordonnees
           avec ses anciennes).
   ?sens : vaut gauche si on insere a gauche, droite sinon.
   ?y_ret : position ou on veut inserer le port ; y_ret peut etre modifie,
   en raison de l'espacement minimum entre 2 ports
   ?y_ext_h : l'ordonne de l'extremite haute de la boite
   ?y_ext_b : l'ordonne de l'extremite bas de la boite (entree) 
   !rang : rang du port dans la liste ?p si le port est insere (ports
   ordonnes de haut en bas).
*/
boolean
graphicalTrees::insert_port (ntrees p, int port, int sens, int *y_ret, int y_ext_b, int y_ext_h, int *rang)
{
  boolean trouve;
  int y1, y2, d1, d2, ajust, noport, nbcoord1, nbcoord2;
  typcoord tcoord[maxcoord], tcoord_rel[maxcoord];
  ntrees s_arb;

  s_arb = p;
  *rang = 0;
  trouve = pk_false;
  s_arb = getFirstPort (p);
  getCoord (s_arb, &nbcoord2, tcoord_rel, 1);
  if (sens == gauche)
    {
      tcoord[0].y = y_ext_h;
      while (tcoord_rel[0].x < 0 && s_arb != null && !trouve)
	{
	  y1 = tcoord[0].y;
	  noport = tree_get_graphObj (s_arb);
	  if (noport == port)
	    {
	      s_arb = ntrees_getNext (s_arb);
	      if (s_arb != null)
		getCoord (s_arb, &nbcoord2, tcoord_rel, 1);
	    }
	  else
	    {
	      getCoord (s_arb, &nbcoord1, tcoord, 0);
	      if (tcoord[0].y >= *y_ret)
		{
		  trouve = pk_true;
		  y2 = tcoord[0].y;
		}
	      else
		{
		  (*rang)++;
		  s_arb = ntrees_getNext (s_arb);
		  if (s_arb != null)
		    getCoord (s_arb, &nbcoord2, tcoord_rel, 1);
		};
	    };
	};
      if (!trouve)
	{
	  y1 = tcoord[0].y;
	  y2 = y_ext_b;
	}
    }
  else				/* sens = droite */
    {
      while (s_arb != null && tcoord_rel[0].x < 0)
	{
	  noport = tree_get_graphObj (s_arb);
	  s_arb = ntrees_getNext (s_arb);
	  if (s_arb != null)
	    getCoord (s_arb, &nbcoord2, tcoord_rel, 1);
	  if (noport == port)
	    {
	      if (s_arb != null)
		{
		  s_arb = ntrees_getNext (s_arb);
		  if (s_arb != null)
		    getCoord (s_arb, &nbcoord2, tcoord_rel, 1);
		}
	    }
	  else
	    (*rang)++;
	};
      if (s_arb == null)
	{
	  if (*y_ret < y_ext_h + taillepave_y)
	    *y_ret = y_ext_h + taillepave_y;
	  if (*y_ret > y_ext_b - taillepave_y)
	    *y_ret = y_ext_b - taillepave_y;
	  return (pk_true);
	};
      tcoord[0].y = y_ext_h;
      while (s_arb != null && !trouve)
	{
	  y1 = tcoord[0].y;
	  noport = tree_get_graphObj (s_arb);
	  if (noport == port)
	    s_arb = ntrees_getNext (s_arb);
	  else
	    {
	      getCoord (s_arb, &nbcoord1, tcoord, 0);
	      if (tcoord[0].y >= *y_ret)
		{
		  trouve = pk_true;
		  y2 = tcoord[0].y;
		}
	      else
		{
		  s_arb = ntrees_getNext (s_arb);
		  (*rang)++;
		}
	    }
	};
      if (!trouve)
	{
	  y1 = tcoord[0].y;
	  y2 = y_ext_b;
	}
    };
  ajust = inter_pave;
  if (y1 == y_ext_h || y2 == y_ext_b)
    ajust = taillepave_y;
  if (y2 - y1 >= (inter_pave + ajust))
    {
      d1 = *y_ret - y1;
      d2 = y2 - *y_ret;
      if (d2 <= d1)
	if (y2 == y_ext_b)
	  {
	    if (d2 < ajust)
	      *y_ret = y2 - ajust;
	  }
	else
	  {
	    if (d2 < inter_pave)
	      *y_ret = y2 - inter_pave;
	  }
      else if (y1 == y_ext_h)
	{
	  if (d1 < ajust)
	    *y_ret = y1 + ajust;
	}
      else
	{
	  if (d1 < inter_pave)
	    *y_ret = y1 + inter_pave;
	};
      return (pk_true);
    }
  else
    return (pk_false);
}



/**
  * Calcul d'un halo(!halo) en fonction des coordonnees dans la structure
  * de donnees (?arb_gra).
  * L'objet graphique est soit un segment soit une boite
  */
boolean
graphicalTrees::calcul_halo (ntrees arb_gra, typcoord * halo)
{
  int form, larg, haute, milx, mily, ope, nbcoord;
  typcoord tcoord[maxcoord];

  ope = noper (arb_gra);
  if (ope != opchemin && ope != gpkOp_porte && ope != gpkOp_ports)
    {
      form = getCoord (nchild (2, arb_gra), &nbcoord, tcoord, 0);
      switch (form)
	{
	case oprect:
	  halo[0].x = tcoord[0].x - (tcoord[1].x + halo_bte);
	  halo[0].y = tcoord[0].y - (tcoord[1].y + halo_bte);
	  halo[1].x = tcoord[0].x + (tcoord[1].x + halo_bte);
	  halo[1].y = tcoord[0].y + (tcoord[1].y + halo_bte);
	  return (pk_true);
	  break;
	case opseg:
	  if (tcoord[0].x == tcoord[1].x)
	    {
	      larg = halo_segm;
	      haute = abs ((tcoord[0].y - tcoord[1].y) / 2);
	      milx = tcoord[0].x;
	      mily = maxi (tcoord[0].y, tcoord[1].y) - haute;
	    }
	  else
	    {
	      larg = abs ((tcoord[0].x - tcoord[1].x) / 2);
	      haute = halo_segm;
	      milx = maxi (tcoord[0].x, tcoord[1].x) - larg;
	      mily = tcoord[0].y;
	    };
	  halo[0].x = milx - larg;
	  halo[0].y = mily - haute;
	  halo[1].x = milx + larg;
	  halo[1].y = mily + haute;
	  return (pk_true);
	  break;
	}
    }
  else
    {
      halo[0].x = 0;
      halo[0].y = 0;
      halo[1].x = 0;
      halo[1].y = 0;
      return (pk_false);
    }
  return (pk_false);
}


/**
 * Calcul d'un halo (!halo) en fonction des coordonnees (?tcoord)
 * et du type d'objet (?typ_obj) (segment ou boite)
 */
void
graphicalTrees::calcul_nouv_halo (int typ_obj, typcoord * tcoord, typcoord * halo)
{
  int larg, haute, milx, mily;

  switch (typ_obj)
    {

    case typ_bte:
      halo[0].x = tcoord[0].x - (tcoord[1].x + halo_bte);
      halo[0].y = tcoord[0].y - (tcoord[1].y + halo_bte);
      halo[1].x = tcoord[0].x + (tcoord[1].x + halo_bte);
      halo[1].y = tcoord[0].y + (tcoord[1].y + halo_bte);
      break;

    case typ_seg:
      if (tcoord[0].x == tcoord[1].x)
	{
	  larg = halo_segm;
	  haute = abs ((tcoord[0].y - tcoord[1].y) / 2);
	  milx = tcoord[0].x;
	  mily = maxi (tcoord[0].y, tcoord[1].y) - haute;
	}
      else
	{
	  larg = abs ((tcoord[0].x - tcoord[1].x) / 2);
	  haute = halo_segm;
	  milx = maxi (tcoord[0].x, tcoord[1].x) - larg;
	  mily = tcoord[0].y;
	};
      halo[0].x = milx - larg;
      halo[0].y = mily - haute;
      halo[1].x = milx + larg;
      halo[1].y = mily + haute;
      break;
    }
}


/**
  * cette procedure annule ou positionne le halo des segments
  * arrivant ou partant d'une liste de ports.
  * ? nbport,prem_port: nombre de ports et le premier port de la liste
  * ? interne: a vrai on s'occupe des segments internes de la boite
  *            sinon on s'occupe des segments externes a la boite
  * ?bougetail : utilise lors d'un changement de taille pour representer
  * les cotes deplaces:
  * pour x: -1 cote gauche, 1 cote droit et 0 x inchange.
  * pour y: -1 cote haut, 1 cote bas et 0 y inchange.
  * ?force a vrai si en changement de taille les ports sont deplaces
  * lors d'un deplacement : bougetail vaut (1,1) et force vrai
  */
void
graphicalTrees::annuler_halo_seg_ports (int nbports, ntrees prem_port, boolean intern, boolean corrige, typcoord * p_bougetail, boolean force)
{
  ntrees lien, chemin, ieme_port;
  int i, nbcoord;
  boolean port_recept, saute;
  typcoord pcoord[1], bougetail;

  bougetail = *p_bougetail;
  ieme_port = prem_port;
  for (i = 1; i <= nbports; i++)
    {
      getCoord (ieme_port, &nbcoord, pcoord, 1);
      switch (bougetail.x)
	{
	case 0:
	  saute = !force;
	  break;

	case 1:
	  if (pcoord[0].x > 0)
	    saute = pk_false;
	  else
	    saute = (!force || bougetail.y == 0);
	  break;
	case -1:
	  if (pcoord[0].x < 0)
	    saute = pk_false;
	  else
	    saute = (!force || bougetail.y == 0);
	}
      if (!saute)
	{
	  port_recept = (noper (ieme_port) == gpkOp_porte);
	  if ((port_recept && !intern) || (!port_recept && intern))
	    {
	      if ((lien = nchild (4, ieme_port)) != null)
		{
		  chemin = the_graphicTablesManager->graphObj_get_tree ((int) nintof (lien));
		  annuler_halo_chemins (chemin, corrige);
		}
	    }
	  else if ((chemin = nchild (3, ieme_port)) != null)
	    annuler_halo_chemins (chemin, corrige);
	}
      ieme_port = ntrees_getNext (ieme_port);
    }
}


 /**
  * cette procedure annule ou positionne le halo des segments arrivant ou 
  * partant des ports rattaches a la boite ?bte
  * ? interne: a vrai segments a l'interieur de la boite sinon exterieur
  * ? corrige: a vrai on positionne le halo sinon on annule
  * ? bougetail et ?force voir la procedure ci-dessus.
  */
void
graphicalTrees::annuler_halo_seg_bte (int bte, boolean intern, boolean corrige, typcoord * bougetail, boolean force)
{
  int nbports;
  ntrees p_bte, prem_port;

  p_bte = the_graphicTablesManager->graphObj_get_tree (bte);
  nbports = get_nb_ports (p_bte);	/* modif 30/12/98 */
  if (nbports != 0)
    {
      prem_port = getFirstPort (p_bte);
      annuler_halo_seg_ports (nbports, prem_port, intern, corrige, bougetail, force);
    }
}


/**
 * Cette procedure annule les connexions externes des ports de la boite ?p_gra
 */
void
graphicalTrees::annuler_chem_ext (ntrees p_gra)
{
  ntrees port, lports;
  int nbports;

  if (noper (p_gra) == opschema)
    lports = nchild (2, p_gra);
  else
    lports = ntrees_getNext (p_gra);
  if (lports != null)
    {
      nbports = ntreearity (lports);
      if (nbports > 0)
	port = nchild (1, lports);
      else
	port = null;
      while (port != null)
	{
	  /* on annule les connexions externes du port */
	  if (noper (port) == gpkOp_porte)
	    nchangchild (port, null, 4);
	  else
	    nchangchild (port, null, 3);
	  port = ntrees_getNext (port);
	}
    }
}


/**
  * parcourt l'arbre des chemins ?chemin, accede au port et au
  * lien contenus aux extremites et remplace dans l'arbre port
  * le numero du lien par celui rencontre dans le chemin.
  */
void
graphicalTrees::repositionner_liens (ntrees chemin)
{
  ntrees l_chem, filsc, extrm, port;
  int nbchem, numlien;

  if ((l_chem = nchild (3, chemin)) != null)
    {
      nbchem = ntreearity (l_chem);
      if (nbchem > 0)
	filsc = nchild (1, l_chem);
      else
	filsc = null;
      while (filsc != null)
	{
	  repositionner_liens (filsc);
	  filsc = ntrees_getNext (filsc);
	}
    }
  extrm = nchild (2, chemin);
  if (extrm != null)
    {
      port = the_graphicTablesManager->graphObj_get_tree ((int) nintof (nchild (2, extrm)));
      numlien = tree_get_graphObj (extrm);
      nchangchild (port, LL_make_lien_atom (numlien), 4);
    }
}


/**
 *  cette fonction rend vrai si la boite ?bte  n'a pas de connection
 *  avec d'autres boites, soit internes (?intern a vrai), soit externes
 *  (?intern a faux).
 */
boolean
graphicalTrees::is_boite_non_liee (int bte, boolean intern)
{
  ntrees port, lien, chemin, arb_gra;
  boolean result, port_recept;

  result = pk_true;
  arb_gra = the_graphicTablesManager->graphObj_get_tree (bte);
  if (nfather (arb_gra) != null)
    {
      port = nchild (1, ntrees_getNext (arb_gra));
      while (port != null && result)
	{
	  if (noper (port) == gpkOp_porte)
	    port_recept = pk_true;
	  else
	    port_recept = pk_false;
	  if ((port_recept && !intern) || (!port_recept && intern))
	    {
	      if ((lien = nchild (4, port)) != null)
		result = pk_false;
	    }
	  else if ((chemin = nchild (3, port)) != null)
	    result = pk_false;
	  port = ntrees_getNext (port);
	}
    }

  return (result);
}


/**
  Cette fonction rend le chemin racine du sous-chemin ou du segment
  ?seg_ou_chem, c'est-a-dire le chemin qui part directement du port 
  diffuseur.
*/
ntrees
graphicalTrees::get_racine_chemin (ntrees seg_ou_chem)
{
  int ope;
  ntrees p;

  p = seg_ou_chem;
  if (ISNOT_NILTREE (p))
    {
      ope = noper (nfather (p));
      while ((ope != gpkOp_porte) && (ope != gpkOp_ports))
	{
	  p = nfather (p);
	  ope = noper (nfather (p));
	}
    }
  return (p);
}



/**
  cette fonction rend le numero du port recepteur du lien auquel appartient
  le segment.
  ATTENTION, cette fonction ne peut s'appliquer que si le segment 
  appartient a un lien simple, c'est-a-dire un seul port diffuseur, un 
  seul port recepteur, sinon il faut utiliser la fonction suivante.
*/
int
graphicalTrees::get_no_port_recept (ntrees seg_ou_chem)
{
  ntrees chemin;

  if (noper (seg_ou_chem) != opchemin)
    chemin = get_chemin (seg_ou_chem);
  else
    chemin = seg_ou_chem;
  return ((int) nintof (nchild (2, nchild (2, chemin))));
}


/**
   Cette fonction recursive rend dans ?l_ports la liste des numeros des
   ports recepteurs du chemin ou du segment ?seg_ou_chem (un chemin peut 
   alimenter plusieurs ports recepteurs.
   L'arbre l_ports est une liste des numeros de port; a l'appel de la
   procedure, l_ports est une liste vide.
*/
void
graphicalTrees::get_ports_recept (ntrees seg_ou_chem, ntrees l_ports)
{
  ntrees lsch, sch, chemin, p_extr;
  int nbsch, no_port;

  if (noper (seg_ou_chem) != opchemin)
    chemin = get_chemin (seg_ou_chem);
  else
    chemin = seg_ou_chem;
  if (chemin_simple (chemin))
    {
      no_port = (int) nintof (nchild (2, nchild (2, chemin)));
      // ???     return(ntreeList_insertFirst(LL_make_port_atom(no_port), l_ports));
      ntreeList_insertFirst (LL_make_port_atom (no_port), l_ports);
    }
  else
    {
      p_extr = nchild (2, chemin);
      if (ISNOT_NILTREE (p_extr))	/* on est dans un chemin terminal */
	l_ports = ntreeList_insertFirst (LL_make_port_atom (nintof (nchild (2, p_extr))), l_ports);
      else
	{			/* on a des sous chemins */
	  lsch = nchild (3, chemin);
	  nbsch = ntreearity (lsch);
	  if (nbsch > 0)
	    {
	      sch = nchild (1, lsch);
	      while (sch != null)
		{
		  p_extr = nchild (2, sch);
		  if (p_extr == null)
		    get_ports_recept (sch, l_ports);
		  else
		    l_ports = ntreeList_insertFirst (LL_make_port_atom (nintof (nchild (2, p_extr))), l_ports);
		  sch = ntrees_getNext (sch);
		}
	    }
	}
    }
}


/**
  * cette fonction rend la liste des numeros des ports qui sont connectes
  * au port passe en parametre ?nport.
  * elle rend aussi l'arbre representant le chemin(!c) et le booleen
  * indiquant si les ports appartiennent a la  meme boite(!boucle) 
  * ?intern: a vrai si on s'interesse aux ports lies en interne.
  */
ntrees
graphicalTrees::get_ports_lies (int nport, boolean intern, ntrees * chemin, boolean * boucle)
{
  int port_dep, bte, ope;
  boolean port_recept;
  ntrees lien, port, l_ports_recept;

  // printf("DEBUG entering graphicalTrees::get_ports_lies gpkport ident is %i\n", nport); fflush(stdout);
  SET_NILTREE (l_ports_recept);
  port = the_graphicTablesManager->graphObj_get_tree (nport);
  ope = noper (port);
  port_recept = (((ope == gpkOp_porte) && !intern) || ((ope == gpkOp_ports) && intern));
  *boucle = pk_false;
  *chemin = null;
  bte = tree_get_graphObj (boite_mere (nport));
  if (port_recept)		/* le port est un port recepteur */
    {
      if ((lien = nchild (4, port)) != null)
	{
	  *chemin = the_graphicTablesManager->graphObj_get_tree ((int) nintof (lien));
	  if (ISNOT_NILTREE (*chemin))	/* blindage */
	    {
	      *chemin = get_racine_chemin (*chemin);
	      port_dep = get_no_port_diff (*chemin);
	      if ((tree_get_graphObj (boite_mere (port_dep)) == bte) && (ope == gpkOp_porte))
		*boucle = pk_true;
	      l_ports_recept = LL_make_lports_list (LL_make_port_atom (port_dep));
	    }
	}
    }
  else
    {				/* le port est un port diffuseur */
      if ((*chemin = nchild (3, port)) != null)
	{
	  l_ports_recept = LL_make_lports_emptyList ();
	  get_ports_recept (*chemin, l_ports_recept);
	  if (ntreearity (l_ports_recept) == 1)	/* on a un chemin simple */
	    {
	      port_dep = (int) nintof (nchild (1, l_ports_recept));
	      if ((tree_get_graphObj (boite_mere (port_dep)) == bte) && (noper (port) == gpkOp_ports))
		*boucle = pk_true;
	    }
	}
    }
  return (l_ports_recept);
}


 /**
   Fonction appelée lors du déplacement d'un segment:
       - si le segment appartient à un chemin multiple, rend NILTREE
        (pour le moment, déplacement de ce type de segment non autorisé)
       - si le segment relie directement le port récepteur au port 
         emetteur (le lien est composé d'un seul segment), rend le
	 port emetteur et, le port récepteur dans !port_bis;
       - si le segment est directement relié au port emetteur, rend
         le port récepteur et NILTREE dans !port_bis;
       - si le segment est directement relié au port récepteur, rend
         le port récepteur et NILTREE dans !port_bis.
 */
ntrees
graphicalTrees::getMovedPorts (int seg, ntrees * port_bis)
{
  ntrees result, p_seg;

  *port_bis = null;
  p_seg = the_graphicTablesManager->acces_segment (seg);
  if (chemin_simple (p_seg))
    if (ntrees_getPrevious (p_seg) == null)
      {				/* le segment demarre du port diffuseur */
	if (ntrees_getNext (p_seg) == null)
	  /* le segment arrive sur un port, donc c'est un segment entre 2 ports */
	  *port_bis = the_graphicTablesManager->graphObj_get_tree (get_no_port_recept (p_seg));
	result = get_port_diff (p_seg);
      }
    else if (ntrees_getNext (p_seg) == null)	/* le segment arrive sur le port recepteur */
      result = the_graphicTablesManager->graphObj_get_tree ((int) nintof (nchild (2, ntrees_getIPrevious (2, (nfather (p_seg))))));
    else
      result = get_port_diff (p_seg);
  else
    result = null;
  return (result);
}


/**
  * Cette procedure modifie la hauteur et la largeur de la boite ?bte
  * et modifie les coordonnees des ports
  * ?force a vrai les ports sont deplaces proportionnelement au
  * changement de taille sinon ils gardent leurs anciennes coordonnees absolues 
  */
void
graphicalTrees::modif_port_bte (ntrees bte, typcoord * oldcoord, typcoord * tcoord, boolean force)
{
  ntrees p, filsc;
  int nbcoord, hauteur, largeur, reste, y_abs, y_centre;
  typcoord pcoord_rel[1], newcoord[1];

  hauteur = tcoord[1].y;
  largeur = tcoord[1].x;
  p = nchild (2, nchild (2, bte));	/* on modifie la hauteur et la largeur de la boite */
  nmodifvalint (nchild (1, p), largeur);
  nmodifvalint (nchild (2, p), hauteur);
  filsc = getFirstPort (bte);
  while (filsc != null)
    {
      getCoord (filsc, &nbcoord, pcoord_rel, 1);
      if (pcoord_rel[0].x <= 0)
	newcoord[0].x = -largeur;
      else
	newcoord[0].x = largeur;
      if (force)
	/* on modifie les coordonnees du port proportionnellement */
	if (hauteur == oldcoord[1].y)
	  newcoord[0].y = pcoord_rel[0].y;
	else
	  {
	    newcoord[0].y = (hauteur * pcoord_rel[0].y) / oldcoord[1].y;
	    reste = (hauteur * pcoord_rel[0].y) % oldcoord[1].y;
	    if ((abs (reste * 10) / oldcoord[1].y) > 5)
	      if (reste > 0)
		newcoord[0].y++;
	      else
		newcoord[0].y--;
	  }
      else
	{
	  y_centre = tcoord[0].y + decal_cadre_y;
	  y_abs = pcoord_rel[0].y + oldcoord[0].y;
	  newcoord[0].y = y_abs - y_centre;
	}
      p = nchild (2, filsc);
      nmodifvalint (nchild (1, p), newcoord[0].x);
      nmodifvalint (nchild (2, p), newcoord[0].y);
      filsc = ntrees_getNext (filsc);
    }
}


///
void
graphicalTrees::modif_coord_btes_non_prop (ntrees arb_bte, typcoord * oldcoord, typcoord * nouvcoord)
{
  ntrees p, arb_sig, arb_gra, p_att, pfils, p_coord;
  int form, hauteur, y_centre, nbcoord;
  typcoord tcoord[2];

  p = nfather (arb_bte);
  if (p == null)
    p = arb_bte;
  arb_sig = nattrcarrier (p);
  if (noper (arb_sig) == oplfeed)
    {
      pfils = nchild (1, arb_sig);
      while (pfils != null)
	{
	  if ((p_att = desc_expr_sig (pfils)) != null)
	    if ((arb_gra = nattrntrees (p_att, GRAPHI_ATTRNUMBER)) != null)
	      {
		if (noper (arb_gra) == opschema)
		  arb_gra = nchild (1, arb_gra);
		p_coord = nchild (2, arb_gra);
		form = getCoord (p_coord, &nbcoord, tcoord, 1);
		y_centre = tcoord[0].y + abs (nouvcoord[0].y - oldcoord[0].y);
		hauteur = tcoord[1].y - abs (nouvcoord[1].y - oldcoord[1].y);
		nmodifvalint (nchild (2, nchild (1, p_coord)), y_centre);
		nmodifvalint (nchild (2, nchild (2, p_coord)), hauteur);
	      }
	  pfils = ntrees_getNext (pfils);
	}
    }
}



/**
  * verifie que les 2 ports ?port1 et ?port2 peuvent etre connectes entre eux.
  * Regles de connexion :
  *    si un des ports appartient a la boite englobante et l'autre a une boite
  * interne, les 2 ports doivent etre de meme type (c'est-a-dire entree-entree
  * ou sortie-sortie).
  *   si les 2 ports appartiennent a des boites internes, ils doivent
  * obligatoirement etre de type contraire.
  * bte_cou : boite courante englobante
  * "invers" (parametre de retour) :s'il est a vrai, indique qu'il faut inver-
  * ser l'ordre des 2 ports (les ports sont ordonnes suivant le sens du
  * flot de donnees).
  */
boolean
graphicalTrees::verif_typ_port (int port1, int port2, ntrees bte_cou, boolean * invers)
{
  int ope1, ope2;
  ntrees p1, p2;

  p1 = the_graphicTablesManager->graphObj_get_tree (port1);
  p2 = the_graphicTablesManager->graphObj_get_tree (port2);
  ope1 = noper (p1);
  ope2 = noper (p2);
  if (isPortInBox (p1, bte_cou) && isPortInBox (p2, bte_cou))
    {				/* connexion directe entre 2 ports de la boite englobante */
      *invers = (ope1 == gpkOp_ports);
      return (ope1 != ope2);
    }
  else
    {
      if (isPortInBox (p1, bte_cou))	/* 1ier port: boite englobante */
	{			/* 2ieme port: boite interne   */
	  *invers = (ope1 == gpkOp_ports);
	  return (ope1 == ope2);
	}
      else if (isPortInBox (p2, bte_cou))	/* 1ier port: boite interne */
	{			/* 2ieme port: boite englobante   */
	  *invers = (ope1 == gpkOp_porte);
	  return (ope1 == ope2);
	}
      else
	{			/* les 2 ports : boites internes */
	  *invers = (ope2 == gpkOp_ports);
	  return (ope1 != ope2);
	}
    }
}



/**
  * Rend le premier identificateur rencontre dans la liste de declaration 
  * de signaux ?l_l_list et recopie dans !decattr la declaration du
  * signal.
  * rend null si le fils n'existe pas
  * Remarque : l_l_list peut etre une liste de listes 
  * (exemple l_l_list peut etre egal a : integer a,b; real d; event e,f)
  */
ntrees
graphicalTrees::first_list_dec (ntrees l_l_list, ntrees * decattr)
{
  ntrees l_list, idsig, nomattr, typ, typattr;

  if (l_l_list != null)
    if ((l_list = nchild (1, l_l_list)) != null)
      {
	if ((l_list = nchild (2, l_list)) != null)
	  {
	    typ = ntrees_getPrevious (l_list);
	    typattr = COPY_ATTRIBUTEDSTREE (typ);
	    idsig = nchild (1, l_list);
	    nomattr = COPY_ATTRIBUTEDSTREE (idsig);
	    *decattr = LL_make_decsig_tree (typattr, LL_make_lsign_list (nomattr));
	    return (idsig);
	  }
	else
	  return (null);
      }
    else
      return (null);
  else
    return (null);
}


/**
  * Rend le frere droit de l'identificateur ?idsig dans la liste de
  * declarations de signaux 
  * et recopie dans (!decattr) l'arbre representant sa declaration
  * rend null si le frere n'existe pas
  * Cette fonction est utilisee avec la precedente.
  */
ntrees
graphicalTrees::next_list_dec (ntrees idsig, ntrees * decattr)
{
  ntrees l_list, f_idsig, nomattr, result, typ, typattr;

  result = null;
  if (idsig != null)
    {
      f_idsig = ntrees_getNext (idsig);
      if (f_idsig == null)	/* la sous-liste courante est terminee */
	{
	  if ((l_list = ntrees_getNext (nfather (nfather (idsig)))) != null)
	    {			/* la liste de declarations suivante est non vide */
	      result = idsig = nchild (1, nchild (2, l_list));
	      typ = nchild (1, l_list);
	    }
	  else
	    result = null;	/* c'est fini */
	}
      else
	{
	  typ = ntrees_getPrevious (nfather (f_idsig));
	  result = idsig = f_idsig;
	}
      if (result != null)
	{
	  typattr = COPY_ATTRIBUTEDSTREE (typ);
	  nomattr = COPY_ATTRIBUTEDSTREE (idsig);
	  *decattr = LL_make_decsig_tree (typattr, LL_make_lsign_list (nomattr));
	}
    }
  return (result);
}


/**
  Rend le nombre d'elements dans une liste de declarations de signaux 
  ?list_decl (?list_decl est une liste de listes)
*/
int
graphicalTrees::nb_elem_list_decl (ntrees list_decl)
{
  int i, nb, total;
  ntrees p;

  total = 0;
  if (list_decl != null)
    {
      nb = ntreearity (list_decl);
      for (i = 1; i <= nb; i++)
	{
	  p = nchild (i, list_decl);
	  if (p != null)
	    total = total + ntreearity (nchild (2, p));
	}
    }
  return (total);
}


/**
  Fonction récursive appelee lorsqu'on charge un fichier Signal textuel
  sous l'interface graphique.
   Elle transforme tous les modèles externes en modèle avec un corps vide.
   ?p_sig arbre Signal provenant du fichier apres analyse syntaxique
   Cette fonction rend l'arbre Signal modifie.
*/
ntrees
graphicalTrees::conv_text_gra (ntrees arb_sig)
{
  ntrees fils, l_proc;
  int nb, i;

  if (STree_isExternModel (arb_sig))
    /* cas des processus externes : on remplace leur corps */
    nchangchild (nchild (3, arb_sig), STree_mkEmptyConfinedExpr (), 2);
  else if (STree_isModel (arb_sig))
    {
      l_proc = STreeModel_getLocalDeclarations (arb_sig);
      if (l_proc != null && noper (l_proc) != opnil)
	{
	  nb = ntreearity (l_proc);
	  for (i = 1; i <= nb; i++)
	    {
	      fils = nchild (i, l_proc);
	      if (STree_isModel (fils))
		fils = conv_text_gra (fils);
	    }
	}
    }
  return (arb_sig);
}


/**
  * detruit l'attribut de type ?typ_attr accroche aux ports de la boite ?bte
  */
void
graphicalTrees::detr_attr_ports_bte (int bte, short typ_attr)
{
  ntrees arb_gra, arb_port_bte;

  arb_gra = the_graphicTablesManager->graphObj_get_tree (bte);
  if (nfather (arb_gra) != null)
    {
      arb_port_bte = nchild (1, ntrees_getNext (arb_gra));
      while (arb_port_bte != null)
	{
	  ndestroyattr (arb_port_bte, typ_attr);
	  arb_port_bte = ntrees_getNext (arb_port_bte);
	}
    }
}


/**
  * cette procedure associe les declarations se trouvant dans 
  * l'interface (?interface) aux ports d'une boite.
  * ?prem_fils : premier port de la boite
  */
void
graphicalTrees::accr_decl_ports_bte (ntrees prem_fils, ntrees interface)
{
  ntrees port, idsige, idsigs, decattre, decattrs;

  port = prem_fils;
  idsige = first_list_dec (nchild (1, interface), &decattre);
  idsigs = first_list_dec (nchild (2, interface), &decattrs);
  while (port != null)
    {
      if (noper (port) == gpkOp_porte)
	{
	  if (idsige != null)
	    {
	      nmakeattrntrees (port, SIGNAL_ATTRNUMBER, decattre);
	      idsige = next_list_dec (idsige, &decattre);
	    }
	  else
	    ndestroyattr (port, SIGNAL_ATTRNUMBER);
	}
      else
	{
	  if (idsigs != null)
	    {
	      nmakeattrntrees (port, SIGNAL_ATTRNUMBER, decattrs);
	      idsigs = next_list_dec (idsigs, &decattrs);
	    }
	  else
	    ndestroyattr (port, SIGNAL_ATTRNUMBER);
	}
      port = ntrees_getNext (port);
    }
}


/**
  Cette procedure insere dans la liste de declaration ?l_liste
  la declaration de signal ?declid si le signal n'est pas deja present.
  (?declid est forcemment un arbre d'opérateur opdecsig).
*/
void
graphicalTrees::inser_l_dec_sig (ntrees l_liste, ntrees declid)
{
  int i, j, nblp, nbp, typ, typ_list;
  boolean trouver, typexist;
  ntrees liste, sous_liste, tree_to_ins, lsign_liste, idc, p, boite, arb_typ, arb_typ2, nom_to_ins, nom_cour;

  nom_to_ins = STree_getNameOfDeclaration (declid);
  nblp = ntreearity (l_liste);
  trouver = pk_false;
  typexist = pk_false;
  typ = STreeDecl_getType (declid, &arb_typ);
  i = 1;
  if (noper (l_liste) == opnil)
    {
      i = nposit (l_liste);
      p = LL_make_ldecs_emptyList ();
      boite = nattrntrees (l_liste, GRAPHI_ATTRNUMBER);
      if (boite != null)
	nmakeattrntrees (p, GRAPHI_ATTRNUMBER, boite);
      nchangchild (nfather (l_liste), p, i);
      l_liste = p;
      trouver = pk_false;
    }
  else
    while ((i <= nblp) && !trouver)
      {
	sous_liste = nchild (i, l_liste);
	if (STree_isSignalDecl (sous_liste))	/* on ne s'occupe que des declarations de signaux */
	  {
	    typ_list = STreeDecl_getType (sous_liste, &arb_typ2);
	    liste = nchild (2, sous_liste);
	    if (typ == typ_list)
	      if (STree_isTreeEq (ntrees_getPrevious (liste), arb_typ))
		{
		  typexist = pk_true;
		  lsign_liste = liste;
		}
	    nbp = ntreearity (liste);
	    j = 1;
	    while ((j <= nbp) && !trouver)
	      {
		idc = nchild (j, liste);
		nom_cour = STree_getNameOfDeclaration (idc);
		trouver = STree_isIdentEq (nom_to_ins, nom_cour);
		j++;
	      }
	  }
	i++;
      }
  if (!trouver)
    if (typexist)
      {				/* on insere la declaration sans le type */
	tree_to_ins = STree_getDeclarationWithoutType (declid);
	lsign_liste = ntreeList_appendLast (lsign_liste, tree_to_ins);
      }
    else
      l_liste = ntreeList_appendLast (l_liste, declid);
}


/**
  Cette fonction insère dans la liste de declarations ?l_liste le signal ou
  la constante defini par ?arb_def et de type ?arb_typ, s'il n'est pas 
  deja present.
  Remarque : ?arb_def peut être un nom (pour un signal), un nom avec une
  initialisation (pour un signal), une affectation de cste (pour une cste).
*/
void
graphicalTrees::addFlowDeclaration (ntrees l_liste, ntrees arb_typ, ntrees arb_def)
{
  int i, j, nblp, nbp, typ, typ_list, ope;
  boolean trouver, typexist;
  ntrees liste, sous_liste, lsign_liste, idc, p, boite, arb_typ2, nom_cour, nom;

  nom = STree_getNameOfDeclaration (arb_def);
  ope = noper (arb_def);
  nblp = ntreearity (l_liste);
  trouver = pk_false;
  typexist = pk_false;
  typ = noper (arb_typ);
  i = 1;
  if (noper (l_liste) == opnil)
    {
      i = nposit (l_liste);
      p = LL_make_ldecs_emptyList ();
      boite = nattrntrees (l_liste, GRAPHI_ATTRNUMBER);
      if (boite != null)
	nmakeattrntrees (p, GRAPHI_ATTRNUMBER, boite);
      nchangchild (nfather (l_liste), p, i);
      l_liste = p;
    }
  else
    while ((i <= nblp) && !trouver)
      {
	sous_liste = nchild (i, l_liste);
	if ((STree_isSignalDecl (sous_liste) && (ope != opcons)) || (STree_isConstantDecl (sous_liste) && (ope == opcons)))
	  {
	    typ_list = STreeDecl_getType (sous_liste, &arb_typ2);
	    liste = nchild (2, sous_liste);
	    if (typ == typ_list)
	      if (STree_isTreeEq (ntrees_getPrevious (liste), arb_typ))
		{
		  typexist = pk_true;
		  lsign_liste = liste;
		}
	    nbp = ntreearity (liste);
	    j = 1;
	    while ((j <= nbp) && !trouver)
	      {
		idc = nchild (j, liste);
		nom_cour = STree_getNameOfDeclaration (idc);
		trouver = STree_isIdentEq (nom, nom_cour);
		j++;
	      }
	  }
	i++;
      }
  if (!trouver)
    if (typexist)
      lsign_liste = ntreeList_appendLast (lsign_liste, arb_def);
    else if (ope == opcons)
      l_liste = ntreeList_appendLast (l_liste, STree_mkConstantDeclaration (arb_def, arb_typ));
    else
      l_liste = ntreeList_appendLast (l_liste, STree_mkSignalDeclaration (arb_def, arb_typ, NILTREE));
}


/**
  Insertion des  déclarations contenues dans ?l_decl_a_inserer
  dans la liste de déclarations ?l_decl.
*/
void
graphicalTrees::mergeLocalDecl (ntrees l_decl, ntrees l_decl_a_inserer)
{
  ntrees p_typ, elem, sous_liste, l_sig;
  int i, j, nb, nb_elem, typ;

  if (ISNOT_NILTREE (l_decl_a_inserer) && (noper (l_decl_a_inserer) != opnil))
    {
      nb = ntreearity (l_decl_a_inserer);
      for (i = 1; i <= nb; i++)
	{
	  sous_liste = nchild (i, l_decl_a_inserer);
	  if (STree_isModel (sous_liste))	/* modif 27/1/99 */
	    {			/* cas d'une declaration de modele */
	      if (!isModelInList (STreeModelModule_getName (sous_liste), l_decl, &elem))
		STree_addDeclarationToList (l_decl, COPY_ATTRIBUTEDSTREE (sous_liste));
	    }
	  else if (STree_isSignalDecl (sous_liste) || STree_isConstantDecl (sous_liste))
	    {			/* cas des declarations de signaux */
	      typ = STreeDecl_getType (sous_liste, &p_typ);
	      l_sig = nchild (2, sous_liste);
	      nb_elem = ntreearity (l_sig);
	      for (j = 1; j <= nb_elem; j++)
		{
		  elem = nchild (j, l_sig);
		  addFlowDeclaration (l_decl, COPY_ATTRIBUTEDSTREE (p_typ), COPY_ATTRIBUTEDSTREE (elem));
		}
	    }
	}
    }
}


/**
  Cette procedure attribue au port (?arb_gra) la declaration de signaux
  ?arb_sig (si ?arb_sig est un arbre non vide)
  Si le nouvel arbre declaration ne contient qu'un seul signal, de type 
  implicite et de meme nom que le nom deja existant, la procedure ne 
  fait rien (evite de detruire le type); on ne fait cela uniquement si
  la declaration associee au port ne contient qu'un seul nom.
*/
void
graphicalTrees::assoc_decl_port (ntrees arb_gra, ntrees arb_sig)
{
  ntrees arb_sig_old, p_typ;
  integer ope_sig, ope_sig_old, ope_typ, nb_sig, nb_sig_old;
  boolean is_meme_nom;

  if (arb_sig != null)
    {
      arb_sig_old = nattrntrees (arb_gra, SIGNAL_ATTRNUMBER);
      if (arb_sig_old != null)
	ope_sig_old = noper (arb_sig_old);
      else
	ope_sig_old = opnil;
      if (ope_sig_old != opnil)
	{
	  if (ope_sig_old == opdecsig)
	    nb_sig_old = 1;
	  else if (ntreearity (arb_sig_old) > 1)
	    nb_sig_old = 2;
	  else
	    nb_sig_old = ntreearity (nchild (2, nchild (1, arb_sig_old)));
	  ope_sig = noper (arb_sig);
	  ope_typ = STreeDecl_getType (arb_sig, &p_typ);
	  if (ope_sig == opdecsig)
	    nb_sig = 1;
	  else if (ope_sig == opldecs)
	    if (ntreearity (arb_sig) > 1)
	      nb_sig = 2;
	    else
	      nb_sig = ntreearity (nchild (2, nchild (1, arb_sig)));
	  is_meme_nom = ((ope_typ == opimplicit) && (nb_sig == 1) && (nb_sig_old == 1) && meme_nom (arb_sig_old, arb_sig));
	  if (!is_meme_nom)
	    nmakeattrntrees (arb_gra, SIGNAL_ATTRNUMBER, arb_sig);
	}
      else
	nmakeattrntrees (arb_gra, SIGNAL_ATTRNUMBER, arb_sig);
    }
}





/**
  Rend vrai si l'objet de numero ?no appartient a la liste d'objets
  ?liste_no (actuellement cette fonction est appelee pour le "glue"
  et le "break" et la liste contient des numeros de boites).
*/
boolean
graphicalTrees::is_in_list (int no, ntrees liste_no)
{
  boolean trouve;
  ntrees fils;

  trouve = pk_false;
  fils = nchild (1, liste_no);
  while (fils != null && !trouve)
    {
      trouve = ((int) nintof (fils) == no);
      fils = ntrees_getNext (fils);
    }
  return (trouve);
}


/**
  Les 4 fonctions qui suivent concernent l'encapsulation de boites par une autre
  (bouton "glue" du menu) ou la fonction inverse (bouton "break")

  Rend vrai si, a partir du port diffuseur ?arb_port_diff, on peut atteindre
  un port appartenant a une boite qui n'est pas dans la liste des 
  boites encapsulees ?l_bte_enc. Si le port appartient deja a une boite qui 
  n'est pas  dans ?l_bte_enc, rend vrai.
  ?bte : le numero de la boite qui encapsule.
*/
boolean
graphicalTrees::existe_port_lie_non_encaps (int bte, ntrees arb_port_diff, ntrees l_bte_enc)
{
  ntrees l_no_port_lie, arb_no_port_lie, arb_port_lie, ch;
  int no_bte_lie;
  boolean trouve, intern, b;

  trouve = pk_false;
  no_bte_lie = no_boite_mere (arb_port_diff);
  if (!is_in_list (no_bte_lie, l_bte_enc))
    return (pk_true);
  else
    intern = isExternPort (the_graphicTablesManager->graphObj_get_tree (bte), arb_port_diff);
  l_no_port_lie = get_ports_lies (tree_get_graphObj (arb_port_diff), intern, &ch, &b);
  arb_no_port_lie = nchild (1, l_no_port_lie);
  while (arb_no_port_lie != null && !trouve)
    {
      arb_port_lie = the_graphicTablesManager->graphObj_get_tree ((int) nintof (arb_no_port_lie));
      no_bte_lie = no_boite_mere (arb_port_lie);
      if (!is_in_list (no_bte_lie, l_bte_enc))
	return (pk_true);
      else
	arb_no_port_lie = ntrees_getNext (arb_no_port_lie);
    }
  return (trouve);
}


/**
  Cette procedure tente d'inserer un port d'ordonnee (?!y_p) sur le
  bord gauche ou droit (?sens) de la boite (?bte) d'ordonnees (?b_inf)
  et (?b_sup) : on tente l'insertion avec y_p puis s'il y a echec, on 
  boucle sur l'insertion avec y_p modifiee situee dans l'intervalle 
  representant le trou le plus proche.
  (!y_p) l'ordonnee modifiee
  (!rang)  le rang du port a inserer
*/
boolean
graphicalTrees::place_trou_port (int bte, int sens, int b_inf, int b_sup, int *y_p, int *rang)
{
  ntrees arb_bte, l_trous, b_trou_p, b_trou_i;
  int m_trou_p, m_trou_i, d_p, d_i, ind_p, ind_i, y1, y2, d_y1, d_y2, ajust;
  boolean place;

  arb_bte = the_graphicTablesManager->graphObj_get_tree (bte);
  if (ISNOT_NILTREE (nfather (arb_bte)))
    {
      place = insert_port (arb_bte, 0, sens, y_p, b_sup, b_inf, rang);
      if (!place)
	{
	  l_trous = LL_make_ord_emptyList ();
	  calcul_trous_port (b_inf, b_sup, l_trous, sens, arb_bte);
	}
      while (!place && !nisemptyList (l_trous))
	{
	  b_trou_p = nchild (1, l_trous);
	  y1 = (int) nintof (nchild (1, b_trou_p));
	  y2 = (int) nintof (nchild (2, b_trou_p));
	  m_trou_p = (y2 - y1) / 2;
	  d_p = m_trou_p - (*y_p);
	  ind_p = 1;
	  ind_i = 1;
	  b_trou_i = ntrees_getNext (b_trou_p);
	  while (b_trou_i != null)
	    {
	      ind_i++;
	      y1 = (int) nintof (nchild (1, b_trou_i));
	      y2 = (int) nintof (nchild (2, b_trou_i));
	      m_trou_i = (y2 - y1) / 2;
	      d_i = m_trou_i - (*y_p);
	      if (abs (d_i) <= abs (d_p))
		{
		  d_p = d_i;
		  ind_p = ind_i;
		}
	      b_trou_i = ntrees_getNext (b_trou_i);
	    }
	  b_trou_p = nchild (ind_p, l_trous);
	  y1 = (int) nintof (nchild (1, b_trou_p));
	  y2 = (int) nintof (nchild (2, b_trou_p));
	  d_y1 = y1 - (*y_p);
	  d_y2 = y2 - (*y_p);
	  if (abs (d_y1) < abs (d_y2))
	    {
	      if (y1 == b_inf)
		ajust = taillepave_y;
	      else
		ajust = inter_pave;
	      *y_p = y1 + ajust;
	    }
	  else
	    {
	      if (y2 == b_sup)
		ajust = taillepave_y;
	      else
		ajust = inter_pave;
	      *y_p = y2 - ajust;
	    }
	  place = insert_port (arb_bte, 0, sens, y_p, b_sup, b_inf, rang);
	  ntreesDestroy (b_trou_p);
	}
    }
  else
    {
      if (*y_p < b_inf + taillepave_y)
	*y_p = b_inf + taillepave_y;
      else if (*y_p > b_sup - taillepave_y)
	*y_p = b_sup - taillepave_y;
      *rang = 0;
      place = pk_true;
    }
  return (place);
}


/**
  * enleve le numero de port contenu dans la liste (?l_ports_lies)
  * dont le numero de boite mere ne figure pas dans la liste
  * (?l_bte_dec) des boites internes
  */
void
graphicalTrees::oter_port_lie_non_decaps (ntrees l_ports_lies, ntrees l_bte_dec)
{
  ntrees arb_no_port_lie;
  int bte_lie;

  arb_no_port_lie = nchild (1, l_ports_lies);
  while (arb_no_port_lie != null)
    {
      bte_lie = tree_get_graphObj (boite_mere ((int) nintof (arb_no_port_lie)));
      if (!is_in_list (bte_lie, l_bte_dec))
	{
	  ntreesDestroy (arb_no_port_lie);
	  arb_no_port_lie = nchild (1, l_ports_lies);
	}
      else
	arb_no_port_lie = ntrees_getNext (arb_no_port_lie);
    }
}


/**
  * enleve le numero de port contenu dans la liste (?l_ports_lies)
  * dont le numero de boite mere ne figure pas dans la liste
  * (?l_bte_dec) des boites internes
  */
void
graphicalTrees::oter_port_lie_bte (int bte, ntrees l_ports_lies)
{
  ntrees arb_no_port_lie;
  int bte_lie;
  boolean trouver;

  trouver = pk_false;
  arb_no_port_lie = nchild (1, l_ports_lies);
  while (arb_no_port_lie != null && trouver == pk_false)
    {
      bte_lie = tree_get_graphObj (boite_mere ((int) nintof (arb_no_port_lie)));
      if (bte_lie == bte)
	{
	  ntreesDestroy (arb_no_port_lie);
	  trouver = pk_true;
	}
      else
	arb_no_port_lie = ntrees_getNext (arb_no_port_lie);
    }
}


/**
  * rend vrai si la boite de numero (?bte) a au moins une 
  * boite interne.  
  * Rend la liste des numeros des boites internes (?l_bte_enc)
  */
boolean
graphicalTrees::une_boite_decaps (int bte, ntrees * l_bte_enc)
{
  ntrees arb_gra, arb_sig;
  int no_bte;

  arb_gra = the_graphicTablesManager->graphObj_get_tree (bte);
  arb_sig = get_arb_sig (&arb_gra);
  arb_sig = desc_expr_sig (nchild (1, arb_sig));
  if (arb_sig != null)
    {
      arb_gra = nattrntrees (arb_sig, GRAPHI_ATTRNUMBER);
      if (arb_gra == null)
	return (pk_false);
      else
	{
	  *l_bte_enc = LL_make_design_emptyList ();
	  while (arb_gra != null && arb_sig != null)
	    {
	      no_bte = tree_get_graphObj (arb_gra);
	      ntreeList_insertFirst (LL_make_design_atom (no_bte), *l_bte_enc);
	      arb_sig = ntrees_getNext (arb_sig);
	      if (arb_sig != null)
		arb_gra = nattrntrees (arb_sig, GRAPHI_ATTRNUMBER);
	    }
	}
      return (pk_true);
    }
  else
    return (pk_false);
}


/**
  Fonction utilisee pour la repartion de programmes (l'utilisateur
  a clique sur le bouton "SET ARCHI"). Elle est appelee par la fonction 
  "traiter_repartition".
  Cette fonction rajoute un port a une boite correspondant à l'appel du 
  processus ?p_appel, soit en entree, soit en sortie suivant ?en_entree et 
  lui associe un nom et un type ?p_decl. 
  Ce port correspond a un signal d'interface cree par le compilo;
*/
void
graphicalTrees::add_signal_proc_appe (ntrees p_appel, ntrees p_decl, boolean en_entree)
{
  ntrees p_boite;
  int no_boite, no_port, nbcoord, y, rg, b_inf, b_sup, sens;
  typcoord coord[2], pcoord;

  p_boite = nattrntrees (p_appel, GRAPHI_ATTRNUMBER);
  if (ISNOT_NILTREE (p_boite))	/* blindage, en principe on a forcemment une boite */
    {
      if (noper (p_boite) == opschema)
	p_boite = nchild (1, p_boite);
      getCoord (p_boite, &nbcoord, coord, 0);
      y = coord[0].y - coord[1].y + taillepave_y;
      if (en_entree)
	{
	  sens = gauche;
	  pcoord.x = coord[0].x - coord[1].x;
	}
      else
	{
	  sens = droite;
	  pcoord.x = coord[0].x + coord[1].x;
	}
      b_inf = coord[0].y - coord[1].y;
      b_sup = coord[0].y + coord[1].y;
      no_boite = tree_get_graphObj (p_boite);
      if (place_trou_port (no_boite, sens, b_inf, b_sup, &y, &rg))
	pcoord.y = y;
      else			// s'il n'y a plus de place, on crée le port en bas de la boite
	{
	  pcoord.y = coord[0].y + coord[1].y - (taillepave_y / 2);
	  if (en_entree)
	    rg = getNbLeftPorts (no_boite);
	  else
	    rg = getNbRightPorts (no_boite);
	}
    }
  no_port = creatept_connex (no_boite, pcoord, en_entree, rg);
  assoc_decl_port (the_graphicTablesManager->graphObj_get_tree (no_port), p_decl);
}


/**
   CONSTRUCTION DE LA TABLE D'ACCES AUX OBJETS GRAPHIQUES lorsqu'on charge
   un fichier contenant un programme graphique 
   OU MODIFICATION DE LA TABLE D'ACCES AUX OBJETS GRAPHIQUES 
      soit, apres l'insertion d'un fichier contenant un arbre graphique
      soit apres une copie d'un sous-arbre  (?undo a faux)
      soit, pour revenir a un etat anterieur (?undo est a vrai et ?p
            est l'arbre Signal sauvegarde).
      pour avoir unicite des numeros de tous les objets graphiques.
   ?correction : a vrai, on applique un facteur de correction sur les
                 coordonnees des objets de ?f_x sur les abscisses
                 et de ?f_y sur les ordonnees. (ut<ilise en cas de 
                 changement de version de l'editeur).
   Les numeros des objets graphiques sont modifies de telle facon qu'ils
   soient toujours uniques.
*/
void
graphicalTrees::reconstruire_tab (ntrees gpkSourceModeul, boolean undo, boolean correction, float f_x, float f_y)
{
  // nDEBUG_subtree(4, 10, "entering graphicalTrees::reconstruire_tab srcTree is", srcTree);
  the_graphicTablesManager->init_convnum ();
  innerRebuild_gpkTree (gpkSourceModeul, undo, correction, f_x, f_y);
  // nDEBUG_subtree(4, 10, " exiting graphicalTrees::reconstruire_tab srcTree is", srcTree);
}


///
void
graphicalTrees::ajout_proc_hierar (ntrees p)
{

}



/// called in undo::set_undo_archi
void
graphicalTrees::replaceInHierar (ntrees model)
{
  C2Java_deleteSelectedNode ();	// on détruit le noeud sélectionné et on remonte dans le JTree
  add_modelInJtree (model, pk_true);
}
