/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
#define COPY_ATTRIBUTEDSTREE(ATREE) ncopytree(ATREE, pk_true)

_PUBLIC_
/**
  Cette procedure annule ou corrige le halo du ?chemin
  Cette fonction est appelee par la fonction  annuler_halo_seg_ports
*/
  void
graphicalTrees::annuler_halo_chemins (ntrees chemin, boolean corrige)
{
  ntrees l_chem, lseg, filsc;
  int nbchem, nbseg, no_seg;
  typcoord halo[2];

  if ((l_chem = nchild (3, chemin)) != null)
    {
      nbchem = ntreearity (l_chem);
      if (nbchem > 0)
	filsc = nchild (1, l_chem);
      else
	filsc = null;
      while (filsc != null)
	{
	  annuler_halo_chemins (filsc, corrige);
	  filsc = ntrees_getNext (filsc);
	}
    }
  lseg = nchild (4, chemin);
  nbseg = ntreearity (lseg);
  if (nbseg > 0)
    filsc = nchild (1, lseg);
  else
    filsc = null;
  while (filsc != null)
    {
      no_seg = tree_get_graphObj (filsc);
      if (corrige)
	{
	  calcul_halo (filsc, halo);
	  the_graphicTablesManager->setSegHalo (no_seg, halo);
	}
      else
	the_graphicTablesManager->annuler_halo_seg (no_seg);
      filsc = ntrees_getNext (filsc);
    }
}


/**
  Method called only by innerRebuild_gpkTree to modify the coordinates of the graphical objects
*/
void
graphicalTrees::changCoord (ntrees pObj, float fcor_x, float fcor_y)
{
  ntrees p1, p2, pcoord, chemin, sous_chem;
  int i, j, nbcoord, ope;
  typcoord tcoord[2];

  pcoord = nchild (2, pObj);
  getCoord (pcoord, &nbcoord, tcoord, 1);
  ope = noper (pObj);
  switch (ope)
    {
    case gpkOp_bte:
      p1 = nchild (1, pcoord);
      nmodifvalint (nchild (1, p1), (int) (tcoord[0].x * fcor_x));
      nmodifvalint (nchild (2, p1), (int) (tcoord[0].y * fcor_y));
      p2 = nchild (2, pcoord);
      nmodifvalint (nchild (1, p2), (int) (tcoord[1].x * fcor_x));
      nmodifvalint (nchild (2, p2), (int) (tcoord[1].y * fcor_y));
      break;
    case gpkOp_porte:		// input or output port
    case gpkOp_ports:
      p1 = nchild (1, pcoord);
      nmodifvalint (p1, (int) (tcoord[0].x * fcor_x));
      p2 = nchild (2, pcoord);
      nmodifvalint (p2, (int) (tcoord[0].y * fcor_y));
      break;
    case opseg:
      p1 = nchild (1, pcoord);
      p2 = nchild (2, pcoord);
      i = (int) (tcoord[0].x * fcor_x);
      j = (int) (tcoord[1].x * fcor_x);
      if (premiertour)
	{
	  if (i > j)
	    i = i + 2;
	  else
	    i = i - 2;
	  premiertour = pk_false;
	};
      chemin = get_chemin (pObj);
      sous_chem = nchild (2, chemin);
      if ((ntrees_getNext (pObj) == null) && (sous_chem == null))
	if (j > i)
	  j = j + 2;
	else
	  j = j - 2;
      nmodifvalint (nchild (1, p1), i);
      nmodifvalint (nchild (1, p2), j);
      nmodifvalint (nchild (2, p1), (int) (tcoord[0].y * fcor_y));
      nmodifvalint (nchild (2, p2), (int) (tcoord[1].y * fcor_y));
      break;
    }
}



/**
   Cette procedure parcourt le bord gauche ou droite (?sens) de la 
   boite d'ordonnees ?b_inf et ?b_sup et insere dans une liste (?l_trou)
   les arbres binaires representants les ordonnees d'un espace de taille
   suffisement grande pour inserer un port.
   (?arb_l_ports) la liste des ports
*/
void
graphicalTrees::calcul_trous_port (int b_inf, int b_sup, ntrees l_trous, int sens, ntrees arb_bte)
{
  int y1, y2, ajust, nbcoord1, nbcoord2;
  typcoord tcoord[maxcoord], tcoord_rel[maxcoord];
  ntrees arb_port, b_trou;

  y1 = b_inf;
  y2 = b_sup;
  arb_port = getFirstPort (arb_bte);
  getCoord (arb_port, &nbcoord2, tcoord_rel, 1);
  if (sens == gauche)
    {
      while (tcoord_rel[0].x < 0 && arb_port != null)
	{
	  getCoord (arb_port, &nbcoord1, tcoord, 0);
	  y2 = tcoord[0].y;
	  if (y1 == b_inf || y2 == b_sup)
	    ajust = taillepave_y;
	  else
	    ajust = inter_pave;

	  if (y2 - y1 >= (inter_pave + ajust))
	    {
	      b_trou = LL_make_ordBin (LL_make_ord_atom (y1), LL_make_ord_atom (y2));
	      ntreeList_insertFirst (b_trou, l_trous);
	    }
	  y1 = y2;
	  arb_port = ntrees_getNext (arb_port);
	  if (arb_port != null)
	    getCoord (arb_port, &nbcoord2, tcoord_rel, 1);
	}
      y2 = b_sup;
      if (y2 - y1 >= (inter_pave + taillepave_y))
	{
	  b_trou = LL_make_ordBin (LL_make_ord_atom (y1), LL_make_ord_atom (y2));
	  ntreeList_insertFirst (b_trou, l_trous);
	}
    }
  else				/* sens = droite */
    {
      while (arb_port != null && tcoord_rel[0].x < 0)
	{
	  arb_port = ntrees_getNext (arb_port);
	  if (arb_port != null)
	    getCoord (arb_port, &nbcoord2, tcoord_rel, 1);
	}
      while (arb_port != null)
	{
	  getCoord (arb_port, &nbcoord1, tcoord, 0);
	  y2 = tcoord[0].y;
	  if (y1 == b_inf || y2 == b_sup)
	    ajust = taillepave_y;
	  else
	    ajust = inter_pave;

	  if (y2 - y1 >= (inter_pave + ajust))
	    {
	      b_trou = LL_make_ordBin (LL_make_ord_atom (y1), LL_make_ord_atom (y2));
	      ntreeList_insertFirst (b_trou, l_trous);
	    }
	  y1 = y2;
	  arb_port = ntrees_getNext (arb_port);
	}
      y2 = b_sup;
      if (y2 - y1 >= (inter_pave + taillepave_y))
	{
	  b_trou = LL_make_ordBin (LL_make_ord_atom (y1), LL_make_ord_atom (y2));
	  ntreeList_insertFirst (b_trou, l_trous);
	}
    }
}


///
int
graphicalTrees::setToBoxTable (ntrees pGra, boolean undo, boolean correction, float f_x, float f_y)
{
  ntrees p_design;
  int n;
  typcoord halo[2];

  p_design = nchild (1, pGra);
  calcul_halo (pGra, halo);
  if (!undo)
    {
      if (noper (pGra) == gpkOp_bte)
	n = the_graphicTablesManager->convnumbte (p_design);
      else
	n = the_graphicTablesManager->convnumport (p_design);
      if (correction)
	changCoord (pGra, f_x, f_y);
      the_graphicTablesManager->inser_obj (n, pGra, halo);
    }
  else
    {
      n = (int) nintof (p_design);
      the_graphicTablesManager->enlever_obj (n);
      the_graphicTablesManager->inser_obj (n, pGra, halo);
    }
  return (n);
}


/**
  Reconstruit la table d'acces aux segments et la table d'acces aux liens 
  pour le chemin ?p_chemin qui part du port de numero ? noport.
  ?undo a vrai on modifie uniquement dans la table les adresses des
  objets de l'arbre par celles des objets de l'arbre sauvegarde 
  Cette procedure est appelee par innerRebuild_gpkTree et par 
  set_undo_arb_chemin.
  Si ?correction est a vrai, on modifie les positions grace au facteur de 
  correction ?fcor_x et fcor_y (utilise uniquement pour un changement de version
*/
void
graphicalTrees::reconstruire_tabs_chem (ntrees p_chemin, int noport, boolean undo, boolean correction, float fcor_x, float fcor_y)
{
  int form, n, nbcoord, sens;
  typcoord tcoord[maxcoord], halo[2];
  ntrees p, p_design, sous_chem;

  p = nchild (1, nchild (4, p_chemin));	// first segment of the link
  sous_chem = nchild (2, p_chemin);
  while (p != null)
    {
      form = getCoord (nchild (2, p), &nbcoord, tcoord, 1);
      if (tcoord[0].x == tcoord[1].x)
	sens = vertic;
      else
	sens = horiz;
      p_design = nchild (1, p);
      calcul_halo (p, halo);
      if (undo)
	{
	  n = (int) nintof (p_design);
	  the_graphicTablesManager->enlever_seg (n);
	  the_graphicTablesManager->inser_seg (n, sens, noport, p, halo);
	}
      else
	{
	  n = the_graphicTablesManager->convnumseg (p_design);
	  if (correction)
	    changCoord (p, fcor_x, fcor_y);
	  the_graphicTablesManager->inser_seg (n, sens, noport, p, halo);
	}
      p = ntrees_getNext (p);
    }
  p = sous_chem;
  if (p == null)
    {
      p_chemin = nchild (1, nchild (3, p_chemin));
      p_design = nchild (1, p_chemin);
      if (!undo)
	n = the_graphicTablesManager->convnumchem (p_design);
      while (p_chemin != null)
	{
	  reconstruire_tabs_chem (p_chemin, noport, undo, correction, fcor_x, fcor_y);
	  p_chemin = ntrees_getNext (p_chemin);
	}
    }
  else
    {
      p_design = nchild (1, p);
      calcul_halo (p_chemin, halo);
      if (undo)
	{
	  n = (int) nintof (p_design);
	  the_graphicTablesManager->enlever_obj (n);
	  the_graphicTablesManager->inser_obj (n, p_chemin, halo);
	}
      else
	{
	  n = the_graphicTablesManager->convnumlien (p_design);
	  the_graphicTablesManager->inser_obj (n, p_chemin, halo);
	  p_design = nchild (2, p);
	  n = the_graphicTablesManager->convnumport (p_design);
	}
    }
}

///
void
graphicalTrees::buildGraphicalTables (ntrees pGra, boolean undo, boolean correction, float fcor_x, float fcor_y)
{
  ntrees pBox, pPort, p_chemin, p_lien;
  int n, noPort;

  if (noper (pGra) == opschema)
    {
      pBox = nchild (1, pGra);
      pPort = getFirstPort (pGra);
    }
  else
    {
      pBox = pGra;
      SET_NILTREE (pPort);
    };
  if (pBox != null)
    setToBoxTable (pBox, undo, correction, fcor_x, fcor_y);
  while (pPort != null)
    {
      noPort = setToBoxTable (pPort, undo, correction, fcor_x, fcor_y);
      p_lien = nchild (4, pPort);
      if (p_lien != null && !undo)
	n = the_graphicTablesManager->convnumlien (p_lien);
      p_chemin = nchild (3, pPort);
      if (p_chemin != null)
	{
	  premiertour = pk_true;
	  reconstruire_tabs_chem (p_chemin, noPort, undo, correction, fcor_x, fcor_y);
	  premiertour = pk_true;
	};
      pPort = ntrees_getNext (pPort);
    }
}

/**
*/
void
graphicalTrees::innerRebuild_gpkTree (ntrees gpkTree, boolean undo, boolean correction, float fcor_x, float fcor_y)
{
  ntrees p_fils, graphTree;

  if ((graphTree = nattrntrees (gpkTree, GRAPHI_ATTRNUMBER)) != null)
    buildGraphicalTables (graphTree, undo, correction, fcor_x, fcor_y);
  if (STree_isModel (gpkTree))
    {
      if ((p_fils = STreeModel_getBody (gpkTree)) != null)
	innerRebuild_gpkTree (p_fils, undo, correction, fcor_x, fcor_y);
      if ((p_fils = STreeModel_getLocalDeclarations (gpkTree)) != null)
	innerRebuild_gpkTree (p_fils, undo, correction, fcor_x, fcor_y);
      return;
    }
  if (STree_isModule (gpkTree))
    {
      if ((p_fils = STreeModel_getLocalDeclarations (gpkTree)) != null)
	innerRebuild_gpkTree (p_fils, undo, correction, fcor_x, fcor_y);
    }
  switch (noper (gpkTree))
    {
    case oplfeed:
      gpkTree = nchild (1, gpkTree);
      while (gpkTree != null)
	{
	  if (nattrntrees (gpkTree, GRAPHI_ATTRNUMBER) != null)
	    innerRebuild_gpkTree (gpkTree, undo, correction, fcor_x, fcor_y);
	  gpkTree = ntrees_getNext (gpkTree);
	}
      break;
    case opldecl:
      gpkTree = nchild (1, gpkTree);
      while (gpkTree != null)
	{
	  if (STree_isModel (gpkTree))
	    innerRebuild_gpkTree (gpkTree, undo, correction, fcor_x, fcor_y);
	  gpkTree = ntrees_getNext (gpkTree);
	};
      break;
    }
}

/**
  Procedure recursive appelee par la fonction create_modelInJTree pour construire
  la hierarchie du processus ou du module ?p apres le chargement d'un fichier.
  Le processus est inséré au rang ?rang au lieu d'être inséré en queue de liste.
*/
void
graphicalTrees::create_modelJTree (ntrees p, boolean setInAtEnd)
{
  ntrees l_proc, proc;
  int i, nb;
  ptrepstr strnom;

  // nDEBUG_subtree(4, 2,"entering graphicalTrees::create_modelJTree p is", p);
  strnom = STreeModelModule_getNameStr (p);
  i = no_boite_proc (p);
  if (i <= 0)
    {
      printf ("## INTERNAL ERROR ## in graphicalTrees::create_modelJTree, no_boite_proc is negative, ignored\n");
      fflush (stdout);
      return;
    }
  if (strnom == nilstr)
    {
      C2Java_appendNewJNode ("", i, setInAtEnd);
    }				// java world
  else
    {
      C2Java_appendNewJNode (repstr2string (strnom), i, setInAtEnd);
    }				// java world
  l_proc = STreeModel_getLocalDeclarations (p);
  l_proc = STree_removeNilTree (l_proc);	// cleaning tree
  nb = ntreearity (l_proc);
  i = 1;
  while (i <= nb)
    {
      proc = nchild (i, l_proc);
      if (STree_isModel (proc))
	create_modelJTree (proc, pk_true);
      i++;
    }
  C2Java_upTempSelectedNode ();	// java world
  // nDEBUG_subtree(4, 2," exiting graphicalTrees::create_modelJTree p is", p);
}

///
void
graphicalTrees::add_modelInJtree (ntrees p, boolean setInAtEnd)
{
  int no_boite;
  // printf("DEBUG entering graphicalTrees::add_modelInJtree\n"); fflush(stdout);
  create_modelJTree (p, setInAtEnd);
  no_boite = no_boite_proc (p);
  C2Java_SetCurrentTree (no_boite);	// java world
  // printf("DEBUG  exiting graphicalTrees::add_modelInJtree\n"); fflush(stdout);
}
