/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
///
_PUBLIC_
  /*
     Creation d'un lien :
     Cette procedure est appelee autant de fois qu'un point est clique
     jusqu'a arriver sur un port ou le bord d'une boite.
     ?(x_reticule,y_reticule) point clique par l'utilisateur.
   */
  void
builder::cr_lien (int x_ret, int y_ret)
{
  int i, x, trait, pt_neg, pt_posi, y, noseg, direc, ope, rg, sens_arr;
  typcoord pcoord[2], halo[2];
  objet_type typ_obj2;
  ntrees arb_gra;
  boolean invers, ok, b, connexion;

  ok = pk_true;
  if (premiertour)
    {
      pt_dep = pCurObjManager->reper_obj (x_ret, y_ret, &typ_obj1);
      if (pt_dep != 0 && (typ_obj1 == typ_port || typ_obj1 == typ_port_ext))
	{
	  x = x_ret;
	  y = y_ret;
	  if (typ_obj1 == typ_port)
	    i = pCurObjManager->reper_bte (&x, &y, &sens_dep, &rg, 0, &b);
	  /* on appelle pCurObjManager->reper_bte uniquement pour savoir si le port est a gauche ou
	     a droite de la boite (pour le routage automatique) dans le cas d'une
	     boite interne */
	  else if (x_ret < abs_mil_cadre)
	    sens_dep = droite;
	  else
	    sens_dep = gauche;	/* boite externe */
	  arb_gra = the_graphicTablesManager->graphObj_get_tree (pt_dep);
	  ope = noper (arb_gra);
	  if (typ_obj1 == typ_port_ext)
	    connexion = !((ope == gpkOp_ports) && (nchild (4, arb_gra) != null));
	  else
	    connexion = !((ope == gpkOp_porte) && (nchild (4, arb_gra) != null));
	  if (connexion)
	    {
	      depart = pk_true;
	      premiertour = pk_false;
	      get_coord_seg_port (arb_gra, typ_obj1, &x_pred, &y_pred);
	      tcoord[0].x = x_pred;
	      tcoord[0].y = y_pred;
	      nbcoord = 1;
	      sens = horiz;
	    }
	  else
	    {
	      ok = pk_false;
	      the_messageTool->error_mess (crlienimppc);
	    }
	}
    }
  else
    {
      pt_arr = pCurObjManager->reper_obj (x_ret, y_ret, &typ_obj2);
      if (pt_arr == 0 || typ_obj2 == typ_seg)
	/* cas ou on clique un point qui n'est pas un port ou une boite */
	dessiner_seg_intermediaire (x_ret, y_ret);
      else
	{			/* cas ou on a clique sur un port ou une boite */
	  if (typ_obj2 == typ_bte || typ_obj2 == typ_cadre)
	    {
	      if (nbcoord == 1)	/*on essaie d'aligner uniquement en routage automatique */
		if (abs (y_ret - tcoord[0].y) <= 2 * halo_segm)
		  y_ret = tcoord[0].y;
	      ok = createport_auto (x_ret, y_ret, &typ_obj2);
	    }
	  if (ok)
	    ok = the_graphicalTrees->verif_typ_port (pt_dep, pt_arr, pCurObjManager->get_bte_cou_engl (), &invers);
	  if (!ok)
	    the_messageTool->error_mess (portincomp);
	  else
	    {
	      x = x_ret;
	      y = y_ret;
	      if (typ_obj2 == typ_port)
		i = pCurObjManager->reper_bte (&x, &y, &sens_arr, &rg, 0, &b);
	      /* on appelle reper_bte pour savoir si le port est a gauche ou
	         a droite de la boite (pour le routage automatique) dans le cas d'une
	         boite interne */
	      else if (x_ret < abs_mil_cadre)
		sens_arr = droite;
	      else
		sens_arr = gauche;	/* boite externe */
	      arb_gra = the_graphicTablesManager->graphObj_get_tree (pt_arr);
	      get_coord_seg_port (arb_gra, typ_obj2, &x_ret, &y_ret);
	      if (!invers)
		{
		  pt_neg = pt_dep;
		  pt_posi = pt_arr;
		}
	      else
		{
		  pt_neg = pt_arr;
		  pt_posi = pt_dep;
		};
	      if (pt_posi == pt_arr && nchild (4, arb_gra) != null)
		{
		  ok = pk_false;
		  the_messageTool->error_mess (crlienimppc);
		}
	      else
		{
		  arb_gra = the_graphicTablesManager->graphObj_get_tree (pt_neg);
		  if ((nchild (3, arb_gra) == null))
		    {		/* le port de sortie n'est pas deja connecte : lien simple */
		      if (nbcoord == 1)	/* routage automatique */
			{
			  ok = lancer_routage (invers, x_ret, y_ret, sens_arr, x_pred, y_pred, sens_dep, tcoord, &nbcoord, 0, &noseg);
			  if (ok)
			    {
			      pDraw->dessiner_lien (nbcoord + 1, tcoord, pk_false, COL_DEF);
			      createchemin (pt_neg, pt_posi, nbcoord, tcoord);
			    }
			  else
			    the_messageTool->error_mess (routimpptint);
			}
		      else	/* routage manuel */
			{
			  dessiner_dern_seg (x_ret, y_ret);
			  if (invers)
			    inverser (tcoord, nbcoord);
			  i = 0;
			  while (i < nbcoord && ok)
			    {
			      pcoord[0].x = tcoord[i].x;
			      pcoord[1].x = tcoord[i + 1].x;
			      pcoord[0].y = tcoord[i].y;
			      pcoord[1].y = tcoord[i + 1].y;
			      if (tcoord[i].y == tcoord[i + 1].y)
				direc = horiz;
			      else
				direc = vertic;
			      the_graphicalTrees->calcul_nouv_halo (typ_seg, pcoord, halo);
			      ok = pCurObjManager->test_cr_seg (halo, direc);
			      i++;
			    }
			  if (ok)
			    createchemin (pt_neg, pt_posi, nbcoord, tcoord);
			  else
			    the_messageTool->error_mess (crlienimp);
			}
		    }
		  else		/*  cas ou il y a recouvrement */
		    {
		      if (nbcoord != 1)	/* routage manuel */
			{
			  if (x_pred == tcoord[nbcoord - 2].x)
			    tcoord[nbcoord - 1].y = y_ret;
			  tcoord[nbcoord].x = x_ret;
			  tcoord[nbcoord].y = y_ret;
			  if (pt_arr == pt_posi)
			    inverser (tcoord, nbcoord);
			  if (tcoord[nbcoord - 2].x == tcoord[nbcoord - 1].x)
			    trait = vertic;
			  else
			    trait = horiz;
			  ok = recouv_lien (pt_neg, pt_posi, &nbcoord, tcoord, trait, pk_false);
			  if (!ok)
			    the_messageTool->error_mess (crlienimp);
			}
		      else
			{	/* routage automatique */
			  if (lancer_routage (invers, x_ret, y_ret, sens_arr, x_pred, y_pred, sens_dep, tcoord, &nbcoord, pt_neg, &noseg))
			    {
			      inverser (tcoord, nbcoord);
			      if (tcoord[nbcoord].x == tcoord[nbcoord - 1].x)
				trait = vertic;
			      else
				trait = horiz;
			      nbcoord++;
			      ok = recouv_lien (pt_neg, pt_posi, &nbcoord, tcoord, trait, pk_true);
			      if (!ok)
				the_messageTool->error_mess (crlienimp);
			    }
			  else
			    {
			      ok = pk_false;
			      the_messageTool->error_mess (routimpptint);
			    }
			}
		    }
		}
	    }
	  nettoyage ();
	  premiertour = pk_true;
	  if (ok)
	    C2Java_refresh (pk_true);
	}
    }
}
