/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
///
_PRIVATE_
/*
  Cette procedure calcule les coordonnees du segment qui arrive
   ou qui part d'un port, en fonction du type du port (port
   externe ou non) et en fonction de son placement par rapport a la
   boite (a gauche ou a droite).
   ?arb_gra : sous-arbre representant le port
      ?typ_p : type du port (normal ou externe).
      !xseg,!yseg : coordonnees calculees
*/
  void
builder::get_coord_seg_port (ntrees arb_port, int typ_p, int *xseg, int *yseg)
{
  typcoord pcoord[10], pcoord_r[10];
  int npt, form;
  ntrees p_bte;
  float f_ech_x, f_ech_y;

  p_bte = the_graphicalTrees->boite_mere (the_graphicalTrees->tree_get_graphObj (arb_port));
  calcul_fact_ech (p_bte, &f_ech_x, &f_ech_y);
  form = the_graphicalTrees->getCoord (arb_port, &npt, pcoord_r, 1);
  if (typ_p == typ_port_ext)
    {
      pcoord[0].x = (int) (pcoord_r[0].x * f_ech_x) + abs_mil_cadre;
      pcoord[0].y = (int) (pcoord_r[0].y * f_ech_y) + ord_mil_cadre;
    }
  else
    form = the_graphicalTrees->getCoord (arb_port, &npt, pcoord, 0);
  if ((pcoord_r[0].x < 0 && typ_p == typ_port) || (pcoord_r[0].x > 0 && typ_p == typ_port_ext))
    *xseg = pcoord[0].x - taillepave_x - 1;
  else
    *xseg = pcoord[0].x + taillepave_x + 1;
  *yseg = pcoord[0].y;
}


/**
   Creation de l'arbre associe a un chemin et insertion dans la table d'acces
   pt_neg : port diffuseur
   pt_posi: port recepteur
   ?nbcoord : nombre de segments composant le chemin.
   ?tcoord : leurs coordonnees (attention, il faut nbcoord+1 coordonnees).
*/
void
builder::createchemin (int pt_neg, int pt_posi, int nbcoord, typcoord * tcoord)
{
  ntrees arb_gra, s1, s2, s3, s4, chemin;
  graphObj_type ch_design, li_design;
  typcoord halo[2];

  ch_design = the_graphicTablesManager->newGraphObjIdentity ();
  li_design = the_graphicTablesManager->newGraphObjIdentity ();
  arb_gra = the_graphicTablesManager->graphObj_get_tree (pt_neg);
  s1 = LL_make_design_atom (ch_design);
  s2 = LL_make_extm (LL_make_lien_atom (li_design), LL_make_port_atom (pt_posi));
  s3 = LL_make_chemins_emptyList ();
  s4 = createListSeg (tcoord, nbcoord, pt_neg);
  chemin = LL_make_chemin (s1, s2, s3, s4);
  nchangchild (arb_gra, chemin, 3);
  the_graphicalTrees->calcul_halo (chemin, halo);
  the_graphicTablesManager->inser_obj (li_design, chemin, halo);
  arb_gra = the_graphicTablesManager->graphObj_get_tree (pt_posi);
  s2 = LL_make_lien_atom (li_design);
  nchangchild (arb_gra, s2, 4);
}


/**
  Fonction appellee par detruire_lien:
  elle detruit le chemin passe en parametre ?ch et la reference a ce 
  chemin dans le port recepteur.
  Si ?intern est a faux, le chemin est designable et on l'efface
  de l'ecran. Sinon, il faudra faire un rafraichissement de l'ecran
  pour l'effacer (lien interne).
*/
void
builder::detruire_chemin (ntrees ch, boolean intern)
{
  ntrees e, lchemins, pseg, prem_fils, chemin, pere, lsegs, lsegs_pere, l;
  int arite, id_seg, lien_p, no, nbfils, port, nbcoord;
  typcoord vcoord[2], tcoord[2], halo[2];

  // nDEBUG_subtree(0, 2 ,"entering builder::detruire_chemin", ch);
  if (ch != null)
    {
      enlever_liens_segs (ch, intern);
      pere = nfather (ch);
      if (noper (pere) == gpkOp_chemins)
	{
	  ntreesDestroy (ch);
	  if (ntreearity (pere) == 1)
	    {
	      chemin = nchild (1, pere);
	      pere = nfather (pere);
	      e = nchild (2, chemin);
	      if (e != null)
		{
		  lien_p = the_graphicalTrees->tree_get_graphObj (e);
		  port = (int) nintof (nchild (2, e));
		  l = LL_make_extm (LL_make_lien_atom (lien_p), LL_make_port_atom (port));
		  nchangchild (pere, l, 2);
		  nouv_chemin (lien_p, pere);
		};
	      lchemins = nchild (3, chemin);
	      if (ntreearity (lchemins) != 0)
		nchangchild (pere, lchemins, 3);
	      else
		nchangchild (pere, null, 3);
	      lsegs = nchild (4, chemin);
	      if ((lsegs != null) && ntreearity (lsegs) != 0)
		{
		  lsegs_pere = nchild (4, pere);
		  no = 1;
		  arite = ntreearity (lsegs_pere);
		  if (arite != 0)
		    {
		      pseg = nchild (arite, lsegs_pere);
		      /*  si le premier fils du sous-chemin est le prolongement du 
		         dernier segment du chemin pere, on les fusionne */
		      e = nchild (2, pseg);
		      the_graphicalTrees->getCoord (e, &nbcoord, tcoord, 0);
		      verif_lseg (lsegs, intern);
		      prem_fils = nchild (1, lsegs);
		      if (ISNOT_NILTREE (prem_fils))
			{	/*il faut tester si prem_fils n'est
				   pas a NIL car  verif_lseg peut
				   detruire des segments */
			  the_graphicalTrees->getCoord (nchild (2, prem_fils), &nbcoord, vcoord, 0);
			  if (tcoord[0].x == tcoord[1].x && vcoord[0].x == vcoord[1].x && tcoord[0].x == vcoord[0].x)
			    {
			      tcoord[1].y = vcoord[1].y;
			      no = 2;
			    }
			  else if (tcoord[0].y == tcoord[1].y && vcoord[0].y == vcoord[1].y && vcoord[0].y == tcoord[0].y)
			    {
			      tcoord[1].x = vcoord[1].x;
			      no = 2;
			    };
			  if (no == 2)
			    {
			      id_seg = the_graphicalTrees->tree_get_graphObj (prem_fils);
			      if (!intern)
				pCurObjManager->enlever_l_seg (id_seg);
			      the_graphicTablesManager->enlever_seg (id_seg);
			      the_graphicalTrees->calcul_nouv_halo (typ_seg, tcoord, halo);
			      modif_coord (the_graphicalTrees->tree_get_graphObj (pseg), typ_seg, halo, 2, tcoord, 0);
			    }
			}
		    }
		  nbfils = ntreearity (lsegs);
		  while (no <= nbfils)
		    {
		      e = nchild (no, lsegs);
		      ntreeList_appendLast (lsegs_pere, e);
		      no++;
		    };
		}
	    }
	}			/* if (noper(pere) == gpkOp_chemins) */
      else
	{			/* (noper(pere) != gpkOp_chemins) */
	  nchangchild (pere, null, 3);
	}
    }				/* if (ch != null) */
}


/**
  Detruit les liens qui ont pour port de depart ou d'arrivee le port 
  passe en parametre.
   ?p_gra : arbre representant le port
   ?typ_obj : soit typ_port, soit typ_port_ext.
*/
void
builder::detruire_liens (ntrees p_gra, int typ_obj)
{
  ntrees p_lien, p_ch;
  int intern, ope;

  if (p_gra != null)
    {
      ope = noper (p_gra);
      if (ope == gpkOp_porte || ope == gpkOp_ports)
	{
	  p_ch = nchild (3, p_gra);
	  p_lien = nchild (4, p_gra);
	  if (p_ch != null)
	    {
	      intern = !((((ope == gpkOp_ports) && (typ_obj == typ_port)) || ((ope == gpkOp_porte) && (typ_obj == typ_port_ext))));
	      detruire_chemin (p_ch, intern);
	    }
	  if (p_lien != null)
	    {
	      p_ch = the_graphicTablesManager->graphObj_get_tree ((int) nintof (p_lien));
	      intern = !((((ope == gpkOp_porte) && (typ_obj == typ_port)) || ((ope == gpkOp_ports) && (typ_obj == typ_port_ext))));
	      detruire_chemin (p_ch, intern);
	    }
	}
    }
}



/**
  * modifie la liste de segments du chemin par (?nbcoord) segments
  * de coordonnees(?tcoord) de port diffuseur(?pt_neg).
  * L'arbre et les tables d'acces sont mises a jour
  */
void
builder::modifier_chemin (int pt_neg, int nbcoord, typcoord * tcoord)
{
  ntrees arb_gra, s4, lseg, seg;
  int nbseg, no_seg;

  arb_gra = the_graphicTablesManager->graphObj_get_tree (pt_neg);
  lseg = nchild (4, nchild (3, arb_gra));
  nbseg = ntreearity (lseg);
  if (nbseg > 0)
    seg = nchild (1, lseg);
  else
    seg = null;
  while (seg != null)
    {
      ntreesDestroy (seg);
      no_seg = the_graphicalTrees->tree_get_graphObj (seg);
      pCurObjManager->enlever_l_seg (no_seg);
      the_graphicTablesManager->enlever_seg (no_seg);
      seg = nchild (1, lseg);
    }
  s4 = createListSeg (tcoord, nbcoord, pt_neg);
  nchangchild (nchild (3, arb_gra), s4, 4);
}


/**
  Cette procedure detruit les chemins partant des ports d'entree de
  la boite ?arb_bte.
  Si ?detruit_port est a vrai, elle detruit egalement les ports 
  d'entree et  de sortie de la boite.
  Utilise lors du remplacement d'une boite par du texte.
*/
void
builder::detruire_chemin_restant (ntrees arb_bte, boolean detruit_port)
{
  ntrees l_ports, p_port, p_chem;

  l_ports = the_graphicalTrees->getPorts (arb_bte);
  if (ISNOT_NILTREE (l_ports))
    {
      p_port = nchild (1, l_ports);
      while (p_port != null)
	{
	  if (noper (p_port) == gpkOp_porte)
	    {
	      if ((p_chem = nchild (3, p_port)) != null)
		detruire_chemin (p_chem, pk_true);
	    }
	  else
	    nchangchild (p_port, null, 4);
	  if (detruit_port)
	    {
	      the_graphicTablesManager->enlever_obj (the_graphicalTrees->tree_get_graphObj (p_port));
	      ntreesDestroy (p_port);
	      p_port = nchild (1, l_ports);
	    }
	  else
	    p_port = ntrees_getNext (p_port);
	}
    }
}


/**
  Cree la liste de segments composee des arbres representant
   les (?nbcoord) segments de coordonnees(?tcoord) 
   et de port diffuseur (?pt_neg) et inserer les segments dans
   la liste des segments designables.
*/
ntrees
builder::createListSeg (typcoord * tcoord, int nbcoord, int pt_neg)
{
  ntrees lsegs, pSeg;
  graphObj_type noSeg;
  int i;
  typcoord nouvtcoord[maxcoord], pcoord[2];

  transfcoord (nbcoord + 1, tcoord, nouvtcoord);
  lsegs = LL_make_lsegs_emptyList ();
  i = 0;
  while (i < nbcoord)
    {
      pcoord[0].x = nouvtcoord[i].x;
      pcoord[0].y = nouvtcoord[i].y;
      pcoord[1].x = nouvtcoord[i + 1].x;
      pcoord[1].y = nouvtcoord[i + 1].y;
      pSeg = the_graphicalTrees->createSeg (pcoord, pt_neg, &noSeg);
      ntreeList_appendLast (lsegs, pSeg);
      pCurObjManager->inser_l_seg (noSeg);
      i++;
    };
  return (lsegs);
}


/**
   Creation d'un lien direct entre le port ?port_a et le port ?p_rout_b.

   ATTENTION, cette fonction ne s'occupe pas des halos, ce qui fait que
    le lien peut passer à travers une boite. Elle est appelée lorsque le 
   routage a échoué ou lorsqu'on est sûr qu'il n'y aura pas de problème.
   ?ext_a (ext_b) : boolean a vrai si le port ?port_a (?port_b)
   est externe, a faux sinon (le calcul des coordonnees n'est pas le meme).
*/
void
builder::createlien_direct (ntrees port_a, ntrees port_b, boolean ext_a, boolean ext_b)
{
  int t, ope, x1, x2, y1, y2;
  typcoord tcoord_lien[4];

  if (ext_a)
    t = typ_port_ext;
  else
    t = typ_port;
  get_coord_seg_port (port_a, t, &x1, &y1);
  if (ext_b)
    t = typ_port_ext;
  else
    t = typ_port;
  get_coord_seg_port (port_b, typ_port, &x2, &y2);
  tcoord_lien[0].x = x1;
  tcoord_lien[0].y = y1;
  tcoord_lien[1].x = (x1 + x2) / 2;
  tcoord_lien[1].y = y1;
  tcoord_lien[2].x = tcoord_lien[1].x;
  tcoord_lien[2].y = y2;
  tcoord_lien[3].x = x2;
  tcoord_lien[3].y = y2;
  ope = noper (port_a);
  if (((ope == gpkOp_porte) && ext_a) || ((ope == gpkOp_ports) && !ext_a))
    createchemin (the_graphicalTrees->tree_get_graphObj (port_a), the_graphicalTrees->tree_get_graphObj (port_b), 3, tcoord_lien);
  else
    {
      inverser (tcoord_lien, 3);
      createchemin (the_graphicalTrees->tree_get_graphObj (port_b), the_graphicalTrees->tree_get_graphObj (port_a), 3, tcoord_lien);
    }
}

/**
   Fonction utilisee pour le routage automatique : rend les coordonnees
   (!p_x), (!p_y) et la direction (droite ou gauche !sens_p) du lien
   qui part du port ?p_rout, selon qu'il s'agisse d'un port externe
   ou non (?port_ext).
*/
void
builder::gener_coord_rout_port (ntrees p_rout, boolean port_ext, int *p_x, int *p_y, int *sens_p)
{
  typcoord pcoord[1];
  int nbcoord;

  if (port_ext)
    {
      get_coord_seg_port (p_rout, typ_port_ext, p_x, p_y);
      if (*p_x <= abs_mil_cadre)
	*sens_p = droite;
      else
	*sens_p = gauche;
    }
  else
    {
      the_graphicalTrees->getCoord (p_rout, &nbcoord, pcoord, 1);
      if (pcoord[0].x < 0)
	*sens_p = gauche;
      else
	*sens_p = droite;
      get_coord_seg_port (p_rout, typ_port, p_x, p_y);
    }
}


/**
   Creation d'un lien entre le port recepteur ?port_a et le port 
   diffuseur ?p_rout_b par routage automatique.
   Rend vrai (et construit le lien) si le routage a reussi, rend faux 
   sinon.
   ?ext_a (ext_b) : boolean a vrai si le port ?port_a (?port_b)
   est externe, a faux sinon (le calcul des coordonnees n'est pas le meme).
*/
boolean
builder::rout_port_port (ntrees port_a, ntrees port_b, boolean ext_a, boolean ext_b)
{
  typcoord p_b, p_a, rcoord[30];
  int sens_b, sens_a, nbcoord, npr, no_seg, direc;
  boolean ok;

  gener_coord_rout_port (port_a, ext_a, &p_a.x, &p_a.y, &sens_a);
  gener_coord_rout_port (port_b, ext_b, &p_b.x, &p_b.y, &sens_b);
  if (nchild (3, port_b) != null)
    {
      npr = the_graphicalTrees->tree_get_graphObj (port_b);	/* le port diffuseur est deja connecte */
    }
  else
    npr = 0;
  ok = pRoutor->applyRoutor (p_a.x, p_a.y, sens_a, p_b.x, p_b.y, sens_b, rcoord, &nbcoord, npr, &no_seg);
  if (ok)
    if (npr == 0)
      createchemin (the_graphicalTrees->tree_get_graphObj (port_b), the_graphicalTrees->tree_get_graphObj (port_a), nbcoord, rcoord);
    else
      {
	inverser (rcoord, nbcoord);
	if (rcoord[nbcoord - 1].x == rcoord[nbcoord].x)
	  direc = vertic;
	else
	  direc = horiz;
	nbcoord++;
	ok = recouv_lien (npr, the_graphicalTrees->tree_get_graphObj (port_a), &nbcoord, rcoord, direc, pk_true);
      }
  else if (npr == 0)
    {				/* on ne construit un lien direct que si le port diffuseur n'est pas déjà connecté */
      //      the_messageTool->info_mess(manualrout);
      createlien_direct (port_a, port_b, ext_a, ext_b);
      ok = pk_true;
    }
  return (ok);
}


/**
  Detruit le segment ?seg dans la liste de segments de  l'arbre
  representant le chemin .
  Rend l'arbre representant le segment
*/
ntrees
builder::destr_seg (int seg)
{
  ntrees p_seg;

  p_seg = the_graphicTablesManager->acces_segment (seg);
  ntreesDestroy (p_seg);
  return (p_seg);
}


/**
  Coupe le segment ?seg en deux : utilisee pour les chemins
  multiples
  ?direc_seg : sa direction, ?ext : endroit ou on coupe
*/
void
builder::couper_seg (pt_seg seg, typcoord ext, int direc_seg)
{
  ntrees pSeg, arb_s, nouv_seg;
  graphObj_type noSeg;
  int i, c, c_s, pt_neg;
  typcoord halo[2], tcoord[2], nouvtcoord[2];

  pSeg = seg->arb_seg;
  pt_neg = seg->port_;
  tcoord[0].x = ext.x;
  tcoord[0].y = ext.y;
  transfcoord (1, tcoord, nouvtcoord);
  if (direc_seg == vertic)
    {
      c = nouvtcoord[0].x;
      i = 1;
    }
  else
    {
      c = nouvtcoord[0].y;
      i = 2;
    }
  arb_s = nchild (2, nchild (2, pSeg));
  c_s = (int) nintof (nchild (i, arb_s));
  nmodifvalint (nchild (i, arb_s), c);
  the_graphicalTrees->calcul_halo (pSeg, halo);	/* on met a jour le halo du segment coupe */
  the_graphicTablesManager->setSegElemWithHalo (seg, halo);
  tcoord[0].x = (int) nintof (nchild (1, arb_s));
  tcoord[0].y = (int) nintof (nchild (2, arb_s));
  if (i == 1)
    {
      tcoord[1].x = c_s;
      tcoord[1].y = tcoord[0].y;
    }
  else
    {
      tcoord[1].x = tcoord[0].x;
      tcoord[1].y = c_s;
    };
  nouv_seg = the_graphicalTrees->createSeg (tcoord, pt_neg, &noSeg);
  i = nposit (pSeg);
  ninsertright (i, nouv_seg, nfather (pSeg));
  pCurObjManager->inser_l_seg (noSeg);
}


/**
  Remplace l'arbre representant l'ancien chemin par le nouveau ?ch
  dans la table d'acces pour le lien ?lien
*/
void
builder::nouv_chemin (int lien, ntrees ch)
{
  acces_gra *pt;

  pt = the_graphicTablesManager->getBoxTableElem (lien);
  if (pt != null)
    pt->elem_gra = ch;
}


/**
  Fonction appelee par detruire_chemin:
  Elle enleve les  segments et les liens des tables d'acces et elle enleve
  la reference au lien dans le port recepteur (c'est-a-dire qu'elle met a 
  NIL le 4ieme fils du port recepteur.
  ?ch : chemin diffuseur.
  ?intern : a faux on enleve les segments de la liste des segments designables.
*/
void
builder::enlever_liens_segs (ntrees ch, boolean intern)
{
  ntrees lsegs, e, lchemins;
  int lien, id_seg, no, nbfils;
  if (ch != null)
    {
      e = nchild (2, ch);
      if (e != null)
	{
	  lien = the_graphicalTrees->tree_get_graphObj (e);
	  no = (int) nintof (nchild (2, e));
	  the_graphicTablesManager->enlever_obj (lien);
	  STree_replaceSon (nchild (4, the_graphicTablesManager->graphObj_get_tree (no)), null, 0);
	};
      lsegs = nchild (4, ch);
      if (lsegs != null)
	e = nchild (1, lsegs);
      else
	e = null;
      while (e != null)
	{
	  id_seg = the_graphicalTrees->tree_get_graphObj (e);
	  if (!intern)
	    pCurObjManager->enlever_l_seg (id_seg);
	  the_graphicTablesManager->enlever_seg (id_seg);
	  e = ntrees_getNext (e);
	};
      lchemins = nchild (3, ch);
      nbfils = ntreearity (lchemins);
      no = 1;
      while (no <= nbfils)
	{
	  e = nchild (no, lchemins);
	  enlever_liens_segs (e, intern);
	  no++;
	};
    }
}


/**
   Rajout d'un sous_chemin ch2
   Cette fonction coupe le chemin existant ch qui contient le segment ?seg en deux 
   sous_chemins avec mise a jour de la structure de donnee :
   structure de donnee avant l'appel  :  ch
    "        "    "    apres    "     :  ch ( ch1 , ch2)
    ch devenant la partie commune aux 2 sous_chemins ch1 et ch2
    ?pt_neg : port emetteur
    ?pt_posit : port recepteur
    ?nbcoord, tcoord : coordonnees des segments appartenant a ch2
*/
void
builder::couper_chemin (int pt_neg, int pt_posi, int *nbcoord, typcoord * tcoord, int seg)
{
  ntrees arb_gra, chemin_pere, s1, s2, s3, s4, l, e, se, ch1, ch2, lchemins;
  int lien_p;
  graphObj_type ch_design, li_design;
  typcoord halo[2];

  li_design = the_graphicTablesManager->newGraphObjIdentity ();
  arb_gra = the_graphicTablesManager->acces_segment (seg);
  chemin_pere = the_graphicalTrees->get_chemin (arb_gra);
  ch_design = the_graphicTablesManager->newGraphObjIdentity ();
  s1 = LL_make_design_atom (ch_design);
  l = nchild (2, chemin_pere);
  if (l != null)
    {
      lien_p = the_graphicalTrees->tree_get_graphObj (l);
      e = LL_make_lien_atom (lien_p);
      se = LL_make_port_atom (nintof (nchild (2, l)));
      s2 = LL_make_extm (e, se);
      nchangchild (chemin_pere, null, 2);
    }
  else
    {
      s2 = null;
      lien_p = 0;
    }
  s3 = LL_make_chemins_emptyList ();
  if ((lchemins = nchild (3, chemin_pere)) != null)
    {
      while (!nisemptyList (lchemins))
	ntreeList_appendLast (s3, ndestroyListElem (lchemins, 1));
      nchangchild (chemin_pere, null, 3);
    };
  s4 = LL_make_lsegs_emptyList ();
  while (ntrees_getNext (arb_gra) != null)
    {
      e = ntrees_getNext (arb_gra);
      ntreeList_appendLast (s4, destr_seg (the_graphicalTrees->tree_get_graphObj (e)));
    }
  ch1 = LL_make_chemin (s1, s2, s3, s4);
  if (lien_p != 0)
    nouv_chemin (lien_p, ch1);
  lchemins = LL_make_chemins_list (ch1);
  nchangchild (chemin_pere, lchemins, 3);

  ch_design = the_graphicTablesManager->newGraphObjIdentity ();
  s1 = LL_make_design_atom (ch_design);
  s2 = LL_make_extm (LL_make_lien_atom (li_design), LL_make_port_atom (pt_posi));
  s3 = LL_make_chemins_emptyList ();
  s4 = createListSeg (tcoord, *nbcoord, pt_neg);
  ch2 = LL_make_chemin (s1, s2, s3, s4);
  the_graphicalTrees->calcul_halo (ch2, halo);
  the_graphicTablesManager->inser_obj (li_design, ch2, halo);
  arb_gra = the_graphicTablesManager->graphObj_get_tree (pt_posi);
  s2 = LL_make_lien_atom (li_design);
  nchangchild (arb_gra, s2, 4);
  ntreeList_appendLast (lchemins, ch2);
}

/**
  Fonction apellee par la fonction cr_lien du module builder.
  Elle lance le routage entre (x1,y1) et (x2,y2).
  Si ?inverv est a vrai, elle inverse les points a connecter (le routage
  se fait toujours dans le meme sens recepteur-->diffuseur
*/
boolean
  builder::lancer_routage (boolean invers, int x1, int y1, int sens1, int x2, int y2, int sens2,
			   typcoord * tcoord, int *nbcoord, int pt_neg, int *noseg)
{
  boolean res;

  if (invers)
    res = pRoutor->applyRoutor (x2, y2, sens2, x1, y1, sens1, tcoord, nbcoord, pt_neg, noseg);
  else
    res = pRoutor->applyRoutor (x1, y1, sens1, x2, y2, sens2, tcoord, nbcoord, pt_neg, noseg);
  return (res);
}





/**
  Deplacement des chemins partant ou arrivant d'une boite ?bte.
  les nouvelles coordonnees des ports sont calculees a partir des coordonnees de
  la boite.
  ?bougetail : utilise lors d'un changement de taille pour ne pas deplacer
  les chemins des ports inchanges, il represente les cotes deplaces :
         pour x: -1 cote gauche, 1 cote droit et 0 x inchange.
         pour y: -1 cote haut, 1 cote bas et 0 y inchange.
  ?force a vrai si en changement de taille les ports sont deplaces
  Lors d'un deplacement sans changement de taille : bougetail(1,1) et force a vrai
  Rend vrai si le deplacement est possible, faux sinon.
*/
boolean
builder::depl_bte_chemins (int bte, typcoord * p_bougetail, boolean force)
{
  typcoord bougetail, newcoord[maxports];
  int i, xseg, yseg;
  ntrees p_port, p;

  // printf("DEBUG entering builder::depl_bte_chemins\n"); fflush(stdout);
  bougetail = *p_bougetail;
  p_port = the_graphicalTrees->getFirstPort (the_graphicTablesManager->graphObj_get_tree (bte));
  if (p_port == null)
    return (pk_true);
  p = p_port;
  i = 0;
  while (p != null)
    {
      get_coord_seg_port (p, typ_port, &xseg, &yseg);
      newcoord[i].x = xseg;
      newcoord[i].y = yseg;
      p = ntrees_getNext (p);
      i++;
    }
  return (trans_ext_chemins (p_port, i, newcoord, pk_true, p_bougetail, force));
}


/**
  Cette procedure insere dans la liste (?l_segs) composee de 
   numeros de segments les numeros de segments qui coupent la boite
   de numero (?bte) et epure les segments qui representent le meme
   chemin (on ne garde alors qu'un segment)
*/
void
builder::cherch_segs_coupe_bte (int bte, ntrees l_segs)
{
  ntrees i_arb_seg, i_rac_chem, j_arb_seg, j_rac_chem;
  typcoord tcoord[2], scoord[2], centre_bte;
  int nbcoord, i, j, nb_seg, hauteur, largeur;

  the_graphicalTrees->getCoord (the_graphicTablesManager->graphObj_get_tree (bte), &nbcoord, tcoord, 0);
  centre_bte.x = tcoord[0].x;
  centre_bte.y = tcoord[0].y;
  largeur = tcoord[1].x;
  hauteur = tcoord[1].y;
  scoord[0].x = centre_bte.x - largeur;
  scoord[0].y = centre_bte.y - hauteur;
  scoord[1].x = centre_bte.x - largeur;
  scoord[1].y = centre_bte.y + hauteur;
  pCurObjManager->cherch_segs_coupe_seg (scoord, vertic, l_segs);
  scoord[0].x = centre_bte.x + largeur;
  scoord[1].x = centre_bte.x + largeur;
  pCurObjManager->cherch_segs_coupe_seg (scoord, vertic, l_segs);
  scoord[0].x = centre_bte.x - largeur;
  scoord[1].y = centre_bte.y - hauteur;
  pCurObjManager->cherch_segs_coupe_seg (scoord, horiz, l_segs);
  scoord[0].y = centre_bte.y + hauteur;
  scoord[1].y = centre_bte.y + hauteur;
  pCurObjManager->cherch_segs_coupe_seg (scoord, horiz, l_segs);
  nb_seg = ntreearity (l_segs);
  if (nb_seg > 1)
    {
      i = 1;
      while (i <= nb_seg)
	{
	  i_arb_seg = the_graphicTablesManager->acces_segment ((int) nintof (nchild (i, l_segs)));
	  i_rac_chem = the_graphicalTrees->get_racine_chemin (i_arb_seg);
	  j = i + 1;
	  while (j > i && j <= nb_seg)
	    {
	      j_arb_seg = the_graphicTablesManager->acces_segment ((int) nintof (nchild (j, l_segs)));
	      j_rac_chem = the_graphicalTrees->get_racine_chemin (j_arb_seg);
	      if (i_rac_chem == j_rac_chem)
		{		/* on ne laisse qu'un seul segment pour un meme port diffuseur */
		  ndestroyListElem (l_segs, j);
		  nb_seg--;
		}
	      else
		j++;
	    }
	  i++;
	}
    }
}

/**
  * Cette fonction est appellee par la fonction cr_lien lorsqu'on construit un lien
  * vers un port deja connecte (port diffuseur) ?pt_neg.
  * Elle recherche le segment (appartenant a un des chemins qui partent de ?pt_neg)
  * le plus proche de l'avant dernier segment existant et de direction oppose a
  * ?direc_seg.
  * Si elle trouve un segment, la structure de donnees du nouveau lien est 
  construite, sinon rend faux.
  * ?pt_posit : numero du port recepteur.
  * ?nbcoord, ?tcoord : nombre et tableau des coordonnees.
  * ?direc_seg : direction de l'avant dernier segment du lien construit
  * ?autoRouting : si vrai, on est en routage automatique, sinon en routage manuel 
  */
boolean
builder::recouv_lien (int pt_neg, int pt_posi, int *nbcoord, typcoord * tcoord, int direc_seg, boolean autoRouting)
{
  int i, seg_t, n, direc;
  float d1, d2;
  boolean partage, coupe, ok;
  ntrees pSeg, p;
  pt_seg seg;
  typcoord ext[2], scoord[2], halo[2];

  ext[0].x = tcoord[*nbcoord - 2].x;
  ext[0].y = tcoord[*nbcoord - 2].y;
  ext[1].x = tcoord[*nbcoord - 1].x;
  ext[1].y = tcoord[*nbcoord - 1].y;
  seg = pCurObjManager->cherch_seg (pt_neg, ext, direc_seg, &coupe);
  ok = pk_false;
  if (seg != null)
    {
      pSeg = seg->arb_seg;
      the_graphicalTrees->getCoord (nchild (2, pSeg), &n, scoord, 0);
      if (!coupe)
	{
	  d1 = dist (ext[1], scoord[0]);
	  d2 = dist (ext[1], scoord[1]);
	  i = 0;
	  partage = pk_false;
	  if (d2 < d1)
	    {
	      partage = pk_true;
	      i = 1;
	    };
	  if (direc_seg == vertic)
	    {
	      tcoord[(*nbcoord) - 1].y = tcoord[*nbcoord].y = scoord[i].y;
	      tcoord[*nbcoord].x = scoord[i].x;
	    }
	  else
	    {
	      tcoord[(*nbcoord) - 1].x = tcoord[*nbcoord].x = scoord[i].x;
	      tcoord[*nbcoord].y = scoord[i].y;
	    };
	  if (tcoord[*nbcoord].x == tcoord[(*nbcoord) - 1].x && tcoord[*nbcoord].y == tcoord[(*nbcoord) - 1].y)
	    (*nbcoord)--;
	  if (!partage)
	    {
	      p = ntrees_getPrevious (pSeg);
	      if (p != null)
		pSeg = p;
	      else
		{
		  p = ntrees_getNext (nfather (nfather (nfather (pSeg))));
		  if (p != null)
		    {
		      i = ntreearity (p);
		      pSeg = nchild (i, p);
		    }
		  else
		    pSeg = null;
		}
	    }
	}
      else			// the found segment must be cut
	{
	  if (direc_seg == vertic)
	    tcoord[(*nbcoord) - 1].y = scoord[0].y;
	  else
	    tcoord[(*nbcoord) - 1].x = scoord[0].x;
	  *nbcoord = *nbcoord - 1;
	};
      inverser (tcoord, *nbcoord);
      if (pSeg != null)
	{
	  seg_t = the_graphicalTrees->tree_get_graphObj (pSeg);
	  i = 0;
	  ok = pk_true;
	  if (!autoRouting)	//in manuel routing, the different segments must be created
	    while (i < *nbcoord && ok)
	      {
		scoord[0].x = tcoord[i].x;
		scoord[0].y = tcoord[i].y;
		scoord[1].x = tcoord[i + 1].x;
		scoord[1].y = tcoord[i + 1].y;
		the_graphicalTrees->calcul_nouv_halo (typ_seg, scoord, halo);
		if (tcoord[i].y == tcoord[i + 1].y)
		  direc = horiz;
		else
		  direc = vertic;
		ok = pCurObjManager->test_cr_seg (halo, direc);
		i++;
	      };
	  if (ok)
	    {
	      if (coupe)
		couper_seg (seg, ext[1], direc_seg);
	      couper_chemin (pt_neg, pt_posi, nbcoord, tcoord, seg_t);
	    };
	}
    }
  return (ok);
}


/**
  Translations des chemins externes partant ou arrivant (selon le type
   des ports) d'une liste de ports .
   Rend vrai si tous les chemins sont translatables: le deplacement des
   segments s'effectue par routage
   ?prem_port : premier port de la liste de ports
   ?cpt_port : le nombre de ports
   ?newcoord : L'ensemble des nouvelles coordonnees
   ?bougetail : utilises lors d'un changement de taille 
   pour ne pas deplaces les chemins des ports inchanges
   et represente les cotes deplaces 
   pour x: -1 cote gauche, 1 cote droit et 0 x inchange.
   pour y: -1 cote haut, 1 cote bas et 0 y inchange.
   ?force a vrai si en changement de taille les ports sont deplaces
   lors d'un deplacement sans changement de taille : bougetail(1,1) et force(vrai)
*/
boolean
builder::trans_ext_chemins (ntrees prem_port, int cpt_port, typcoord * newcoord, boolean xmod, typcoord * p_bougetail, boolean force)
{
  int c, nbcoord, noport, noseg;
  ntrees ieme_port, lien, chemin, seg;
  boolean no_pb, saute;
  typcoord scoord[2], pcoord[1], bougetail;

  // printf("\nDEBUG ===============================================================\n"); fflush(stdout);
  // printf("DEBUG entering builder::trans_ext_chemins\n"); fflush(stdout);
  bougetail = *p_bougetail;
  no_pb = pk_true;
  ieme_port = prem_port;
  for (c = 0; (c <= cpt_port - 1) && no_pb; c++)
    {
      the_graphicalTrees->getCoord (ieme_port, &nbcoord, pcoord, 1);
      switch (bougetail.x)
	{
	case 0:
	  saute = !force;
	  break;
	case 1:
	  if (pcoord[0].x > 0)
	    saute = pk_false;
	  else
	    saute = !(force && bougetail.y != 0);
	  break;
	case -1:
	  if (pcoord[0].x < 0)
	    saute = pk_false;
	  else
	    saute = !(force && bougetail.y != 0);
	}
      if (!saute)
	{
	  noport = the_graphicalTrees->tree_get_graphObj (ieme_port);
	  if (noper (ieme_port) == gpkOp_porte)
	    {
	      if ((lien = nchild (4, ieme_port)) != null)
		{
		  chemin = the_graphicTablesManager->graphObj_get_tree ((int) nintof (lien));
		  if (nchild (2, chemin) != null)
		    {
		      seg = nchild (ntreearity (nchild (4, chemin)), nchild (4, chemin));
		      noseg = the_graphicalTrees->tree_get_graphObj (seg);
		      the_graphicalTrees->getCoord (nchild (2, seg), &nbcoord, scoord, 0);
		      scoord[0].y = scoord[1].y = newcoord[c].y;
		      if (xmod)
			scoord[1].x = newcoord[c].x;
		      no_pb = depl_seg_routage (noseg, scoord, noport, pk_false, &bougetail);
		    }
		  else
		    no_pb = pk_false;
		}
	    }
	  else if ((chemin = nchild (3, ieme_port)) != null)
	    {
	      seg = nchild (1, nchild (4, chemin));
	      noseg = the_graphicalTrees->tree_get_graphObj (seg);
	      the_graphicalTrees->getCoord (nchild (2, seg), &nbcoord, scoord, 0);
	      scoord[0].y = newcoord[c].y;
	      if (xmod)
		scoord[0].x = newcoord[c].x;
	      no_pb = depl_seg_routage (noseg, scoord, noport, pk_false, &bougetail);
	    }
	}
      ieme_port = ntrees_getNext (ieme_port);
    }
  if (no_pb)
    {
      // printf("DEBUG  exiting builder::trans_ext_chemins TRUE\n"); fflush(stdout);
      return pk_true;
    }
  // printf("DEBUG  exiting builder::trans_ext_chemins FALSE\n"); fflush(stdout);
  the_messageTool->error_mess (routimpnodepl);
  return pk_false;
}


/**
  * translations des chemins internes a une boite, arrivant
  * ou partant d'une liste de ports
  * Rend vrai si tous les chemins sont translatables
  * idem que trans_ext_chemins sauf:
  * la modification de l'abscisse n'est pas prise en compte pour
  * un segment interne.
  * ?arb_bte : l'arbre graphique de la boite.
  */
boolean
builder::trans_int_chemins (ntrees arb_bte, ntrees prem_port, int cpt_port, typcoord * newcoord)
{
  int noport, c, nbcoord, ytrans;
  ntrees ieme_port, chemin, lien, seg;
  boolean no_pb;
  typcoord tcoord[2], scoord[2], bougetail;
  float fact_ech_x, fact_ech_y;

  calcul_fact_ech (arb_bte, &fact_ech_x, &fact_ech_y);
  no_pb = pk_true;
  bougetail.x = 1;
  bougetail.y = 1;
  the_graphicalTrees->getCoord (arb_bte, &nbcoord, tcoord, 0);
  ieme_port = prem_port;
  for (c = 0; (c <= cpt_port - 1) && no_pb; c++)
    {
      if (noper (ieme_port) == gpkOp_ports)
	{
	  if ((lien = nchild (4, ieme_port)) != null)
	    {
	      noport = the_graphicalTrees->tree_get_graphObj (ieme_port);
	      chemin = the_graphicTablesManager->graphObj_get_tree ((int) nintof (lien));
	      if (nchild (2, chemin) != null)
		{
		  ytrans = (int) ((newcoord[c].y - tcoord[0].y) * fact_ech_y) + ord_mil_cadre;
		  seg = nchild (ntreearity (nchild (4, chemin)), nchild (4, chemin));
		  the_graphicalTrees->getCoord (nchild (2, seg), &nbcoord, scoord, 0);
		  scoord[0].y = scoord[1].y = ytrans;
		  pCurObjManager->descendre (the_graphicalTrees->tree_get_graphObj (arb_bte), pk_true);
		  no_pb = depl_seg_routage (the_graphicalTrees->tree_get_graphObj (seg), scoord, noport, pk_true, &bougetail);
		  pCurObjManager->remonter (pk_true);
		}
	      else
		no_pb = pk_false;
	    }
	}
      else
	{
	  if ((chemin = nchild (3, ieme_port)) != null)
	    {
	      noport = the_graphicalTrees->tree_get_graphObj (ieme_port);
	      ytrans = (int) ((newcoord[c].y - tcoord[0].y) * fact_ech_y) + ord_mil_cadre;
	      seg = nchild (1, nchild (4, chemin));
	      the_graphicalTrees->getCoord (nchild (2, seg), &nbcoord, scoord, 0);
	      scoord[0].y = scoord[1].y = ytrans;
	      pCurObjManager->descendre (the_graphicalTrees->tree_get_graphObj (arb_bte), pk_true);
	      no_pb = depl_seg_routage (the_graphicalTrees->tree_get_graphObj (seg), scoord, noport, pk_true, &bougetail);
	      pCurObjManager->remonter (pk_true);
	    }
	}
      ieme_port = ntrees_getNext (ieme_port);
    }
  if (!no_pb)
    the_messageTool->error_mess (routimpnodepl);
  return (no_pb);
}



/**
   Rend vrai s'il est possible d'effectuer le routage entre le port
   ?arb_port_bte, le port (?arb_port_diff) et les ports lies a arb_port_diff
   (?l_no_ports_lies).
   ?intern a vrai on effectue le routage a l'interieur de la boite
   (arb_port_bte est alors un port externe)
   (?l_bte_enc) la liste des numeros des boites encapsulees
   ?bte le numero de la boite qui encapsule
*/
boolean
builder::rout_ports_lies_encaps (int bte, ntrees arb_port_bte, ntrees arb_port_diff, ntrees l_no_ports_lies, ntrees l_bte_enc, boolean intern)
{
  ntrees arb_no_port_lie, arb_port_lie, p_rout_a, p_rout_b;
  boolean p_ext_a, p_ext_b, cont, r_port_diff, port_bte_sortie;
  int no_bte_lie, no_bte_diff;

  cont = pk_true;
  no_bte_diff = the_graphicalTrees->no_boite_mere (arb_port_diff);
  if (noper (arb_port_bte) == gpkOp_ports)
    port_bte_sortie = pk_true;
  else
    port_bte_sortie = pk_false;
  r_port_diff = ((port_bte_sortie && intern) || (!port_bte_sortie && !intern));
  if (!port_bte_sortie && !intern)
    arb_no_port_lie = null;
  else
    arb_no_port_lie = nchild (1, l_no_ports_lies);
  while (arb_no_port_lie != null && cont)
    {
      arb_port_lie = the_graphicTablesManager->graphObj_get_tree ((int) nintof (arb_no_port_lie));
      no_bte_lie = the_graphicalTrees->no_boite_mere (arb_port_lie);
      if ((intern && the_graphicalTrees->is_in_list (no_bte_lie, l_bte_enc)) || (!intern && !the_graphicalTrees->is_in_list (no_bte_lie, l_bte_enc)))
	{
	  if (intern && noper (arb_port_bte) != noper (arb_port_lie))
	    {
	      p_rout_b = arb_port_diff;
	      p_ext_b = pk_false;
	    }
	  else
	    {
	      p_rout_b = arb_port_bte;
	      p_ext_b = intern;
	    }
	  p_rout_a = arb_port_lie;
	  if (!intern)
	    p_ext_a = the_graphicalTrees->isExternPort (the_graphicTablesManager->graphObj_get_tree (bte), p_rout_a);
	  else
	    p_ext_a = pk_false;
	  cont = rout_port_port (p_rout_a, p_rout_b, p_ext_a, p_ext_b);
	}
      arb_no_port_lie = ntrees_getNext (arb_no_port_lie);
    }

  if (r_port_diff)
    if (!intern)
      {
	p_ext_b = the_graphicalTrees->isExternPort (the_graphicTablesManager->graphObj_get_tree (bte), arb_port_diff);
	cont = rout_port_port (arb_port_bte, arb_port_diff, pk_false, p_ext_b);
      }
    else
      cont = rout_port_port (arb_port_bte, arb_port_diff, pk_true, pk_false);

  return (cont);
}


/**
   Fonction appelee lors de l'encapsulation ou le mapping:
   elle  parcourt les ports de la boite creee (?bte), accede a l'attribut
   de type algo de chaque port (s'il existe) et effectue le 
   routage entre ce port, le premier fils de l'attribut representant 
   le port diffuseur et le deuxieme fils representant la liste des
   anciens ports lies au port diffuseur.
   ?intern a vrai on effectue le routage a l'interieur de
   la boite  (descente et remontee simulee) et on effectue en plus le 
   routage des ports diffuseurs, dont les numeros sont dans la liste 
   (?l_n_p_d_e), vers leurs ports lies (ces ports ne sont connectes
   qu'avec des boites encapsulees. POURQUOI FAIRE UN ROUTAGE????
   (?l_bte_enc) la liste des numeros des boites encapsulees
*/
boolean
builder::rout_ports_bte_encaps (int bte, ntrees l_bte_enc, ntrees l_n_p_d_e, boolean intern)
{
  ntrees arb_gra, p_port, arb_port_diff, arb_no_port_diff, l_no_ports_lies, attr_port_bte, ch;
  boolean cont, b;

  cont = pk_true;
  arb_gra = the_graphicTablesManager->graphObj_get_tree (bte);
  p_port = the_graphicalTrees->getFirstPort (arb_gra);
  if (intern)
    pCurObjManager->descendre (bte, pk_true);
  while ((p_port != null) && cont)
    {
      attr_port_bte = nattrntrees (p_port, ALGO_ATTRNUMBER);
      if (attr_port_bte != null)
	{
	  arb_no_port_diff = nchild (1, attr_port_bte);
	  arb_port_diff = the_graphicTablesManager->graphObj_get_tree ((int) nintof (arb_no_port_diff));
	  l_no_ports_lies = nchild (2, attr_port_bte);
	  cont = rout_ports_lies_encaps (bte, p_port, arb_port_diff, l_no_ports_lies, l_bte_enc, intern);
	}
      p_port = ntrees_getNext (p_port);
    }
  if (intern)
    {
      arb_no_port_diff = nchild (1, l_n_p_d_e);
      while ((arb_no_port_diff != null) && cont)
	{
	  arb_port_diff = the_graphicTablesManager->graphObj_get_tree ((int) nintof (arb_no_port_diff));
	  l_no_ports_lies = the_graphicalTrees->get_ports_lies (the_graphicalTrees->tree_get_graphObj (arb_port_diff), pk_false, &ch, &b);
	  detruire_chemin (nchild (3, arb_port_diff), pk_false);
	  cont = rout_port_diff_ports_lies (arb_port_diff, l_no_ports_lies, 0);
	  arb_no_port_diff = ntrees_getNext (arb_no_port_diff);
	}
      pCurObjManager->remonter (pk_true);
    }
  return (cont);
}


/**
  Cette fonction detruit les chemins qui seront recontruits
   apres routage lors d'un encapsulage.
   on parcourt les ports de la boite creee (?bte)
   si le port est un port d'entree et (?intern) a faux, on 
   accede au deuxieme fils de l'attribut de type algo representant
   les ports lies et pour chaque port lie appartenant a une boite
   encapsulee on detruit le chemin.
   si le port est un port de sortie et intern a vrai 
   on accede au premier fils de l'attribut de type algo representant
   le port diffuseur et on  detruitle chemin
   (?l_bte_enc) la liste des numeros des boites encapsulees
   (descente et remontee simulee si intern est vrai)
*/
void
builder::detr_liens_ports_encaps (int bte, ntrees l_bte_enc, boolean intern)
{
  ntrees p_port, attr_port_bte, arb_port_diff, l_no_ports_lies, arb_no_port_lie, arb_port_lie, ch;
  int no_arb_port_diff, no_bte_lie;

  p_port = the_graphicalTrees->getFirstPort (the_graphicTablesManager->graphObj_get_tree (bte));
  if (p_port != null)
    {
      if (intern)
	pCurObjManager->descendre (bte, pk_true);
      while (p_port != null)
	{
	  attr_port_bte = nattrntrees (p_port, ALGO_ATTRNUMBER);
	  if (attr_port_bte != null)
	    {
	      no_arb_port_diff = the_graphicalTrees->tree_get_graphObj (attr_port_bte);
	      arb_port_diff = the_graphicTablesManager->graphObj_get_tree (no_arb_port_diff);
	      if (!intern && noper (p_port) == gpkOp_porte)
		{
		  l_no_ports_lies = nchild (2, attr_port_bte);
		  arb_no_port_lie = nchild (1, l_no_ports_lies);
		  while (arb_no_port_lie != null)
		    {
		      arb_port_lie = the_graphicTablesManager->graphObj_get_tree ((int) nintof (arb_no_port_lie));
		      no_bte_lie = the_graphicalTrees->no_boite_mere (arb_port_lie);
		      if (the_graphicalTrees->is_in_list (no_bte_lie, l_bte_enc))
			{
			  ch = the_graphicTablesManager->graphObj_get_tree ((int) nintof (nchild (4, arb_port_lie)));
			  detruire_chemin (ch, pk_false);
			}
		      arb_no_port_lie = ntrees_getNext (arb_no_port_lie);
		    }
		}
	      else if (noper (p_port) == gpkOp_ports)
		{
		  ch = nchild (3, arb_port_diff);
		  detruire_chemin (ch, pk_false);
		}
	    }
	  p_port = ntrees_getNext (p_port);
	}
      if (intern)
	pCurObjManager->remonter (pk_true);
    }
}



/**
  Cette fonction detruit les liens qui seront recontruits apres 
  routage lors d'un mapping (fonction calquee sur la fonction
  detr_liens_ports_encaps. On ne detruit que les liens necessaires,
  c'est-a-dire les liens avec des boites qui n'appartiennent pas
  a l'ensemble des boites mappees ?l_bte_map.
  Pour cela, on parcourt les ports de la boite ?bte dans laquelle
  on fait le mapping et on recupere l'attribut de type "algo".
  Attention, les numeros des objets contenus dans l'attribut "algo"
  ne correspondent pas aux numeros des objets qu'on manipule, vu
  qu'il y a eu copie des arbres avec nouveaux numeros, il faut
  passer par la fonction the_graphicTablesManager->locinconvnum.
*/
void
builder::detr_liens_mapping (int bte, ntrees l_bte_map)
{
  ntrees p_port, p_mapping, arb_port_diff, l_no_ports_lies, l_nouv_ports_lies, arb_no_port_lie, arb_port_lie, ch, p, p_nouv_port;
  int i, no, nb, no_port_diff, no_bte_lie, nouv_port_lie, nouv_port_diff;
  boolean alinterieur, intern, b;

  p_port = the_graphicalTrees->getFirstPort (the_graphicTablesManager->graphObj_get_tree (bte));
  if (p_port != null)
    {
      while (p_port != null)
	{
	  p_mapping = nattrntrees (p_port, ALGO_ATTRNUMBER);
	  if (p_mapping != null)
	    {
	      no_port_diff = the_graphicalTrees->tree_get_graphObj (p_mapping);
	      arb_port_diff = the_graphicTablesManager->graphObj_get_tree (no_port_diff);

	      l_no_ports_lies = nchild (2, p_mapping);
	      arb_no_port_lie = nchild (1, l_no_ports_lies);
	      while (arb_no_port_lie != null)
		{
		  no = (int) nintof (arb_no_port_lie);
		  arb_port_lie = the_graphicTablesManager->graphObj_get_tree (no);
		  no_bte_lie = the_graphicalTrees->no_boite_mere (arb_port_lie);
		  alinterieur = the_graphicalTrees->is_in_list (no_bte_lie, l_bte_map);
		  if ((alinterieur && (noper (p_port) == gpkOp_porte)) || (!alinterieur && (noper (p_port) == gpkOp_ports)))
		    {
		      nouv_port_lie = the_graphicTablesManager->locinconvnum (no, 0);
		      nouv_port_diff = the_graphicTablesManager->locinconvnum (no_port_diff, 0);
		      p_nouv_port = the_graphicTablesManager->graphObj_get_tree (nouv_port_lie);
		      intern = (noper (arb_port_lie) == gpkOp_ports);
		      l_nouv_ports_lies = the_graphicalTrees->get_ports_lies (nouv_port_lie, intern, &ch, &b);
		      nb = ntreearity (l_nouv_ports_lies);
		      for (i = 1; i <= nb; i++)
			if ((int) nintof (nchild (i, l_nouv_ports_lies)) == nouv_port_diff)
			  if ((p = nchild (4, p_nouv_port)) != null)
			    {
			      ch = the_graphicTablesManager->graphObj_get_tree ((int) nintof (p));
			      detruire_chemin (ch, !alinterieur);
			    }
		    }
		  arb_no_port_lie = ntrees_getNext (arb_no_port_lie);
		}
	    }
	  p_port = ntrees_getNext (p_port);
	}
    }
}



/**
  Cette fonction effectue si possible le deplacement du segment "obj"
     obj : numero du segment, vcoord ses nouvelles coordonnees.
     obj1 : numero du segment precedent ou successeur suivant la
            valeur de prec, vcoord1 ses coordonnees.
     obj2 : numero du segment successeur (si different de 0), vcoord2
            ses coordonnees.
     prec : booleen utilise que lorsque obj2 vaut 0 ; si prec est a vrai,
            obj1 est le segment precedant obj, sinon c'est le segment
            succedant a obj.
     p_bougetail :
   Remarque : si obj2 vaut 0, alors obj arrive ou part d'un port.
              si obj1=obj2=0, alors obj part et arrive sur un port.
*/
boolean
  builder::depl_seg (int obj, typcoord * vcoord, int obj1, typcoord * vcoord1, int obj2, typcoord * vcoord2, boolean prec, typcoord * p_bougetail)
{
  int obj_prec, nbcoord, noport, noport1, sens, sens1, rg, rg1, i, x, x1, y, y1;
  boolean b1, b2, b3, b4, result, align;
  ntrees pseg, pseg1, pseg2, pseg_prec, arb_bte, p_port, p_port1;
  typcoord halo[2], halo_prec[2], halo_succ[2], scoord[2], tcoord[2], bougetail;

  bougetail = *p_bougetail;
  result = pk_false;
  align = pk_false;
  pseg = the_graphicTablesManager->acces_segment (obj);
  if (obj1 != 0)
    pseg1 = the_graphicTablesManager->acces_segment (obj1);
  if (obj2 != 0)
    pseg2 = the_graphicTablesManager->acces_segment (obj2);
  if (vcoord[0].x == vcoord[1].x)
    {
      sens = vertic;
      sens1 = horiz;
    }
  else
    {
      sens = horiz;
      sens1 = vertic;
    };
  if ((obj1 == 0) && (obj2 == 0))	/* cas d'un segment entre 2 ports */
    {
      x = vcoord[0].x + 1;
      y = vcoord[0].y;
      scoord[0].x = x;
      scoord[0].y = y;
      noport = the_graphicalTrees->get_no_port_diff (pseg);
      p_port = the_graphicTablesManager->graphObj_get_tree (noport);
      arb_bte = the_graphicalTrees->boite_mere (noport);
      if (deplac_port (noport, &x, &y, &rg))
	{
	  the_graphicalTrees->annuler_halo_seg_ports (1, p_port, pk_true, pk_false, &bougetail, pk_true);
	  if (trans_int_chemins (arb_bte, p_port, 1, scoord))
	    {
	      x1 = vcoord[1].x + 1;
	      y1 = vcoord[1].y;
	      scoord[0].x = x1;
	      scoord[0].y = y1;
	      noport1 = the_graphicalTrees->get_no_port_recept (pseg);
	      p_port1 = the_graphicTablesManager->graphObj_get_tree (noport1);
	      arb_bte = the_graphicalTrees->boite_mere (noport1);
	      if (deplac_port (noport1, &x1, &y1, &rg1) && (y1 == y))
		{
		  the_graphicalTrees->annuler_halo_seg_ports (1, p_port1, pk_true, pk_false, &bougetail, pk_true);
		  if (trans_int_chemins (arb_bte, p_port1, 1, scoord))
		    {
		      pCurObjManager->enlever_l_seg (obj);
		      vcoord[0].y = vcoord[1].y = y;
		      the_graphicalTrees->calcul_nouv_halo (typ_seg, vcoord, halo);
		      if (pCurObjManager->test_cr_seg (halo, sens))
			{
			  scoord[0].x = x;
			  scoord[0].y = y;
			  modif_coord (noport, typ_port, halo, 1, scoord, rg);
			  scoord[0].x = x1;
			  scoord[0].y = y1;
			  modif_coord (noport1, typ_port, halo, 1, scoord, rg1);
			  vcoord[1].y = vcoord[0].y = y;
			  modif_coord (obj, typ_seg, halo, 2, vcoord, rg);
			  pCurObjManager->inser_l_seg (obj);
			  result = pk_true;
			}
		      else
			the_messageTool->error_mess (deplimphalo);
		    }
		}
	      else
		the_messageTool->error_mess (deplportimp);
	    }
	}
      else
	the_messageTool->error_mess (deplportimp);
    }
  else if (obj2 == 0)		/* cas d'un segment qui a un seul segment voisin */
    {
      if (prec)
	{
	  i = 1;
	  noport = the_graphicalTrees->get_no_port_recept (pseg);	/* port recepteur */
	  pseg2 = ntrees_getPrevious (pseg1);
	}
      else
	{
	  i = 0;
	  noport = the_graphicalTrees->get_no_port_diff (pseg);	/* le port est un port diffuseur */
	  pseg2 = ntrees_getNext (pseg1);
	}
      arb_bte = the_graphicalTrees->boite_mere (noport);
      x = vcoord[i].x + 1;
      if (abs (vcoord1[1].y - vcoord1[0].y) <= (halo_segm + 1))	/* Modif 16-10-98 */
	{
	  align = pk_true;
	  y = vcoord1[1 - i].y;
	}
      else
	y = vcoord[i].y;
      scoord[0].x = x;
      scoord[0].y = y;
      p_port = the_graphicTablesManager->graphObj_get_tree (noport);
      if (deplac_port (noport, &x, &y, &rg))
	{
	  the_graphicalTrees->annuler_halo_seg_ports (1, p_port, pk_true, pk_false, &bougetail, pk_true);
	  if (trans_int_chemins (arb_bte, p_port, 1, scoord))
	    {
	      if (align && (y == vcoord1[1 - i].y))
		{
		  obj2 = the_graphicalTrees->tree_get_graphObj (pseg2);
		  the_graphicalTrees->getCoord (nchild (2, pseg2), &nbcoord, tcoord, 0);
		  tcoord[i].x = vcoord[i].x;
		  the_graphicalTrees->calcul_nouv_halo (typ_seg, tcoord, halo_prec);
		  pCurObjManager->enlever_l_seg (obj);
		  pCurObjManager->enlever_l_seg (obj1);
		  pCurObjManager->enlever_l_seg (obj2);
		  if (pCurObjManager->test_cr_seg (halo_prec, sens))
		    {
		      scoord[0].x = x;
		      scoord[0].y = y;
		      modif_coord (noport, typ_port, halo, 1, scoord, rg);
		      modif_coord (obj2, typ_seg, halo_prec, nbcoord, tcoord, rg);
		      ntreesDestroy (pseg);
		      ntreesDestroy (pseg1);
		      the_graphicTablesManager->enlever_seg (obj);
		      the_graphicTablesManager->enlever_seg (obj1);
		      pCurObjManager->inser_l_seg (obj2);
		      result = pk_true;
		    }
		  else
		    the_messageTool->error_mess (deplimphalo);
		}
	      else
		{
		  if (y != vcoord[i].y)
		    vcoord1[i].y = vcoord[1 - i].y = vcoord[i].y = y;
		  pCurObjManager->enlever_l_seg (obj);
		  pCurObjManager->enlever_l_seg (obj1);
		  the_graphicalTrees->calcul_nouv_halo (typ_seg, vcoord, halo);
		  the_graphicalTrees->calcul_nouv_halo (typ_seg, vcoord1, halo_prec);
		  if ((pCurObjManager->test_cr_seg (halo, sens)) && (pCurObjManager->test_cr_seg (halo_prec, sens1)))
		    {
		      scoord[0].x = x;
		      scoord[0].y = y;
		      modif_coord (noport, typ_port, halo, 1, scoord, rg);
		      modif_coord (obj, typ_seg, halo, nbcoord, vcoord, rg);
		      modif_coord (obj1, typ_seg, halo_prec, nbcoord, vcoord1, rg);
		      pCurObjManager->inser_l_seg (obj);
		      pCurObjManager->inser_l_seg (obj1);
		      result = pk_true;
		    }
		  else
		    the_messageTool->error_mess (deplimphalo);
		}
	    }
	}
      else
	the_messageTool->error_mess (deplportimp);
    }

  else
    {				/* cas d'un segment entre 2 segments */
      pCurObjManager->enlever_l_seg (obj);
      b1 = abs (vcoord1[1].y - vcoord1[0].y) <= halo_segm && sens == horiz;
      b2 = abs (vcoord2[1].y - vcoord2[0].y) <= halo_segm && sens == horiz;
      b3 = (abs (vcoord1[1].x - vcoord1[0].x) <= halo_segm) && (ntrees_getPrevious (pseg1) != null) && sens == vertic;
      b4 = (abs (vcoord2[1].x - vcoord2[0].x) <= halo_segm) && (ntrees_getNext (pseg2) != null) && sens == vertic;
      if ((b1) || (b2) || (b3) || (b4))
	{
	  if ((b1) || (b3))
	    {
	      pseg_prec = ntrees_getPrevious (pseg1);
	      i = 1;
	      if (b1)
		vcoord1[1].x = vcoord1[0].x = vcoord2[0].x;
	      else
		vcoord1[1].y = vcoord1[0].y = vcoord2[0].y;
	    }
	  else
	    {
	      pseg_prec = ntrees_getNext (pseg2);
	      i = 0;
	    }
	  obj_prec = the_graphicalTrees->tree_get_graphObj (pseg_prec);
	  the_graphicalTrees->getCoord (nchild (2, pseg_prec), &nbcoord, tcoord, 0);
	  if (sens == horiz)
	    {
	      vcoord1[1].y = vcoord2[1].y;
	      tcoord[i].x = vcoord[i].x;
	    }
	  else
	    {
	      vcoord1[1].x = vcoord2[1].x;
	      tcoord[i].y = vcoord[i].y;
	    }
	  the_graphicalTrees->calcul_nouv_halo (typ_seg, vcoord1, halo_succ);
	  the_graphicalTrees->calcul_nouv_halo (typ_seg, tcoord, halo_prec);
	  pCurObjManager->enlever_l_seg (obj1);
	  pCurObjManager->enlever_l_seg (obj_prec);
	  pCurObjManager->enlever_l_seg (obj2);
	  if (pCurObjManager->test_cr_seg (halo_prec, sens) && pCurObjManager->test_cr_seg (halo_succ, sens1))
	    {
	      modif_coord (obj_prec, typ_seg, halo_prec, nbcoord, tcoord, rg);
	      modif_coord (obj1, typ_seg, halo_succ, nbcoord, vcoord1, rg);
	      ntreesDestroy (pseg2);
	      the_graphicTablesManager->enlever_seg (obj2);
	      ntreesDestroy (pseg);
	      the_graphicTablesManager->enlever_seg (obj);
	      align = pk_true;
	      result = pk_true;
	    }
	  else
	    the_messageTool->error_mess (deplimphalo);
	  pCurObjManager->inser_l_seg (obj1);
	  pCurObjManager->inser_l_seg (obj_prec);
	}
      else
	{
	  the_graphicalTrees->calcul_nouv_halo (typ_seg, vcoord, halo);
	  the_graphicalTrees->calcul_nouv_halo (typ_seg, vcoord1, halo_prec);
	  the_graphicalTrees->calcul_nouv_halo (typ_seg, vcoord2, halo_succ);
	  pCurObjManager->enlever_l_seg (obj1);
	  pCurObjManager->enlever_l_seg (obj2);
	  if (pCurObjManager->test_cr_seg (halo, sens) && pCurObjManager->test_cr_seg (halo_prec, sens1) &&
	      pCurObjManager->test_cr_seg (halo_succ, sens1))
	    {
	      modif_coord (obj, typ_seg, halo, nbcoord, vcoord, rg);
	      modif_coord (obj1, typ_seg, halo_prec, nbcoord, vcoord1, rg);
	      modif_coord (obj2, typ_seg, halo_succ, nbcoord, vcoord2, rg);
	      result = pk_true;
	    }
	  else
	    the_messageTool->error_mess (deplimphalo);
	  pCurObjManager->inser_l_seg (obj1);
	  pCurObjManager->inser_l_seg (obj2);
	}
      if (!align && result)
	pCurObjManager->inser_l_seg (obj);
    }
  return (result);
}


/**
  Deplacement d'un chemin par routage entre deux ports ou plusieurs
   si le port de rattachement est un port diffuseur.
   on effectue le routage entre les ports dans le sens port recepteur 
   vers port emetteur. 
   si le routage est possible on modifie le chemin existant si le chemin 
   est simple sinon on detruit le sous_chemin puis on recouvre.
   ?obj : le premier segment a deplacer 
   ?vcoord ses nouvelles coordonnees
   ?port le port de rattachement
   ?intern si les segments sont a l'interieur de la boite
   ?p_bougetail : structure permettant de savoir si la boite a change
   de taille et dans quel sens (hauteur, largeur)
*/
boolean
builder::depl_seg_routage (int segmentGpkIdent, typcoord * vcoord, int portGpkIdent, boolean intern, typcoord * p_bougetail)
{
  boolean b, rout, isInputPort, reception, boucle, p_ext;
  int sens_p, sens_pl, nbcoord, nbrcoord, direc, i;
  int rgl, p_depx, p_depy, noseg, nbports, no_pd, p_rout;
  ntrees p_dep, l_p_dep, c, filsc, portGpkTree, segmentGpkTree;
  typcoord pcoord[2], plcoord[2], rcoord[30], bougetail;

  // printf("DEBUG entering builder::depl_seg_routage segmentGpkIdent is %i, portGpkIdent is %i\n", segmentGpkIdent, portGpkIdent); fflush(stdout);
  bougetail = *p_bougetail;
  portGpkTree = the_graphicTablesManager->graphObj_get_tree (portGpkIdent);
  segmentGpkTree = the_graphicTablesManager->acces_segment (segmentGpkIdent);
  isInputPort = (noper (portGpkTree) == gpkOp_porte);
  reception = ((isInputPort && !intern) || (!isInputPort && intern));
  the_graphicalTrees->getCoord (portGpkTree, &nbcoord, pcoord, 1);
  if (intern)
    {
      if (pcoord[0].x < 0)
	sens_p = droite;
      else
	sens_p = gauche;
    }
  else
    {
      if (pcoord[0].x < 0)
	sens_p = gauche;
      else
	sens_p = droite;
    }
  l_p_dep = the_graphicalTrees->get_ports_lies (portGpkIdent, intern, &c, &boucle);
  nbports = ntreearity (l_p_dep);
  i = 1;
  if (nbports > 0)
    {
      filsc = nchild (1, l_p_dep);
      while (filsc != null)
	{
	  p_dep = the_graphicTablesManager->graphObj_get_tree ((int) nintof (filsc));
	  p_ext = the_graphicalTrees->isExternPort (the_graphicalTrees->boite_mere (portGpkIdent), p_dep);
	  the_graphicalTrees->getCoord (p_dep, &nbcoord, rcoord, 1);
	  if (the_graphicalTrees->boite_mere ((int) nintof (filsc)) == the_graphicalTrees->boite_mere (portGpkIdent))
	    if (the_graphicalTrees->boite_mere (portGpkIdent) != pCurObjManager->get_bte_cou_engl ())
	      boucle = pk_true;
	    else
	      {
		boucle = pk_false;	/* si la bte contenant le portGpkIdent est la bte englobante */
		p_ext = pk_true;	/* alors il s'agit d'une connexion entre 2 ports externes */
	      }
	  else
	    boucle = pk_false;
	  if (!intern && isInputPort && boucle)	/* s'il existe un rebouclage */
	    switch (raction)
	      {
	      case rout_bte:
		// printf("DEBUG exiting builder::depl_seg_routage 1 TRUE\n"); fflush(stdout);
		return pk_true;	/* deplacement d'une boite : le routage se fera en partant */
		break;		/* de l'autre port */
	      case rout_tail_bte:
		if (!((bougetail.y == 0) && ((bougetail.x == -1 && pcoord[0].x < 0 && rcoord[0].x > 0)
					     || (bougetail.x == 1 && pcoord[0].x > 0 && rcoord[0].x < 0))))
		  {
		    // printf("DEBUG exiting builder::depl_seg_routage 2 TRUE\n"); fflush(stdout);
		    return pk_true;
		  }		/* pas de routage si on change la taille de l'autre cote du port */
	      }
	  the_graphicalTrees->getCoord (p_dep, &nbcoord, plcoord, 0);
	  pCurObjManager->reper_bte (&plcoord[0].x, &plcoord[0].y, &sens_pl, &rgl, 0, &b);
	  if (p_ext)
	    {
	      get_coord_seg_port (p_dep, typ_port_ext, &p_depx, &p_depy);
	      the_graphicalTrees->getCoord (p_dep, &nbcoord, plcoord, 1);
	      if (plcoord[0].x < 0)
		sens_pl = droite;
	      else
		sens_pl = gauche;
	    }
	  else
	    get_coord_seg_port (p_dep, typ_port, &p_depx, &p_depy);
	  /* on effectue le routage entre les ports dans le sens port recepteur vers port 
	     emeteur; on affecte p_rout a 0 si le chemin est simple */
	  no_pd = the_graphicalTrees->tree_get_graphObj (p_dep);
	  if (reception)
	    {
	      if (the_graphicalTrees->chemin_simple (segmentGpkTree))
		p_rout = 0;
	      else
		p_rout = no_pd;
	      rout = pRoutor->applyRoutor (vcoord[1].x, vcoord[1].y, sens_p, p_depx, p_depy, sens_pl, rcoord, &nbrcoord, p_rout, &noseg);
	    }
	  else
	    {
	      if (i == 1)
		p_rout = 0;
	      else
		p_rout = portGpkIdent;
	      rout = pRoutor->applyRoutor (p_depx, p_depy, sens_pl, vcoord[0].x, vcoord[0].y, sens_p, rcoord, &nbrcoord, p_rout, &noseg);
	    }
	  if (rout)
	    {			/* on modifie le chemin si le chemin est simple */
	      /* sinon on detruit puis recouvre */
	      if (reception)
		if (the_graphicalTrees->chemin_simple ((segmentGpkTree)))
		  modifier_chemin (no_pd, nbrcoord, rcoord);
		else
		  {
		    detruire_chemin (nfather (nfather (segmentGpkTree)), pk_false);
		    inverser (rcoord, nbrcoord);
		    if (rcoord[nbrcoord - 1].x == rcoord[nbrcoord].x)
		      direc = vertic;
		    else
		      direc = horiz;
		    nbrcoord++;
		    if (!recouv_lien (no_pd, portGpkIdent, &nbrcoord, rcoord, direc, pk_true))
		      {
			// printf("DEBUG exiting builder::depl_seg_routage 3 FALSE\n"); fflush(stdout);
			return pk_false;
		      }
		  }
	      else if ((i == 1) && the_graphicalTrees->chemin_simple (segmentGpkTree))
		modifier_chemin (portGpkIdent, nbrcoord, rcoord);
	      else if (i == 1)
		{
		  detruire_chemin (nfather (nfather (segmentGpkTree)), pk_false);
		  createchemin (portGpkIdent, no_pd, nbrcoord, rcoord);
		}
	      else
		{
		  inverser (rcoord, nbrcoord);
		  if (rcoord[nbrcoord - 1].x == rcoord[nbrcoord].x)
		    direc = vertic;
		  else
		    direc = horiz;
		  nbrcoord++;
		  if (!recouv_lien (portGpkIdent, no_pd, &nbrcoord, rcoord, direc, pk_true))
		    {
		      // printf("DEBUG exiting builder::depl_seg_routage 4 FALSE\n"); fflush(stdout);
		      return pk_false;
		    }
		}
	    }
	  else
	    {
	      // printf("DEBUG exiting builder::depl_seg_routage 5 FALSE\n"); fflush(stdout);
	      return pk_false;
	    }
	  filsc = ntrees_getNext (filsc);
	  i++;
	}
    }
  // printf("DEBUG exiting builder::depl_seg_routage * TRUE\n"); fflush(stdout);
  return pk_true;
}


/**
  Deplacement d'un segment appartenant a un chemin multiple (on a un chemin
  multiple lorsque du port diffuseur partent plusieurs connexions). Cette
  fonction ne traite pas le cas d'un segment dont le predecesseur et le
  successeur appartiennent au meme chemin (traite par la fonction depl_seg)
  Pour le moment, la fonction n'esp pas implemente : rend faux
*/
boolean
builder::depl_seg_chem_mult (int obj, typcoord * tcoord)
{
  boolean result;

  result = pk_false;
  return (result);
}


/**
 * Cette procedure verifie qu'une liste de segments (dans un chemin) ne
 * comporte pas de segment de longueur nulle. 
 * ?lseg : liste de segments
 * ?intern : si intern est a faux, on enleve le segment de la liste des
 *           segments designables.
 * Cette procedure est appelee par la fonction detruire_chemin (reperage_g.c)
 * cette procedure n'aurait pas lieu d'etre si le deplacement marchait bien,
 * or, il cree parfois des segments de longueur nulle (des points).
 */
void
builder::verif_lseg (ntrees lseg, boolean intern)
{
  int id_seg, i, nbcoord;
  ntrees seg;
  typcoord tcoord[10];

  i = 1;
  while (i <= ntreearity (lseg))
    {
      seg = nchild (i, lseg);
      the_graphicalTrees->getCoord (nchild (2, seg), &nbcoord, tcoord, 0);
      if (tcoord[0].x == tcoord[1].x && tcoord[0].y == tcoord[1].y)
	{
	  id_seg = the_graphicalTrees->tree_get_graphObj (seg);
	  if (!intern)
	    pCurObjManager->enlever_l_seg (id_seg);
	  the_graphicTablesManager->enlever_seg (id_seg);
	  ndestroyListElem (lseg, i);
	}
      else
	i++;
    }
}
