/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
///_PUBLIC__

routor::routor ()
{
}

///
void
routor::initRoutor (ptCurrentObjectsManager pLoc)
{
  pCurObjManager = pLoc;
}

/**
  Fonction appelee lors de la creation d'un lien ou lors d'un deplacement
  de port ou de boite ou lors d'un changement de taille d'une boite.
  S'il n'existe pas de chemin direct entre les points A ?(xa,ya) et
  B ?(xb,yb), elle appelle la fonction calcul_routage.
  Le point A est affecte d'un sens de depart ?sensa et le point B d'un
  sens d'arrivee sur ce point ?sensb. 
  A est toujours le port recepteur et  B le port diffuseur.
  Si cette derniere rend faux, c'est qu'elle n'a pas trouve de chemin, alors
  routage rend faux ; sinon elle remonte l'arbre des chemins  a partir
  de l'arbre courant et insere toutes les coordonnees dans le tableau 
  !tcoord_lien, rend le nombre de coordonnees dans !nbcoord.
  ?port_diffuseur : contient le numéro du port diffuseur s'il est déjà connecté
  avec un autre port, sinon vaut 0.
  !noseg : rend le numero du segment commun au lien qu'on vient de calculer et
           au lien deja connecte au port diffuseur ?port_diffuseur (uniquement
           si port_diffuseur <> 0)
*/
boolean
  routor::applyRoutor (int xa, int ya, int sensa, int xb, int yb, int sensb, typcoord * tcoord_lien, int *nbcoord, int port_diffuseur, int *noseg)
{
  int i, x, y;
  ntrees p;
  boolean ok;

  ntrees_memoryStoreIndex ();
  init_routage (xa, ya, sensa, xb, yb, sensb);
  port_recouv = port_diffuseur;
  ok = chem_direct (xa, ya, sensa, xb, yb, sensb);
  if (ok && (port_recouv == 0))
    {
      tcoord_lien[0].x = xb;
      tcoord_lien[0].y = tcoord_lien[1].y = yb;
      if (ya == yb)
	{
	  tcoord_lien[1].x = xa;
	  *nbcoord = 1;
	}
      else
	{
	  tcoord_lien[1].x = tcoord_lien[2].x = (xb + xa) / 2;
	  tcoord_lien[2].y = tcoord_lien[3].y = ya;
	  tcoord_lien[3].x = xa;
	  *nbcoord = 3;
	}
    }
  else
    {
      switch (sensb)
	{
	case gauche:
	  xb = xb - taille_virage;
	  break;
	case droite:
	  xb = xb + taille_virage;
	};
      if (!ok)
	/* si les points sont trop pres des bords et qu'il n'y a pas de chemin direct,
	   on ne lance pas le routage, on considere qu'il n'y a pas de chemin */
	if ((((xa + xb) / 2) >= abs_mil_cadre + lg_bte_stand - taillepave_x) || (((xa + xb) / 2) <= abs_mil_cadre - lg_bte_stand + taillepave_x))
	  return (ok);
	else
	  ok = calcul_routage (xa, ya, sensa, xb, yb, sensb);
      if (ok)
	{
	  if (port_recouv != 0 && seg_recouv != 0)
	    i = 0;
	  else
	    {
	      tcoord_lien[0].x = x_arr;
	      tcoord_lien[0].y = y_arr;
	      tcoord_lien[1].x = xb;
	      tcoord_lien[1].y = yb;
	      i = 2;
	    };
	  p = routage_cour;
	  while (p != racine_routage)
	    {
	      x = (int) nintof (nchild (1, p));
	      y = (int) nintof (nchild (2, p));
	      if (i > 1 && y == tcoord_lien[i - 1].y && y == tcoord_lien[i - 2].y)
		tcoord_lien[i - 1].x = x;
	      else
		{
		  tcoord_lien[i].x = x;
		  tcoord_lien[i].y = y;
		  i++;
		}
	      p = nfather (nfather (p));
	    };
	  if (ya == tcoord_lien[i - 1].y && ya == tcoord_lien[i - 2].y)
	    {
	      tcoord_lien[i - 1].x = xa;
	      i--;
	    }
	  else
	    {
	      tcoord_lien[i].x = xa;
	      tcoord_lien[i].y = ya;
	    }
	  *nbcoord = i;
	  if (port_recouv != 0)
	    *noseg = seg_recouv;
	}
    }
  ntrees_memoryRestoreIndex ();	//  ndegeler_recup();
  return (ok);
}
