/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */

_PRIVATE_
#define endFuncMess 54
#define NOTHING -2		///<valeur rendue lorsque l'utilisateur n'a rien cliqué dans les fenêtres de compilation
#define END -1			///< valeur rendue lorsque l'utilisateur a cliqué sur le bouton "END"
// The following constants must be the same than the constants declared in CompilerConst.java
#define nb_func    11		///< nbre de fonctionalités du compilo visibles de l'utilisateur (hors générateur)
#define nb_func_gen 7		///< nbre de fonctionalités concernant la génération de code
#define AllFuncInteractive (nb_func+nb_func_gen)
static ptDCG_basic vgraph_auto;	/** Graphe en mode compilation automatique.*/
/*
    typedef enum{func_noopt,func_init,func_retiming,func_booleans2events,func_clockCalculus,func_trains,func_abstraction,func_booleanizing,func_flattening,
    func_serializing,func_gen_SIGNAL,  func_gen_SSME, func_poly2endo, func_gen_DCPlusSubFormat,func_gen_CANSI,func_gen_CPP,func_gen_JAVA,func_gen_Syndex,func_gen_Z3Z,func_unify,func_gen_abstract,
    func_profiling,func_end}function_type;
*/
// The following table must be conform to the graphical view.
static function_type array_of_functions[AllFuncInteractive] = {
  func_retiming,
  func_booleans2events,
  func_unify,
  func_clockCalculus,
  func_poly2endo,
  func_booleanizing,
  func_abstraction,
  func_trains,
  func_serializing,
  func_flattening,
  func_noopt,
  func_gen_SIGNAL,
  func_gen_abstract,
  func_gen_Z3Z,
  func_gen_CANSI,
  func_gen_CPP,
  func_gen_JAVA,
  func_profiling
};

static func_type_color status_of_functions[AllFuncInteractive];
static ptBatchCompiler the_pKBatchCompiler; /** Signal Batch Compiler (one instance). Created once with the first instance of supervis */

/**
  Pre-traitement avant la compilation de l'arbre Signal (sourceSubtree).
  @param typ_comp : type de la compilation demandée par l'utilisateur (compil normale, compil de type embarquée, compil d'une architecture).
  Si sourceSubtree n'est pas une racine (on n'est pas tout en haut de l'arbre),
  alors il s'agit d'une compilation partielle.
  -  on modifie l'arbre Signal (rajout des affectations pour les noms de
     ports, nettoyage de l'arbre... affectues par la fonction "completeTree") après avoir fait une copie.
  -  on demande les parametres si necessaire.
  Rend l'arbre modifié.
*/
ntrees
supervis::ok_comp (ntrees sourceSubtree, integer typ_comp)
{
  ptrepstr str_nom_prog;
  ntrees father, resSubtree;
  boolean is_archi;
  char *fich_sig;

  // nDEBUG_subtree(4, 8, "entering supervis::ok_comp sourceSubtree is", sourceSubtree);
  is_archi = (typ_comp == ty_archi_comp);
  if (is_archi)
    {
      resSubtree = pBuilder->buildArchi (sourceSubtree);
      if (IS_NILTREE (resSubtree))
	{			/* on n'a pas pu construire l'architecture */
	  ok_pre_comp = pk_true;	/* the pre compilation treatment is finished */
	  err_pre_comp = pk_true;
	  err_nom_vide = pk_true;	/* pour ne pas faire de restauration après */
	  the_messageTool->error_mess (archi_not_built);
	  // nDEBUG_subtree(4, 8, " exiting supervis::ok_comp 1 sourceSubtree is", sourceSubtree);
	  // nDEBUG_subtree(4, 8, " exiting supervis::ok_comp 1 resSubtree is", resSubtree);
	  return resSubtree;
	}
    }
  else
    resSubtree = sourceSubtree;
  fich_sig = repstr2string (STreeModelModule_getNameStr (resSubtree));

  if (the_graphicalTrees->GraphicSaving (nom_fich_compil, resSubtree))
    {
      father = STree_getModelFather (resSubtree);
      ntrees xxx = the_graphicalTrees->getRoot ();
      if ((father != xxx) && !STree_isModule (father))
	pGraph_text->prepareLocalCompilation (resSubtree);
      resSubtree = pGraph_text->completeTree (resSubtree);
      if (is_archi || (typ_comp == ty_embed_comp))
	{			/* pour le TEST */
	  if (is_archi)
	    str_nom_prog = string2repstr ("archi");
	  else
	    str_nom_prog = string2repstr ("embed");
	  makeRepstr_repstr_string (str_nom_prog, ".SIG", &str_nom_prog);
	  filebuf->setsize (margeGaucheMax, lgligne);
	  if (!pK_SignalPrinter->fileUnparse (resSubtree, filebuf, str_nom_prog))
	    {
	      ok_pre_comp = pk_true;	/* le traitement pre-compil est fini */
	      err_pre_comp = pk_true;
	      the_messageTool->error_mess (ERR_unparseFile);
	      return NILTREE;
	    }
	}
      if (STree_isModelWithoutParameters (resSubtree))
	{			/* there is no parameter : the pre compilation treatment is finished */
	  ok_pre_comp = pk_true;
	  err_pre_comp = pk_false;
	  SET_NILTREE (p_param);
	}
      else
	{
	  err_pre_comp = pk_false;
	  ok_pre_comp = pk_false;	/* will be set to true when the user will click on the OK button 
				   of the the OK button of the parameters window */
	  sprintf (fich_par, "%s.PAR", fich_sig);
	  displayParam (fich_par);
	  src_fich = fich_sig;
	}
      // nDEBUG_subtree(4, 8, " exiting supervis::ok_comp 2 sourceSubtree is", sourceSubtree);
      // nDEBUG_subtree(4, 8, " exiting supervis::ok_comp 1 resSubtree is", resSubtree);
      return resSubtree;
    }

  /* GraphicSaving failed */
  ok_pre_comp = pk_true;		/* le traitement pre-compil est fini */
  err_pre_comp = pk_true;
  the_messageTool->error_mess (ERR_fileOpening);
  return resSubtree;
}

/**
    Compilation interactive: indique qu'une fonctionalité (func) ne peut pas être appliquée.
*/
void
supervis::FORBIDDEN (function_type func)
{
  func_modify (get_index_function (func), forbidden);
}


/**
    Compilation interactive: indique qu'une fonctionalité (func) peut être appliquée.
*/
void
supervis::ALLOWED (function_type func)
{
  func_modify (get_index_function (func), allowed);
}

/**
   Compilation interactive: indique qu'une fonctionalité (func) n'a plus d'effet (sans effet sur le graphe).
*/
void
supervis::WITHOUTEFFTECT (function_type func)
{
  func_modify (get_index_function (func), withouteffect);
}

/**
   Compilation interactive.
   @return l'index d'une fonction dans le tableau des fonctionalités offertes à l'utilisateur.
*/
int
supervis::get_index_function (function_type vfunction)
{
  int i;

  for (i = 0; i < AllFuncInteractive; i++)
    if (array_of_functions[i] == vfunction)
      return i;
  return -1;
}


/**
   
@see pKServicesAPI::IGraphBooleanizing
boolean supervis::booleanizingAndPruning(ptDCG_xxx vgraph)
{
  if (!pKServicesAPI::IGraphBooleanizing(vgraph)) return pk_false;
  Pruning(vgraph);
  return pk_true;
}
*/
