 /*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */

import java.awt.Container;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;


/**
 * This class describes a window to display the different functionalities of the Signal compiler
 */
public class CompilerFunct extends JFrame {


    private static final long serialVersionUID = 1L;
    protected static final  String[] labelFunct = new String[CompilerConst.AllFunc];
    protected static final int Red = 3;
    protected static final int Green = 4;
    protected static final int Yellow = 6;
    static {
        labelFunct[0] = "  Retiming";
        labelFunct[1] = "  Booleans to events";
        labelFunct[2] = "  Signals Unifications";
        labelFunct[3] = "  Basic -> Poly";
        labelFunct[4] = "  Poly -> Endo";
        labelFunct[5] = "  Endo -> Booleans";
        labelFunct[6] = "  Abstraction";
        labelFunct[7] = "  Clustering";
        labelFunct[8] = "  Sequentializing";
        labelFunct[9] = "  Flattening ";
        labelFunct[10] = " **Generators** ";
        labelFunct[11] = " SIGNAL";
        labelFunct[12] = " SIGNAL ABSTRACTION";
        labelFunct[13] = " Z/3Z(sigali)";
        labelFunct[14] = " C ANSI";
        labelFunct[15] = " C++";
        labelFunct[16] = " Java";
        labelFunct[17] = " Profiling";
    }

    protected JCheckBox[] tfunc;
    protected CompilerWin compWin; // reference of the compiling window

    // draw in ?col (Yellow, Red, Green) the text associated with the
    // functionality ?func
    public void setColorFunct(int func, int col) {
        switch (col) {
        case Yellow:
            tfunc[func].setForeground(Color.orange);
            tfunc[func].setEnabled(true);
            break;
        case Red:
            tfunc[func].setEnabled(false);
            tfunc[func].setBorderPaintedFlat(false);
            tfunc[func].setForeground(Color.red);
            break;
        case Green:
            tfunc[func].setForeground(Color.green);
            tfunc[func].setEnabled(true);
            break;
        }
    }


    // re-initialization from the first to the last CheckBox elements
    private void resetCheckBox(int first, int last) {
        for (int i = first; i < last; i++)
            tfunc[i].setSelected(false);
    }

    public void actionPerformedInteractiveCompiler(String label) {
        if (label == "END") {
            // button of the functionalities window in interactive mode
            resetCheckBox(0, CompilerConst.nbCompFunc);
            resetCheckBox(CompilerConst.nbCompFunc,
                          CompilerConst.AllFuncInteractive);
            compWin.endInteractCompiling();
        } else
            // button of the generators window
            if (label == "CLOSE")
                resetCheckBox(CompilerConst.nbCompFunc - 1,
                              CompilerConst.nbCompFunc);
    }

    public void itemStateChangedInteractiveCompiler(ItemEvent e) {
        JCheckBox source = (JCheckBox) e.getItemSelectable();
        int i = 0;

        boolean found = false;
        while (!found) {
            found = (tfunc[i] == source);
            if (!found)
                i++;
        }
        if (i < CompilerConst.AllFuncInteractive)
            if (i == (CompilerConst.nbCompFunc - 1))
                compWin.itemStateChangedCodeGenerators(e.getStateChange() == ItemEvent.DESELECTED);
            else
                // a functionality has been selected: apply it.
                if (e.getStateChange() != ItemEvent.DESELECTED)
                    compWin.applyFunctionality(i);
    }

    public void see(boolean b) {
        super.setVisible(b);
    }

    public CompilerWin getCompilerWin() {
        return compWin;
    }

    public void functClose() {
        // to close the window
        WindowListener closeWin = new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                setVisible(false); // close the window
                compWin.endInteractCompiling();
            }
        };
        addWindowListener(closeWin);
    }




    protected  void setButtons(int deb, int fin, CheckBoxListener myListener, JPanel funcPanel) {
        int i;
        for (i = deb; i <= fin; i++) {
            JCheckBox button = new JCheckBox(labelFunct[i]);
            button.addItemListener(myListener);
            button.setBackground(Color.white);
            funcPanel.add(button);
            tfunc[i] = button; // the button reference are saved
        }
    }

    private Container Common(CompilerWin cw, String vLabel, String opLabel, JPanel funcPanel, String vtitle) {

        this.compWin = cw;

        Container frame = getContentPane();
        frame.setLayout(new BorderLayout());
        JPanel title = new JPanel();
        setTitle(vtitle);
        title.setLayout(new BorderLayout());
        JLabel title1 = new JLabel(vLabel);
        title1.setForeground(Color.black);
        title.add(title1, BorderLayout.NORTH);
        title1.setFont(new Font("Helvetica", Font.PLAIN, 20));
        boolean isOpt = (opLabel != "");
        if (isOpt) {
            JLabel title2 = new JLabel(opLabel);
            title2.setForeground(Color.black);
            title.add(title2, BorderLayout.SOUTH);
        }
        title.setBorder(BorderFactory.createMatteBorder(20, 20, 20, 20,  pKGUIColors.Color1 ));

        // Creation of the checkboxes
        funcPanel.setLayout(new GridLayout(0, 1));
        funcPanel.setBackground(Color.white);
        funcPanel.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
        frame.add(title, BorderLayout.NORTH);
        frame.add(funcPanel, BorderLayout.CENTER);
        return (frame);
    }


    // Constructor
    public CompilerFunct(LiaisonSignal l, PolychronyWindow gw, CompilerWin cw,
                         int x0, int y0, int deb, int fin, String butLabel, String vjLabel,
                         String opLabel, String vtitle, JCheckBox[] xtfunc) {
        JPanel panel;
        tfunc = xtfunc;
        panel = new JPanel();
        Container frame = Common(cw, vjLabel, opLabel, panel, vtitle);
        CheckBoxListener myListener = new CheckBoxListener(this);
        setButtons(deb, fin, myListener, panel);
        // Build of a panel with a single button
        panel = new JPanel();
        JButton endButton = new JButton(butLabel);
        endButton.addActionListener(myListener);
        panel.add(endButton, BorderLayout.CENTER);
        frame.add(panel, BorderLayout.SOUTH);
        setLocation(x0, y0);
        pack();
        functClose();
    }

}

class CheckBoxListener implements ItemListener, ActionListener {
    CompilerFunct parent;

    public CheckBoxListener(CompilerFunct w) {
        parent = w;
    }

    public void itemStateChanged(ItemEvent e) {
        parent.itemStateChangedInteractiveCompiler(e);
    }

    public void actionPerformed(ActionEvent e) {
        JButton source = (JButton) e.getSource();
        parent.setVisible(false);
        parent.actionPerformedInteractiveCompiler(source.getText());
    }

}
