 /*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */


 

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.GraphicsEnvironment;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.ListSelectionModel;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;


class InternalChooseFont extends JFrame implements ListSelectionListener, ItemListener {
    private static final long serialVersionUID = 1L;
    private JTextArea exemple = new JTextArea(5, 10); // Text zone of the example
    private String fontSelect; // Name of the selected font
    private Integer sizeSelect = new Integer(12); // For the conversion string to int
    private int tailleFont = 12; // Size of the font
    private int style = Font.PLAIN; // Style of the font: bold, italic... default = nothing
    private String contenu = ("ExEmPlE oF tHe sElEcTeD fOnT\n0123456789\naAbBcCdDeEfFgGhHiIjJkKlLmMn" + "NoO\npPqQrRsStTuUvVwWxXyYzZ"); //  Content of the example

    private Font fontSelected; // Selected Font
    private Font fontByDefault; // Previous Font

    // Creation of the checkboxes to select bold or italic
    private JCheckBox bold = new JCheckBox("bold", false);
    private JCheckBox italic = new JCheckBox("italic", false);
    // ComboBox for the size List
    private String[] sizeListe = { "2", "4", "6", "8", "10", "12", "14", "16", "18", "20", "22", "24", "30", "36", "50", "72" };
    private JComboBox size = new JComboBox(sizeListe);
    JButton ok;

    private Color getMyColor1() {
        return pKGUIColors.Color1; // new Color(200, 200, 230);
    }

    public InternalChooseFont(Font defaultFont, String fontName, int fontSize) {
        setTitle("Set Font...");
        setLocation(100, 100);
        setSize(450, 250);
        // saves the value of the old font
        fontByDefault = new Font(defaultFont.getAttributes());
        // Global Panel
        JPanel content = (JPanel) getContentPane();
        //  Color of the window's background
        content.setBackground(getMyColor1());
        content.setLayout(new GridBagLayout());
        content.setBorder(new EmptyBorder(new Insets(5, 5, 5, 5)));

        GridBagConstraints contentC = new GridBagConstraints();
        contentC.insets = new Insets(0, 0, 0, 0);
        contentC.gridx = 0;
        contentC.gridy = 0;
        contentC.gridwidth = 1;
        contentC.gridheight = 1;
        contentC.weighty = 1.0;
        contentC.weightx = 1.0;
        contentC.fill = GridBagConstraints.BOTH;
        contentC.anchor = GridBagConstraints.CENTER;

        //  Panel containing the selection
        JPanel select = new JPanel();
        select.setLayout(new GridBagLayout());
        select.setBorder(new EtchedBorder(EtchedBorder.LOWERED));
        select.setBackground(getMyColor1());

        // Position constraints on the buttons panel
        GridBagConstraints selectC = new GridBagConstraints();
        selectC.insets = new Insets(5, 5, 5, 5);
        selectC.gridx = 0;
        selectC.gridy = 0;
        selectC.gridwidth = 1;
        selectC.gridheight = 3;
        selectC.weighty = 1.0;
        selectC.weightx = 1.0;
        selectC.fill = GridBagConstraints.BOTH;
        selectC.anchor = GridBagConstraints.CENTER;

        // gets the fonts list
        GraphicsEnvironment gE = GraphicsEnvironment.getLocalGraphicsEnvironment();
        String[] listeFont = gE.getAvailableFontFamilyNames();

        // creation of the liste
        JList list = new JList(listeFont);
        list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        // adds a scroll to the list
        JScrollPane listScrollPane = new JScrollPane(list);
        listScrollPane.setPreferredSize(new Dimension(200, 80));
        list.setSelectedValue(fontName, true);
        list.addListSelectionListener(this);
        select.add(listScrollPane, selectC);
        fontSelect = (String) list.getSelectedValue();

        // Properties of the JComboxBox that selects the size
        size.setEditable(true);

        size.setSelectedItem(new Integer(fontSize).toString());
        size.setMaximumRowCount(6);
        size.addItemListener(this);

        selectC.gridx = 1;
        selectC.gridy = 0;
        selectC.gridwidth = 1;
        selectC.gridheight = 1;
        selectC.weighty = 0.0;
        selectC.weightx = 0.0;
        selectC.fill = GridBagConstraints.NONE;
        select.add(size, selectC);

        // Properties of the BOLD JCheckButton
        bold.setBackground(getMyColor1());
        selectC.gridx = 1;
        selectC.gridy = 1;
        selectC.gridwidth = 1;
        selectC.gridheight = 1;
        selectC.weighty = 0.0;
        selectC.weightx = 0.0;
        selectC.fill = GridBagConstraints.NONE;
        select.add(bold, selectC);
        bold.addItemListener(this);

        // Properties of the ITALIC JCheckButton
        italic.setBackground(getMyColor1());
        selectC.gridx = 1;
        selectC.gridy = 2;
        selectC.gridwidth = 1;
        selectC.gridheight = 1;
        selectC.weighty = 0.0;
        selectC.weightx = 0.0;
        selectC.fill = GridBagConstraints.NONE;
        select.add(italic, selectC);
        italic.addItemListener(this);

        // Creation of the ok and cancel buttons
        ok = new JButton("ok");
        selectC.gridx = 2;
        selectC.gridy = 0;
        selectC.gridwidth = 1;
        selectC.gridheight = 1;
        selectC.weighty = 1.0;
        selectC.weightx = 1.0;
        selectC.fill = GridBagConstraints.BOTH;
        select.add(ok, selectC);

        JButton cancel = new JButton("cancel");
        selectC.gridx = 2;
        selectC.gridy = 1;
        selectC.gridwidth = 1;
        selectC.gridheight = 1;
        selectC.weighty = 1.0;
        selectC.weightx = 1.0;
        selectC.fill = GridBagConstraints.BOTH;
        select.add(cancel, selectC);

        ActionListener actCancel = new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                // the selected font is the old font
                fontSelected = new Font(fontByDefault.getAttributes());
                // the font from the calling window isn't modified
                setVisible(false);
            }
        };
        cancel.addActionListener(actCancel);
        content.add(select, contentC);
        exemple.setWrapStyleWord(true);
        exemple.append(contenu);

        // adds a scroll to the text zone of the example
        JScrollPane scrollEx = new JScrollPane(exemple);

        tailleFont = (sizeSelect.decode((String) size.getSelectedItem())).intValue(); // or  Integer.decode ?
        fontSelected = new Font(fontName, style, tailleFont);
        exemple.setFont(fontSelected);

        contentC.gridx = 0;
        contentC.gridy = 1;
        contentC.gridwidth = 1;
        contentC.gridheight = 1;
        contentC.weighty = 1.0;
        contentC.weightx = 1.0;
        contentC.fill = GridBagConstraints.BOTH;
        content.add(scrollEx, contentC);
        setVisible(true);

        WindowListener wndCloser = new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
                InternalChooseFont cw = (InternalChooseFont) e.getSource();
                cw.fontSelected = new Font(fontByDefault.getAttributes());
                cw.setVisible(false);
            }
        };
        addWindowListener(wndCloser);
    }

    /** Changes the type of the font of the example with the font selected in the list */
    public void valueChanged(ListSelectionEvent evt) {
        if (evt.getValueIsAdjusting())
            return;
        fontSelect = ((String) ((JList) evt.getSource()).getSelectedValue());
        fontSelected = new Font(fontSelect, style, tailleFont);
        exemple.setFont(fontSelected);
        contenu = exemple.getText(); // gets the content of the example
        exemple.setText(contenu);
    }

    /** 
	Events of the combobox (size) and the checkbuttons (bold, italic) changes 
	the size of the font with the combobox changes the style of the font with
	the checkbuttons 
    */
    public void itemStateChanged(ItemEvent evt) {
        // determines the object doing the action
        Object source = evt.getSource();
        if (source == size) {
            // string to int transformation to get the selected size
            tailleFont = (sizeSelect.decode((String) ((JComboBox) evt.getSource()).getSelectedItem())).intValue(); // or Integer.decode
            fontSelected = new Font(fontSelect, style, tailleFont);
            exemple.setFont(fontSelected);
            contenu = exemple.getText();
            exemple.setText(contenu);
        } else if (source == bold || source == italic) {
            // the style is modified from the bold and/or italic choice
            style = Font.PLAIN;
            if (bold.isSelected() && italic.isSelected()) {
                style = Font.BOLD + Font.ITALIC;
            } else if (bold.isSelected()) {
                style += Font.BOLD;
            } else if (italic.isSelected())
                style += Font.ITALIC;

            fontSelected = new Font(fontSelect, style, tailleFont);
            exemple.setFont(fontSelected);
            contenu = exemple.getText();
            exemple.setText(contenu);
        }
    }

    public Font getFont() {
        return fontSelected;
    }

    /**  Gets the width of a character in pixels */
    public int getWidth(Font fonte) {
        FontMetrics fmetrics = getFontMetrics(fonte);
	// see the Java API: charWidth doesn't really return the width in pixels of a character
        return (fmetrics.charWidth('L'));
    }

    /** Gets the height of a character in pixels */
    public int getHeight(Font fonte) {
        FontMetrics fmetrics = getFontMetrics(fonte);
        return fmetrics.getHeight();
    }

}
