/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
/*-----------------------------------------------------------------------
                       Module de gestion du "undo"
   
  Les fonctions de ce module sont appelées lorsque l'utilisateur clique
  lui-même sur le bouton "undo" du bandeau ou lorsqu'une action de
  déplacement, justification de ports, changement de taille, 
  encapsulation, décapsulation  a échoué.
  Pour le moment, le "undo" ne permet de revenir qu'à l'état précédent
  (pas de pile de undo).

  Toutes les fonctions de ce module qui commencent par get_undo_... sont
  appelées avant de faire les modifications dans les arbres.
  Toutes les fonctions qui commencent par set_undo_... sont appelées,
  soit au moment où l'utilisateur clique sur "undo", soit lorsque l'action
  courante (déplacement, justification...) a échoué.
 ----------------------------------------------------------------------*/

typedef struct {
  objet_type typ_obj_recup; /* type de l'objet sauvegardé : bte, port, segment, texte */
  int        action;     /* no de la dernière action sur laquelle on pourra faire un undo */
  int        oper_recup; /* operateur Signal de l'arbre sauvegardé */
  int        obj_recup;  /* numéro du port sauvegardé, utilisé uniquement 
			    pour faire le undo sur les déclarations associées aux ports */
  int        niveau_recup;/* vaut 1 ou 0, suivant qu'on se trouve a niveau d'un modèle ou non */
  int        recup_rg;    /* position de l'arbre qu'on sauvegarde par rapport à son père : ième 
			     fils */
  ntrees     arb_recup;   /* copie du sous-arbre sur lequel on veut pouvoir faire un "undo" */
  ntrees     p_recup;     /* pointeur sur le s/arbre sur lequel on veut pouvoir faire un "undo" */
  ntrees     attr_recup;  /* copie de l'attribut graphique lorsqu'on sauvegarde du texte,
			     c'est-à-dire lorsqu'on sauvegarde un s/arbre Signal */
  ntrees     arb_recup_inte; /* copie de l'interface du modèle, lorsqu'on a modifié ses ports */
				
}ty_undo;

typedef ty_undo *pt_ty_undo;


class undo{

#include "undo_m.h"
#include "undo_g.h"

  private :
  ty_undo info_undo;
  ptGraphicTablesManager the_graphicTablesManager;
  ptGraphicTrees the_graphicalTrees;
  ptCurrentObjectsManager pCurObjManager;
};
